/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - ZipAssemblingService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.assembling.zip;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;
import javax.xml.bind.JAXBException;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.ZipTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.assembling.AssemblingService;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.AccessController;
import ch.dlcm.model.Access;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.access.Order.OrderStatus;
import ch.dlcm.model.access.QueryType;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.service.DipStatusService;
import ch.dlcm.service.MetadataService;

@Service
@ConditionalOnBean(AccessController.class)
public class ZipAssemblingService extends AssemblingService {

  private final DipStatusService dipStatusService;
  private final MetadataService metadataService;

  public ZipAssemblingService(DLCMProperties dlcmProperties, MessageService messageService,
          DipStatusService dipStatusService, MetadataService metadataService) {
    super(dlcmProperties, messageService);
    this.dipStatusService = dipStatusService;
    this.metadataService = metadataService;
  }

  @Override
  public void checkOrder(String orderId) {
    // Check if Order exists
    if (!FileTool.checkFile(this.getOrderFolder(orderId))) {
      throw new SolidifyCheckingException(
              this.messageService.get("checking.order.notfound", new Object[] { this.getOrderFolder(orderId).toString() }));
    }
    // List Organizational Units
    for (final Path orgUnitPath : FileTool.scanFolder(this.getOrderFolder(orderId), Files::isDirectory)) {
      String orgUnitId = orgUnitPath.getFileName().toString();
      // List access levels
      for (final Path accessPath : FileTool.scanFolder(orgUnitPath, Files::isDirectory)) {
        Access accessLevel = Access.valueOf(accessPath.getFileName().toString());
        String dipId = this.findDipId(orderId, orgUnitId, accessLevel);
        URI dipURI = this.getDipUri(orderId, orgUnitId, accessLevel, dipId);
        // Check if Order archive exists
        if (!FileTool.checkFile(dipURI)) {
          throw new SolidifyCheckingException(this.messageService.get("checking.order.noarchive", new Object[] { dipURI }));
        }
        // Check order archive
        final ZipTool zipTool = new ZipTool(dipURI);
        if (!zipTool.checkZip()) {
          throw new SolidifyCheckingException(this.messageService.get("checking.zip.notvalid", new Object[] { dipURI }));
        }
      }
    }
  }

  @Override
  public List<DisseminationInfoPackage> createDipPackages(Order order) {
    final List<DisseminationInfoPackage> dipList = new ArrayList<>();
    int count = 0;
    // Sort AIP by orgUnit
    final Map<String, List<ArchivalInfoPackage>> aipsByOrgUnit = this.getAipByOrgUnit(order);
    // For orgUnits
    for (final Entry<String, List<ArchivalInfoPackage>> entryByOrgUnit : aipsByOrgUnit.entrySet()) {
      final String orgUnitId = entryByOrgUnit.getKey();
      final List<ArchivalInfoPackage> aipList = entryByOrgUnit.getValue();
      // Sort AIP by access
      final Map<Access, List<ArchivalInfoPackage>> aipsByAccess = aipList.stream().collect(Collectors.groupingBy(a -> a.getInfo().getAccess()));
      for (final Entry<Access, List<ArchivalInfoPackage>> entryByAccess : aipsByAccess.entrySet()) {
        count++;
        final Access access = entryByAccess.getKey();
        final String dipId = this.findDipId(order.getResId(), orgUnitId, access);
        final List<ArchivalInfoPackage> aip = entryByAccess.getValue();
        // Determinate DIP name
        String name = order.getName();
        if (!order.getQueryType().equals(QueryType.DIRECT)) {
          name += " #" + count;
        }
        final DisseminationInfoPackage dip = this.dipStatusService.createDIP(dipId, orgUnitId, name, access, aip,
                this.getDipUri(order.getResId(), orgUnitId, access, dipId));
        if (dip != null) {
          dipList.add(dip);
        }
      }
    }
    return dipList;
  }

  @Override
  public String generateDipPackages(Order order) {
    // Sort AIP by orgUnit
    final Map<String, List<ArchivalInfoPackage>> aipsByOrgUnit = this.getAipByOrgUnit(order);

    for (final Entry<String, List<ArchivalInfoPackage>> entryByOrgUnit : aipsByOrgUnit.entrySet()) {
      final String orgUnitId = entryByOrgUnit.getKey();
      final List<ArchivalInfoPackage> aipList = entryByOrgUnit.getValue();
      // Sort AIP by access
      final Map<Access, List<ArchivalInfoPackage>> aipsByAccess = aipList.stream().collect(Collectors.groupingBy(a -> a.getInfo().getAccess()));
      for (final Entry<Access, List<ArchivalInfoPackage>> entryByAccess : aipsByAccess.entrySet()) {
        final Access access = entryByAccess.getKey();
        final List<ArchivalInfoPackage> aip = entryByAccess.getValue();
        if (!this.buildDip(this.getDipFolder(order.getResId(), orgUnitId, access).resolve(DLCMConstants.DATA_FOLDER), aip)) {
          return null;
        }
      }
    }
    return order.getResId();
  }

  @Override
  public OrderStatus getStatus(String orderId, URI infoPackage) {
    try {
      this.checkOrder(orderId);
    } catch (final SolidifyCheckingException e) {
      return OrderStatus.IN_PROGRESS;
    }
    return OrderStatus.READY;
  }

  @Override
  public String submitOrder(Order order) throws JAXBException, IOException {
    // List Organizational Unit
    for (final Path orgUnitPath : FileTool.scanFolder(this.getOrderFolder(order.getResId()), Files::isDirectory)) {
      String orgUnitId = orgUnitPath.getFileName().toString();
      // List access level
      for (final Path accessPath : FileTool.scanFolder(orgUnitPath, Files::isDirectory)) {
        Access accessLevel = Access.valueOf(accessPath.getFileName().toString());
        // Generate DIP iD
        String dipId = StringTool.generateResId();
        // Generate dlcm.xml metadata
        this.metadataService.generateMetadata(order, dipId, orgUnitId, accessLevel, this.getDipFolder(order.getResId(), orgUnitId, accessLevel));
        // Create order archive (i.e. DIP) & Purge working files for each orgUnit
        final ZipTool zipTool = new ZipTool(this.getDipUri(order.getResId(), orgUnitId, accessLevel, dipId));
        if (!zipTool.zipFiles(accessPath) || !FileTool.deleteFolder(accessPath.resolve(DLCMConstants.DATA_FOLDER))) {
          return null;
        }
        // Delete dlcm.xml
        FileTool.deleteFile(accessPath.resolve(DLCMConstants.METADATA_FILE));
      }
    }
    return order.getResId();
  }

  // Extract all AIPs
  private boolean buildDip(Path rootFolder, List<ArchivalInfoPackage> aips) {
    for (final ArchivalInfoPackage aip : aips) {
      if (!this.extractArchive(aip.getArchiveContainer(), aip.getDownloadUriForAccess(), rootFolder.resolve(aip.getResId()))) {
        return false;
      }
      if (aip.isCollection()) {
        return this.buildDip(rootFolder.resolve(aip.getResId()), aip.getCollection());
      }
    }
    return true;
  }

  // Extract an AIU
  private boolean extractArchive(ArchiveContainer aipContainer, URI archive, Path destination) {
    final ZipTool zipAip = new ZipTool(archive);
    switch (aipContainer) {
      case BAG_IT:
        return zipAip.unzipFolder(destination, DLCMConstants.DATA_FOLDER + SolidifyConstants.URL_SEP);
      case ZIP:
        return zipAip.unzipFiles(destination);
      default:
        throw new SolidifyCheckingException(this.messageService.get("archival.aip.error.container", new Object[] { aipContainer }));
    }
  }

}
