/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - AccessController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.stream.Stream;

import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import ch.unige.solidify.business.OAIMetadataPrefixService;
import ch.unige.solidify.business.OAISetService;
import ch.unige.solidify.config.CitationProperties;
import ch.unige.solidify.controller.ModuleController;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.security.RootPermissions;
import ch.unige.solidify.security.TrustedUserPermissions;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.UrlPath;
import ch.dlcm.service.AccessRefreshService;

@RestController
@ConditionalOnProperty(prefix = "dlcm.module.access", name = "enable")
@RequestMapping(UrlPath.ACCESS)
public class AccessController extends ModuleController {

  protected DLCMProperties config;

  private final SystemProperties systemProperties;
  private final AccessRefreshService accessRefreshService;

  AccessController(DLCMProperties config, OAIMetadataPrefixService oaiMetadataPrefixService, OAISetService oaiSetService,
          CitationProperties citationProperties, AccessRefreshService accessRefreshService) {
    super(ModuleName.ACCESS);
    this.config = config;
    this.accessRefreshService = accessRefreshService;
    if (this.config.getData().isInit()) {
      try {
        // Init OAI Metadata Prefix
        oaiMetadataPrefixService.initDefaultData();
        // OAI Metadata Prefix: DataCite
        oaiMetadataPrefixService.createIfNotExists(
                DLCMConstants.OAI_DATACITE,
                true,
                DLCMConstants.DATACITE_NAME,
                DLCMConstants.DATACITE_NAME + " 4.3",
                DLCMMetadataVersion.getDefaultVersion().getRepresentationInfoSchema(),
                null,
                DLCMConstants.DATACITE_SCHEMA_4_3,
                DLCMConstants.DATACITE_NAMESPACE_4);

        // OAI Metadata Prefix: schema.org
        oaiMetadataPrefixService.createIfNotExists(
                DLCMConstants.OAI_SCHEMA_ORG,
                false,
                DLCMConstants.OAI_SCHEMA_ORG_NAME,
                DLCMConstants.OAI_SCHEMA_ORG_NAME,
                null,
                "datacite2schema.org.xsl",
                DLCMConstants.OAI_SCHEMA_ORG_SCHEMA,
                DLCMConstants.OAI_SCHEMA_ORG_SCHEMA);
        // Update OAI DC
        oaiMetadataPrefixService.updateOaiDc("datacite2oai_dc.xsl");

      } catch (IOException e) {
        throw new SolidifyRuntimeException(e.getMessage(), e);
      }
      // OAI Set for all OrgUnits
      oaiSetService.createIfNotExists(DLCMConstants.OAI_ORG_UNIT, "Organizational Units", "Set for all organizational units", "*");
    }
    this.systemProperties = new SystemProperties(citationProperties);
  }

  @RootPermissions
  @PostMapping(DLCMActionName.PURGE_CONTENT_TEMP_FOLDER)
  public HttpStatus purgeContentTempFolder() {
    try {
      Path tmpLocation = Paths.get(this.config.getTempLocation(this.config.getAccessLocation()));
      try (Stream<Path> walk = Files.walk(tmpLocation)) {
        walk.sorted(Comparator.reverseOrder())
                .map(Path::toFile)
                .filter(item -> !item.getPath().equals(tmpLocation.toString()))
                .forEach(File::delete);
      }
    } catch (IOException e) {
      return HttpStatus.INTERNAL_SERVER_ERROR;
    }
    return HttpStatus.OK;
  }

  @TrustedUserPermissions
  @PostMapping("/" + DLCMActionName.REFRESH + "/{aipId}")
  public HttpStatus refresh(@PathVariable String aipId) {
    this.accessRefreshService.refreshAip(aipId);
    return HttpStatus.OK;
  }

  @GetMapping(DLCMActionName.SYSTEM_PROPERTIES)
  public HttpEntity<SystemProperties> systemProperties() {
    this.systemProperties.removeLinks();
    this.systemProperties.add(linkTo(this.getClass()).withSelfRel());
    return new ResponseEntity<>(this.systemProperties, HttpStatus.OK);
  }

  private static class SystemProperties extends RepresentationModel<SystemProperties> {
    Map<String, String[]> citations = new HashMap();
    Map<String, String[]> bibliographies = new HashMap();

    public SystemProperties(CitationProperties citationProperties) {
      this.citations.put("languages", citationProperties.getCitations().getLanguages());
      this.citations.put("outputFormats", citationProperties.getCitations().getOutputFormats());
      this.citations.put("styles", citationProperties.getCitations().getStyles());

      this.bibliographies.put("languages", citationProperties.getBibliographies().getLanguages());
      this.bibliographies.put("outputFormats", citationProperties.getBibliographies().getOutputFormats());
      this.bibliographies.put("styles", citationProperties.getBibliographies().getStyles());
    }

    public Map<String, String[]> getCitations() {
      return this.citations;
    }

    public Map<String, String[]> getBibliographies() {
      return this.bibliographies;
    }
  }
}
