/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - ArchiveMetadataDataProvider.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.citation;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import de.undercouch.citeproc.ItemDataProvider;
import de.undercouch.citeproc.csl.CSLItemData;
import de.undercouch.citeproc.csl.CSLItemDataBuilder;
import de.undercouch.citeproc.csl.CSLName;
import de.undercouch.citeproc.csl.CSLNameBuilder;
import de.undercouch.citeproc.csl.CSLType;

import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.service.SearchService;

public class ArchiveMetadataDataProvider implements ItemDataProvider {

  private SearchService searchService;

  public ArchiveMetadataDataProvider(SearchService searchService) {
    this.searchService = searchService;
  }

  /**************************************************************************************/

  @Override
  public CSLItemData retrieveItem(String archiveId) {

    final ArchiveMetadata archiveMetadata = this.searchService.findOne(MetadataVisibility.PUBLIC, archiveId);

    CSLType cslType = this.getCSLTypeFromDataciteType(archiveMetadata.getResourceTypeGeneral());
    CSLName[] authors = this.getAuthors(archiveMetadata.getCreators());

    CSLItemDataBuilder builder = new CSLItemDataBuilder().id(archiveId);
    builder.type(cslType)
            .title(archiveMetadata.getTitle())
            .author(authors)
            .DOI(archiveMetadata.getDoi())
            .publisher(archiveMetadata.getPublisher())
            .abstrct(archiveMetadata.getAbstract());

    if (archiveMetadata.getIssuedDate() != null) {
      OffsetDateTime issuedDate = archiveMetadata.getIssuedDate();
      builder.issued(issuedDate.getYear(), issuedDate.getMonthValue(), issuedDate.getDayOfMonth());
    }

    if (archiveMetadata.getAcceptedDate() != null) {
      OffsetDateTime submittedDate = archiveMetadata.getAcceptedDate();
      builder.submitted(submittedDate.getYear(), submittedDate.getMonthValue(), submittedDate.getDayOfMonth());
    }

    return builder.build();
  }

  @Override
  public Collection<String> getIds() {
    // return an empty array as this method seems to be never called by the library
    return new ArrayList<>();
  }

  /**************************************************************************************/

  private CSLName[] getAuthors(List<Map<String, String>> creators) {
    List<CSLName> cslNames = new ArrayList<>();
    for (Map<String, String> creator : creators) {
      CSLName cslName = new CSLNameBuilder().given(creator.get(DLCMConstants.GIVEN_NAME)).family(creator.get(DLCMConstants.FAMILY_NAME)).build();
      cslNames.add(cslName);
    }
    return cslNames.toArray(new CSLName[0]);
  }

  private CSLType getCSLTypeFromDataciteType(String resourceType) {
    if (!StringTool.isNullOrEmpty(resourceType)) {
      switch (resourceType) {
        case DLCMConstants.DATACITE_TYPE_BOOK:
          return CSLType.BOOK;
        case DLCMConstants.DATACITE_TYPE_BOOK_CHAPTER:
          return CSLType.CHAPTER;
        case DLCMConstants.DATACITE_TYPE_CONFERENCE_PAPER:
          return CSLType.PAPER_CONFERENCE;
        case DLCMConstants.DATACITE_TYPE_DATASET:
          return CSLType.DATASET;
        case DLCMConstants.DATACITE_TYPE_JOURNAL_ARTICLE:
          return CSLType.ARTICLE_JOURNAL;
        case DLCMConstants.DATACITE_TYPE_PEER_REVIEW:
          return CSLType.REVIEW;
        case DLCMConstants.DATACITE_TYPE_REPORT:
          return CSLType.REPORT;
        default:
          return null;
      }
    }
    return null;
  }
}
