/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - OaiPmhTest.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.access;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.JAXBException;

import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.springframework.data.domain.Pageable;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import ch.unige.solidify.OAIConstants;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.service.OAIService;
import ch.unige.solidify.test.oai.OAIServiceTest;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.config.DLCMRepositoryDescription;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.service.HistoryService;
import ch.dlcm.service.IdentifierService;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.SearchMgmt;
import ch.dlcm.service.rest.fallback.FallbackArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveTypeRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackLanguageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackLicenseRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackMetadataTypeRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackPersonRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackPreservationPolicyRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackSubmissionInfoPackageRemoteResourceService;

@ActiveProfiles({ "sec-noauth" })
@ExtendWith(SpringExtension.class)
public class OaiPmhTest extends OAIServiceTest {

  protected DLCMProperties dlcmConfig;
  protected DLCMRepositoryDescription repositoryDescription;
  protected MetadataService metadataService;
  protected SearchMgmt searchMgmt;
  protected IdentifierService identifierService;

  @Mock
  protected HistoryService historyService;
  @Mock
  protected FallbackLanguageRemoteResourceService languageResourceService;
  @Mock
  protected FallbackOrganizationalUnitRemoteResourceService orgUnitResourceService;
  @Mock
  protected FallbackArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  @Mock
  protected FallbackMetadataTypeRemoteResourceService metadataTypeResourceService;
  @Mock
  protected FallbackPersonRemoteResourceService personRemoteResourceService;
  @Mock
  protected FallbackLicenseRemoteResourceService licenseRemoteResourceService;
  @Mock
  protected FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService;
  @Mock
  protected FallbackSubmissionInfoPackageRemoteResourceService fallbackSubmissionInfoPackageRemoteResourceService;
  @Mock
  protected FallbackPreservationPolicyRemoteResourceService preservationPolicyRemoteResourceService;
  @Mock
  protected FallbackArchiveTypeRemoteResourceService fallbackArchiveTypeResourceService;

  @Override
  protected void localSetup() {
    this.dlcmConfig = new DLCMProperties(new SolidifyProperties());
    this.repositoryDescription = new DLCMRepositoryDescription(this.gitInfoProperties);
    this.repositoryDescription.setArchiveHomePage(HOST_URL + SolidifyConstants.URL_SEP);
    // Metadata Service
    this.metadataService = new MetadataService(
            this.languageResourceService,
            this.orgUnitResourceService,
            this.aipRemoteResourceService,
            this.historyService,
            this.dlcmConfig,
            this.repositoryDescription,
            this.messageService,
            this.gitInfoProperties,
            this.metadataTypeResourceService,
            this.personRemoteResourceService,
            this.licenseRemoteResourceService,
            this.archiveMetadataRemoteResourceService,
            this.preservationPolicyRemoteResourceService,
            this.fallbackSubmissionInfoPackageRemoteResourceService,
            this.fallbackArchiveTypeResourceService);
    // Search Service
    this.searchMgmt = new SearchMgmt(
            this.messageService,
            this.aipRemoteResourceService,
            this.archiveMetadataRemoteResourceService);
    // Git Info Service
    this.gitInfoProperties.getBuild().setVersion("1.x");
    this.identifierService = new IdentifierService(this.repositoryDescription);
    // OAI Service
    this.oaiService = new OAIService(
            this.oaiConfig,
            this.repositoryDescription,
            this.oaiMetadataPrefixService,
            this.oaiSetService,
            this.searchMgmt,
            this.archiveMetadataRemoteResourceService,
            this.metadataService);

    // Mock DLCMRestClientService
    try {
      for (DLCMMetadataVersion version : DLCMMetadataVersion.values()) {
        when(this.archiveMetadataRemoteResourceService.getMetadata(MetadataVisibility.PUBLIC, version.getVersion()))
                .thenReturn(this.getMetadata(version));
      }
      when(this.archiveMetadataRemoteResourceService.getMetadataList(any(MetadataVisibility.class), any(String.class), any(String.class),
              any(Pageable.class)))
                      .thenReturn(this.getArchiveMetadataList());
      when(this.archiveMetadataRemoteResourceService.getMetadataList(any(MetadataVisibility.class), any(), any(String.class), any(), any(),
              any(),
              any(), any(Pageable.class)))
                      .thenReturn(this.getArchiveMetadataList());
    } catch (IOException | JAXBException e) {
      throw new SolidifyRuntimeException(e.getMessage(), e);
    }
  }

  @Override
  protected String[] getMetadataList() {
    return new String[] { OAIConstants.OAI_DC, DLCMConstants.OAI_DATACITE, DLCMConstants.OAI_SCHEMA_ORG };
  }

  private ArchiveMetadata getMetadata(DLCMMetadataVersion version) throws IOException, JAXBException {
    ArchiveMetadata archiveMetadata = new ArchiveMetadata();
    archiveMetadata.setIndex(this.dlcmConfig.getIndexing().getIndexName());
    archiveMetadata.setResId(version.getVersion());
    Path xml = this.getPath("dlcm_mets-" + version.getVersion() + SolidifyConstants.XML_EXT);
    this.metadataService.completeArchiveIndex(version, archiveMetadata, xml);
    return archiveMetadata;
  }

  private RestCollection<ArchiveMetadata> getArchiveMetadataList() throws IOException, JAXBException {
    List<ArchiveMetadata> list = new ArrayList<>();
    for (DLCMMetadataVersion version : DLCMMetadataVersion.values()) {
      ArchiveMetadata archiveMetadata = this.getMetadata(version);
      list.add(archiveMetadata);
    }
    return new RestCollection<>(list);
  }

  @Override
  protected List<String> getIdentifierList() {
    List<String> list = new ArrayList<>();
    for (DLCMMetadataVersion version : DLCMMetadataVersion.values()) {
      list.add(version.getVersion());
    }
    return list;
  }

}
