/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - UserService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.business;

import static ch.dlcm.DLCMRestFields.APPLICATION_ROLE_FIELD;
import static ch.dlcm.DLCMRestFields.RES_ID_FIELD;

import java.security.Principal;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Primary;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.auth.client.service.AuthorizationClientProperties;
import ch.unige.solidify.auth.client.service.UserCreatorService;
import ch.unige.solidify.auth.model.ApplicationRole;
import ch.unige.solidify.auth.model.AuthApplicationRole;
import ch.unige.solidify.auth.model.AuthOrcidInfo;
import ch.unige.solidify.auth.model.AuthUser;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.model.SolidifyApplicationRole;
import ch.unige.solidify.rest.ResourceBase;
import ch.unige.solidify.service.HttpRequestInfoProvider;
import ch.unige.solidify.service.ResourceService;
import ch.unige.solidify.service.UserInfoService;
import ch.unige.solidify.util.PropagateRestClientTool;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.TrustedRestClientTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.AdminController;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.Person;
import ch.dlcm.repository.ApplicationRoleRepository;
import ch.dlcm.repository.InstitutionRepository;
import ch.dlcm.repository.PersonRepository;
import ch.dlcm.repository.UserRepository;
import ch.dlcm.specification.UserSpecification;

@Service
@ConditionalOnBean(AdminController.class)
@Primary
public class UserService extends ResourceService<User> implements UserInfoService, UserCreatorService {
  Logger log = LoggerFactory.getLogger(UserService.class);

  private static final String USERS_ENDPOINT = "/users/";
  private static final String DELETE_TOKEN_ENDPOINT = "/delete-token";
  private static final String CHANGE_ROLE_ENDPOINT = "/change-role";
  private static final String CHANGE_ORCID_ENDPOINT = "/change-orcid";

  private final String authModuleUrl;
  private final UserRepository userRepository;
  private final PersonRepository personRepository;
  private final ApplicationRoleRepository applicationRoleRepository;
  private final InstitutionRepository institutionRepository;
  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final TrustedRestClientTool trustedRestClientTool;
  private final PropagateRestClientTool propagateRestClientTool;

  public UserService(
          UserRepository userRepository,
          PersonRepository personRepository,
          ApplicationRoleRepository applicationRoleRepository,
          InstitutionRepository institutionRepository,
          HttpRequestInfoProvider httpRequestInfoProvider,
          AuthorizationClientProperties authProperties,
          TrustedRestClientTool trustedRestClientTool,
          PropagateRestClientTool propagateRestClientTool) {
    this.userRepository = userRepository;
    this.personRepository = personRepository;
    this.applicationRoleRepository = applicationRoleRepository;
    this.institutionRepository = institutionRepository;
    this.authModuleUrl = authProperties.getAuthorizationServerUrl();
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.trustedRestClientTool = trustedRestClientTool;
    this.propagateRestClientTool = propagateRestClientTool;
  }

  @Override
  public User findByExternalUid(String uniqueId) {
    return ((UserRepository) this.itemRepository).findByExternalUid(uniqueId);
  }

  public User findByPersonResId(String personId) {
    return this.userRepository.findByPersonResId(personId);
  }

  /**
   * Method to get a list of user that have a role application level lower than the one specified by
   * the parameter.
   * 
   * @param applicationRole minimum application role from which you want to get all the users
   * @return list of users
   */
  public List<User> findByApplicationRoleLevelHierarchy(ApplicationRole applicationRole) {
    final UserSpecification spec = this.getSpecification(new User());
    spec.setMinimumApplicationRole(applicationRole);

    return this.findAll(spec, Pageable.unpaged()).toList();
  }

  public void revokeAccessAndRefreshTokens(User user) {
    RestTemplate restTemplate = this.trustedRestClientTool.getClient();
    restTemplate.delete(this.authModuleUrl + USERS_ENDPOINT + user.getExternalUid() + DELETE_TOKEN_ENDPOINT);
  }

  @Override
  public User save(User user) {
    if (user.getApplicationRole() == null) {
      user.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.USER));
    }
    return super.save(user);
  }

  @Override
  public UserSpecification getSpecification(User resource) {
    return new UserSpecification(resource);
  }

  @Override
  @Transactional
  public synchronized String createNewUser(String externalUid) {
    if (!this.userRepository.existsByExternalUid(externalUid)) {
      final AuthUser authUser = this.getAuthenticatedUser(externalUid);
      // Create a new user
      User newUser = new User(authUser);
      // Update Last Login
      this.updateLastLogin(newUser, authUser);
      // Create Person
      Person newPerson = new Person();
      newPerson.setFirstName(newUser.getFirstName());
      newPerson.setLastName(newUser.getLastName());
      newPerson.setOrcid(authUser.getOrcid());
      newPerson.setVerifiedOrcid(authUser.isVerifiedOrcid());
      // Set institution
      Optional<Institution> institution = this.institutionRepository.findByEmailSuffix(authUser.getHomeOrganization());
      if (institution.isPresent()) {
        newPerson.addInstitution(institution.get());
      }
      newPerson = this.personRepository.save(newPerson);
      // Update user
      newUser.setPerson(newPerson);
      // Save user
      this.userRepository.save(newUser);
    }
    return externalUid;
  }

  public User getCurrentUser() {
    final Principal principal = this.httpRequestInfoProvider.getPrincipal();
    if (principal != null) {
      return this.findByExternalUid(principal.getName());
    }
    return null;
  }

  public void updateUser(User user) {
    final AuthUser authUser = this.getAuthenticatedUser(user.getExternalUid());
    this.updateLastLogin(user, authUser);
    this.updateApplicationRole(user, authUser);
  }

  public AuthUser updateUserRole(String externalUid, String role) {
    RestTemplate restTemplate = this.propagateRestClientTool.getClient();
    AuthUser authUser = restTemplate.postForObject(this.authModuleUrl + USERS_ENDPOINT + externalUid + CHANGE_ROLE_ENDPOINT, role,
            AuthUser.class);
    if (authUser == null) {
      throw new IllegalStateException();
    }
    return authUser;
  }

  public AuthUser updateUserOrcid(String externalUid, AuthOrcidInfo orcidInfo) {
    RestTemplate restTemplate = this.trustedRestClientTool.getClient();
    AuthUser authUser = restTemplate.postForObject(this.authModuleUrl + USERS_ENDPOINT + externalUid + CHANGE_ORCID_ENDPOINT, orcidInfo,
            AuthUser.class);
    if (authUser == null) {
      throw new IllegalStateException();
    }
    return authUser;
  }

  @Override
  // Override method to avoid auto-REST call when accessing to users locally
  protected ResourceBase afterFind(User user) {
    // Update creator name
    if (!StringTool.isNullOrEmpty(user.getCreatedBy())) {
      User creator = this.findByExternalUid(user.getCreatedBy());
      if (creator != null) {
        user.setCreatorName(creator.getFullName());
      }
    }
    // Update editor name
    if (!StringTool.isNullOrEmpty(user.getUpdatedBy())) {
      User editor = this.findByExternalUid(user.getUpdatedBy());
      if (editor != null) {
        user.setEditorName(editor.getFullName());
      }
    }

    this.hideSensitiveData(user);

    return user;
  }

  private void hideSensitiveData(User user) {
    User currentUser = this.getCurrentUser();
    if (currentUser != null
            && !user.getResId().equals(currentUser.getResId())
            && currentUser.getApplicationRole().getResId().equals(AuthApplicationRole.USER_ID)) {
      user.setLastLoginIpAddress(null);
      user.setLastLoginTime(null);
      user.setApplicationRole(null);
      user.setExternalUid(null);
    }
  }

  public AuthUser getAuthenticatedUser(String externalUid) {
    RestTemplate restTemplate = this.trustedRestClientTool.getClient();
    AuthUser authUser = restTemplate.getForEntity(this.authModuleUrl + USERS_ENDPOINT + externalUid, AuthUser.class).getBody();
    if (authUser == null) {
      throw new IllegalStateException();
    }
    return authUser;
  }

  private void updateLastLogin(User user, AuthUser authUser) {
    if (authUser.getLastLogin() != null) {
      user.setLastLoginIpAddress(authUser.getLastLogin().getLoginIpAddress());
      user.setLastLoginTime(authUser.getLastLogin().getLoginTime());
    } else if (authUser.getFirstLogin() != null) {
      user.setLastLoginIpAddress(authUser.getFirstLogin().getLoginIpAddress());
      user.setLastLoginTime(authUser.getFirstLogin().getLoginTime());
    }
  }

  private void updateApplicationRole(User user, AuthUser authUser) {
    if (!this.haveSameRole(user, authUser)) {
      Optional<SolidifyApplicationRole> role = this.getApplicationRole(authUser.getApplicationRole().getResId());
      if (role.isPresent()) {
        user.setApplicationRole(role.get());
      } else {
        throw new SolidifyRuntimeException("Cannot find application role (" + authUser.getApplicationRole().getResId() + ")");
      }
    }
  }

  public void createDocumentationUser(DLCMProperties.Documentation.User user, Person person) {
    String name = user.getName().toUpperCase();
    String mail = name + "@unige.ch";
    User userDb = ((UserRepository) this.itemRepository).findByEmail(mail);
    if (userDb == null) {
      User newUser = new User();
      newUser.setResId(user.getName().toLowerCase());
      newUser.setFirstName(DLCMConstants.PERMANENT_TEST_DATA_LABEL + name + "_firstName");
      newUser.setLastName(DLCMConstants.PERMANENT_TEST_DATA_LABEL + name + "_lastName");
      newUser.setEmail(mail);
      newUser.setHomeOrganization(name + "_homeOrganization");
      newUser.setExternalUid(name + "_externalUid");
      newUser.setPerson(person);
      if (user.getRole().equalsIgnoreCase(AuthApplicationRole.ADMIN.getName())) {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.ADMIN));
      } else if (user.getRole().equalsIgnoreCase(AuthApplicationRole.ROOT.getName())) {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.ROOT));
      } else if (user.getRole().equalsIgnoreCase(AuthApplicationRole.TRUSTED_CLIENT.getName())) {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.TRUSTED_CLIENT));
      } else {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.USER));
      }
      User createdUser = this.save(newUser);
      this.log.info("Test user '{}' created", createdUser.getResId());
    }
  }

  public boolean containsApplicationRoleId(Map<String, Object> updateMap) {
    if (!updateMap.containsKey(APPLICATION_ROLE_FIELD) ||
            !(updateMap.get(APPLICATION_ROLE_FIELD) instanceof Map)) {
      return false;
    }
    final Map<String, Object> applicationRoleMap = (Map<String, Object>) updateMap.get(APPLICATION_ROLE_FIELD);
    return applicationRoleMap.containsKey(RES_ID_FIELD)
            && ((applicationRoleMap.get(RES_ID_FIELD)) instanceof String || applicationRoleMap.get(RES_ID_FIELD) == null);
  }

  public String getApplicationRoleId(Map<String, Object> updateMap) {
    if (!this.containsApplicationRoleId(updateMap)) {
      return null;
    }
    return (String) ((Map<String, Object>) updateMap.get(APPLICATION_ROLE_FIELD)).get(RES_ID_FIELD);
  }

  public String synchronizeOrcid(String externalUid, Person person) {
    AuthUser user = this.getAuthenticatedUser(externalUid);
    return this.synchronizeOrcid(user, person);
  }

  public User synchronizeUserInfo(User user, AuthUser authUser) {
    // Update login info & role
    this.updateLastLogin(user, authUser);
    this.updateApplicationRole(user, authUser);
    return this.save(user);
  }

  public String synchronizeOrcid(AuthUser user, Person person) {
    if (StringTool.isNullOrEmpty(user.getOrcid()) && StringTool.isNullOrEmpty(person.getOrcid())) {
      // No ORCID => nothing to synchronize
      return null;
    }

    if (StringTool.isNullOrEmpty(user.getOrcid())) {
      // Update user in Authorization
      AuthOrcidInfo orcidInfo = new AuthOrcidInfo();
      orcidInfo.setOrcid(person.getOrcid());
      orcidInfo.setVerifiedOrcid(person.isVerifiedOrcid());
      this.updateUserOrcid(user.getExternalUid(), orcidInfo);
      log.info("User {}: ORCID updated with {}", user.getExternalUid(), person.getOrcid());
      return person.getFullName() + "[" + user.getExternalUid() + "]";
    } else if (StringTool.isNullOrEmpty(person.getOrcid())) {
      // Update person
      person.setOrcid(user.getOrcid());
      person.setVerifiedOrcid(user.isVerifiedOrcid());
      this.personRepository.save(person);
      log.info("Person {}: ORCID updated with {}", person.getFullName(), user.getOrcid());
      return person.getFullName() + "[" + user.getExternalUid() + "]";
    } else {
      if (!user.getOrcid().equals(person.getOrcid())) {
        throw new SolidifyRuntimeException("Insconsistent ORCID (" + user.getOrcid() + " vs " + person.getOrcid() + ") for "
                + person.getFullName() + " [" + user.getExternalUid() + "]");
      }
    }
    return null;
  }

  private Optional<SolidifyApplicationRole> getApplicationRole(String resId) {
    return this.applicationRoleRepository.findById(resId);
  }

  private boolean haveSameRole(User user, AuthUser authUser) {
    return user.getApplicationRole().getResId().equals(authUser.getApplicationRole().getResId());
  }
}
