/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - OrganizationalUnitController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller.admin;

import java.io.IOException;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.authentication.AnonymousAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.security.AdminPermissions;
import ch.unige.solidify.security.NoOnePermissions;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.business.PersonService;
import ch.dlcm.business.RoleService;
import ch.dlcm.controller.AdminController;
import ch.dlcm.controller.DLCMResourceWithLogoController;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.UrlPath;
import ch.dlcm.service.rest.propagate.PropagatePersonRemoteResourceService;

@AdminPermissions
@RestController
@ConditionalOnBean(AdminController.class)
@RequestMapping(UrlPath.ADMIN_UNIT)
public class OrganizationalUnitController extends DLCMResourceWithLogoController<OrganizationalUnit> {

  private static final Logger log = LoggerFactory.getLogger(OrganizationalUnitController.class);

  protected final PersonService personService;

  private final PropagatePersonRemoteResourceService personRemoteResourceService;

  protected final RoleService roleService;

  public OrganizationalUnitController(PersonService personService,
          PropagatePersonRemoteResourceService personRemoteResourceService, RoleService roleService) {
    this.personService = personService;
    this.personRemoteResourceService = personRemoteResourceService;
    this.roleService = roleService;
  }

  @Override
  public HttpEntity<OrganizationalUnit> create(@RequestBody OrganizationalUnit organizationalUnit) {

    final ResponseEntity<OrganizationalUnit> httpResponse = (ResponseEntity<OrganizationalUnit>) super.create(organizationalUnit);
    organizationalUnit = httpResponse.getBody();

    /**
     * Set the authenticated user as a manager on the new organizational unit
     */
    if (httpResponse.getStatusCode() == HttpStatus.CREATED && organizationalUnit != null) {

      final Authentication authentication = SecurityContextHolder.getContext().getAuthentication();

      if (authentication != null && !(authentication instanceof AnonymousAuthenticationToken)) {

        final String personId = this.personRemoteResourceService.getLinkedPersonId(authentication);
        if (personId != null) {

          final Person person = this.personService.findOne(personId);
          final Role role = this.roleService.findOne(Role.MANAGER_ID);

          organizationalUnit.addPersonRole(person, role);
          organizationalUnit = this.itemService.save(organizationalUnit);

          this.addLinks(organizationalUnit);
          return new ResponseEntity<>(organizationalUnit, HttpStatus.CREATED);
        } else {
          log.error("The authentication object '{}' does not correspond to any person", authentication.getName());
        }

        return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
      }
    }

    return httpResponse;
  }

  @UserPermissions
  @Override
  public HttpEntity<OrganizationalUnit> get(@PathVariable String id) {
    return super.get(id);
  }

  @UserPermissions
  @Override
  public HttpEntity<RestCollection<OrganizationalUnit>> list(@ModelAttribute OrganizationalUnit search, Pageable pageable) {
    return super.list(search, pageable);
  }

  @UserPermissions
  @Override
  public HttpEntity<RestCollection<OrganizationalUnit>> advancedSearch(@ModelAttribute OrganizationalUnit resource,
          @RequestParam("search") String search, @RequestParam(value = "match", required = false) String matchtype, Pageable pageable) {
    return super.advancedSearch(resource, search, matchtype, pageable);
  }

  @UserPermissions
  @Override
  public HttpEntity<RestCollection<OrganizationalUnit>> advancedSearch(@RequestBody OrganizationalUnit search,
          @RequestParam(value = "match", required = false) String matchtype, Pageable pageable) {
    return super.advancedSearch(search, matchtype, pageable);
  }

  @PreAuthorize("@organizationalUnitPermissionService.isAllowedToUpdate(#id)")
  @Override
  public HttpEntity<OrganizationalUnit> update(@PathVariable String id, @RequestBody Map<String, Object> updateMap) {
    return super.update(id, updateMap);
  }

  @PreAuthorize("@organizationalUnitPermissionService.isAllowed(#id, 'DELETE')")
  @Override
  public HttpStatus delete(@PathVariable String id) {
    if (this.itemService.findOne(id).getIsEmpty()) {
      return super.delete(id);
    }
    return this.closeOrgUnit(id, null);
  }

  @NoOnePermissions
  @Override
  public HttpStatus deleteList(@RequestBody String[] ids) {
    return HttpStatus.FORBIDDEN;
  }

  @PreAuthorize("@organizationalUnitPermissionService.isAllowed(#id, 'DELETE')")
  @PostMapping(SolidifyConstants.URL_ID_PLUS_SEP + ActionName.CLOSE)
  public HttpStatus close(@PathVariable String id, @RequestParam(value = "closingDate", required = false) String closingDate) {
    return this.closeOrgUnit(id, closingDate);
  }

  private HttpStatus closeOrgUnit(String orgUnitId, String closingDate) {
    final DateTimeFormatter formatter = DateTimeFormatter.ofPattern(StringTool.DATE_FORMAT);
    final OrganizationalUnit orgUnit = this.itemService.findOne(orgUnitId);
    if (orgUnit.isOpen()) {
      if (closingDate != null) {
        orgUnit.setClosingDate(LocalDate.parse(closingDate, formatter));
      } else {
        orgUnit.setClosingDate(LocalDate.now());
      }
      this.itemService.save(orgUnit);
    }
    return HttpStatus.OK;
  }

  @PreAuthorize("@organizationalUnitPermissionService.isAllowed(#id, 'UPDATE')")
  @PostMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.UPLOAD_LOGO)
  @Override
  public HttpEntity<OrganizationalUnit> uploadLogo(@PathVariable String id, @RequestParam(DLCMConstants.FILE) MultipartFile file) throws IOException {
    return super.uploadLogo(id, file);
  }

  @UserPermissions
  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.DOWNLOAD_LOGO)
  @ResponseBody
  @Override
  public HttpEntity<StreamingResponseBody> downloadLogo(@PathVariable String id) throws IOException {
    return super.downloadLogo(id);
  }

  @PreAuthorize("@organizationalUnitPermissionService.isAllowed(#id, 'UPDATE')")
  @DeleteMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.DELETE_LOGO)
  @ResponseBody
  @Override
  public HttpEntity<OrganizationalUnit> deleteLogo(@PathVariable String id) {
    return super.deleteLogo(id);
  }
}
