/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - UserController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller.admin;

import java.text.ParseException;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestHeader;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import com.nimbusds.jwt.SignedJWT;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.controller.ResourceController;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.security.AdminPermissions;
import ch.unige.solidify.security.RootPermissions;
import ch.unige.solidify.security.TrustedUserPermissions;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.business.UserService;
import ch.dlcm.controller.AdminController;
import ch.dlcm.model.security.User;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.UrlPath;

@RestController
@ConditionalOnBean(AdminController.class)
@RequestMapping(UrlPath.ADMIN_USER)
public class UserController extends ResourceController<User> {
  static final Logger log = LoggerFactory.getLogger(UserController.class);

  @UserPermissions
  @Override
  public HttpEntity<User> get(@PathVariable String id) {
    return super.get(id);
  }

  @UserPermissions
  @Override
  public HttpEntity<RestCollection<User>> list(@ModelAttribute User search, Pageable pageable) {
    return super.list(search, pageable);
  }

  @UserPermissions
  @Override
  public HttpEntity<RestCollection<User>> advancedSearch(@ModelAttribute User user,
          @RequestParam("search") String search,
          @RequestParam(value = "match", required = false) String matchType, Pageable pageable) {
    return super.advancedSearch(user, search, matchType, pageable);
  }

  @UserPermissions
  @Override
  public HttpEntity<RestCollection<User>> advancedSearch(@RequestBody User search,
          @RequestParam(value = "match", required = false) String matchtype, Pageable pageable) {
    return super.advancedSearch(search, matchtype, pageable);
  }

  @PreAuthorize("@userPermissionService.isAllowedToUpdate(#id, #updateMap)")
  @Override
  public HttpEntity<User> update(@PathVariable String id, @RequestBody Map<String, Object> updateMap) {
    // Get existing user
    final User existingUser = this.itemService.findOne(id);
    if (updateMap == null) {
      return ResponseEntity.notFound().build();
    }
    // Check if role changed
    String updateApplicationRoleId = ((UserService) this.itemService).getApplicationRoleId(updateMap);
    if (!existingUser.getApplicationRole().getResId().equals(updateApplicationRoleId)) {
      ((UserService) this.itemService).updateUserRole(existingUser.getExternalUid(), updateApplicationRoleId);
    }
    return super.update(id, updateMap);
  }

  @AdminPermissions
  @Override
  public HttpStatus delete(@PathVariable String id) {
    return super.delete(id);
  }

  @AdminPermissions
  @Override
  public HttpStatus deleteList(@RequestBody String[] ids) {
    return super.deleteList(ids);
  }

  @UserPermissions
  @GetMapping(SolidifyConstants.URL_SEP + DLCMActionName.AUTHENTICATED)
  public HttpEntity<User> getAuthenticatedUser() {
    final String userId = this.getAuthenticatedUserId();
    final User user = ((UserService) this.itemService).findByExternalUid(userId);
    this.addLinks(user);
    return new ResponseEntity<>(user, HttpStatus.OK);
  }

  @AdminPermissions
  @PostMapping(SolidifyConstants.URL_SEP + DLCMActionName.REVOKE_ALL_TOKENS + SolidifyConstants.URL_SEP + UrlPath.AUTH_USER_ID)
  public ResponseEntity<Object> revokeAllToken(@PathVariable String externalUid) {

    final User user = ((UserService) this.itemService).findByExternalUid(externalUid);
    if (user == null) {
      return ResponseEntity.notFound().build();
    }
    ((UserService) this.itemService).revokeAccessAndRefreshTokens(user);
    return ResponseEntity.ok().build();
  }

  @UserPermissions
  @PostMapping(SolidifyConstants.URL_SEP + DLCMActionName.REVOKE_MY_TOKENS)
  public ResponseEntity<Object> revokeMyToken(@RequestHeader("Authorization") String authorizationHeader) throws ParseException {
    final String token = authorizationHeader.substring(authorizationHeader.indexOf(' ') + 1);
    final String externalUid = SignedJWT.parse(token).getJWTClaimsSet().getClaim("user_name").toString();
    final User user = ((UserService) this.itemService).findByExternalUid(externalUid);
    ((UserService) this.itemService).revokeAccessAndRefreshTokens(user);
    return ResponseEntity.ok().build();
  }

  @RootPermissions
  @PostMapping(DLCMActionName.REFRESH)
  public HttpEntity<Result> refresh() {
    Result result = new Result("users");
    try {
      Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
      Page<User> userList;
      do {
        userList = this.itemService.findAll(pageable);
        for (User user : userList) {
          ((UserService) this.itemService).updateUser(user);
          this.itemService.save(user);
        }
        pageable = pageable.next();
      } while (!userList.isEmpty());
      result.setStatus(Result.ActionStatus.EXECUTED);
      result.setMesssage(this.messageService.get("message.user.updated"));
    } catch (RuntimeException e) {
      log.error("In updating user information", e);
      result.setStatus(Result.ActionStatus.NOT_EXECUTED);
      result.setMesssage(e.getMessage());
    }
    return this.returnHttpResult(result);
  }

  @RootPermissions
  @PostMapping(DLCMActionName.SYNCHRONIZE)
  public HttpEntity<Result> synchronize() {
    Result result = new Result("synchronize");
    try {
      StringBuilder sb = new StringBuilder(this.messageService.get("message.user.sync"));
      Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
      Page<User> userList;
      do {
        userList = this.itemService.findAll(pageable);
        for (User user : userList) {
          final String who = ((UserService) this.itemService).synchronizeOrcid(user.getExternalUid(), user.getPerson());
          if (!StringTool.isNullOrEmpty(who)) {
            sb.append(" ").append(who).append(" ; ");
          }
        }
        pageable = pageable.next();
      } while (!userList.isEmpty());
      result.setStatus(Result.ActionStatus.EXECUTED);
      result.setMesssage(sb.toString());
    } catch (RuntimeException e) {
      log.error("In synchronising user information", e);
      result.setStatus(Result.ActionStatus.NOT_EXECUTED);
      result.setMesssage(e.getMessage());
    }
    return this.returnHttpResult(result);
  }

  @TrustedUserPermissions
  @PostMapping(SolidifyConstants.URL_SEP + DLCMActionName.NEW_USER + SolidifyConstants.URL_EXTERNAL_UID)
  public HttpStatus createNewUser(@PathVariable String externalUid) {
    ((UserService) this.itemService).createNewUser(externalUid);
    return HttpStatus.OK;
  }
}
