/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - UserSynchronizationService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.auth.model.AuthUser;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.business.UserService;
import ch.dlcm.controller.AdminController;
import ch.dlcm.model.security.User;

@Service
@ConditionalOnBean(AdminController.class)
public class UserSynchronizationService extends DLCMService {
  private static final Logger log = LoggerFactory.getLogger(UserSynchronizationService.class);

  private final UserService userService;

  protected UserSynchronizationService(MessageService messageService, UserService userService) {
    super(messageService);
    this.userService = userService;
  }

  @Scheduled(fixedRate = 24 * SolidifyConstants.MILLISECONDS_IN_HOUR, initialDelay = SolidifyConstants.MILLISECONDS_IN_MINUTE)
  @Transactional
  public void synchronizeUsers() {
    Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
    Page<User> userList;
    do {
      userList = this.userService.findAll(pageable);
      for (User user : userList) {
        try {
          final AuthUser authUser = this.userService.getAuthenticatedUser(user.getExternalUid());
          // Update login info & role
          this.userService.synchronizeUserInfo(user, authUser);
          // Synchronize Orcid
          this.userService.synchronizeOrcid(authUser, user.getPerson());
          log.info("User {} synchronized", user.getFullName());
        } catch (RuntimeException e) {
          log.error("User {} not synchronized", user.getFullName());
        }
      }
      pageable = pageable.next();
    } while (!userList.isEmpty());
  }
}
