/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - AclPermissionService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.security;

import java.security.Principal;
import java.util.Optional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.service.HttpRequestInfoProvider;
import ch.unige.solidify.validation.ValidationError;

import ch.dlcm.controller.AdminController;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.ArchiveACL;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.repository.ArchiveACLRepository;
import ch.dlcm.repository.OrganizationalUnitPersonRoleRepository;
import ch.dlcm.repository.UserRepository;
import ch.dlcm.service.rest.abstractservice.ArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedArchivalInfoPackageRemoteResourceService;

@Service
@ConditionalOnBean(AdminController.class)
public final class AclPermissionService extends AbstractPermissionService {
  Logger log = LoggerFactory.getLogger(AclPermissionService.class);
  private final HttpRequestInfoProvider httpRequestInfoProvider;

  private final ArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  private final ArchiveACLRepository archiveACLRepository;
  private final OrganizationalUnitPersonRoleRepository ouprRepository;
  private final UserRepository userRepository;

  public AclPermissionService(HttpRequestInfoProvider httpRequestInfoProvider,
          TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService,
          ArchiveACLRepository archiveACLRepository,
          UserRepository userRepository,
          OrganizationalUnitPersonRoleRepository ouprRepository) {
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.aipRemoteResourceService = aipRemoteResourceService;
    this.archiveACLRepository = archiveACLRepository;
    this.userRepository = userRepository;
    this.ouprRepository = ouprRepository;
  }

  public boolean hasReadPermissionOnAcl(ArchiveACL archiveACL) {
    return hasReadPermissionOnAclId(archiveACL.getResId());
  }

  public boolean hasReadPermissionOnAclId(String aclId) {
    if (this.isRootOrAdminOrTrustedRole()) {
      return true;
    }
    return hasManagerOrStewardRoleForAcl(aclId);
  }

  public boolean hasWritePermissionOnAclId(String aclId) {
    if (this.isRootOrTrustedRole()) {
      return true;
    }
    return hasManagerOrStewardRoleForAcl(aclId);
  }

  public boolean hasCreationPermissionOnAcl(ArchiveACL archiveACL) {
    if (this.isRootOrTrustedRole()) {
      return true;
    }
    final String aclOrgUnitId = this.getOrgUnitId(archiveACL);
    final String aipId = archiveACL.getAipId();
    if(aipId == null) {
      throw new SolidifyValidationException(new ValidationError("AIP id from ACL is null."));
    }
    final ArchivalInfoPackage aip = this.aipRemoteResourceService.findOne(aipId);
    final String aipOrgUnitId = aip.getOrganizationalUnitId();
    if(!aclOrgUnitId.equals(aipOrgUnitId)) {
      log.warn("OrgUnit Id from ACL {} differs from orgUnit Id from AIP {}.", aclOrgUnitId, aipOrgUnitId);
      return false;
    }
    return this.isManagerOrSteward(aclOrgUnitId);
  }

  public boolean hasListPermissionOnAcl(ArchiveACL searchAcl) {
    if (this.isRootOrAdminOrTrustedRole()) {
      return true;
    }
    final String aclOrgUnitId = this.getOrgUnitId(searchAcl);
    return this.isManagerOrSteward(aclOrgUnitId);
  }

  private boolean hasManagerOrStewardRoleForAcl(String aclId) {
    Optional<ArchiveACL> aclOptional = this.archiveACLRepository.findById(aclId);
    if(aclOptional.isEmpty()) {
      return false;
    }
    final String orgUnitId = aclOptional.get().getOrganizationalUnit().getResId();
    if(orgUnitId == null) {
      return false;
    }
    return this.isManagerOrSteward(orgUnitId);
  }

  private boolean isManagerOrSteward(String orgUnitId) {
    final Principal principal = this.httpRequestInfoProvider.getPrincipal();
    final User currentUser = this.userRepository.findByExternalUid(principal.getName());
    String personId = currentUser.getPerson().getResId();
    for (Role role : this.ouprRepository.findByOrganizationalUnitAndPerson(orgUnitId, personId)) {
      if (Role.MANAGER_ID.equals(role.getResId()) || Role.STEWARD_ID.equals(role.getResId())) {
        return true;
      }
    }
    return false;
  }

  private String getOrgUnitId(ArchiveACL archiveAcl) {
    final OrganizationalUnit aclOrgUnit = archiveAcl.getOrganizationalUnit();
    if(aclOrgUnit == null) {
      throw new SolidifyValidationException(new ValidationError("OrgUnit from ACL is null."));
    }
    final String aclOrgUnitId = aclOrgUnit.getResId();
    if (aclOrgUnitId == null) {
      throw new SolidifyValidationException(new ValidationError("OrgUnit Id from ACL is null"));
    }
    return aclOrgUnitId;
  }

}
