/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - ArchivalStorageEmailProcessingService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import javax.mail.MessagingException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import ch.unige.solidify.service.EmailService;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.business.ArchivalInfoPackageService;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.ArchivalStorageController;
import ch.dlcm.message.EmailMessage;
import ch.dlcm.model.oais.ArchivalInfoPackage;

@Service
@ConditionalOnBean(ArchivalStorageController.class)
@Profile("email-service")
public class ArchivalStorageEmailProcessingService extends EmailProcessingService {
  private static final Logger log = LoggerFactory.getLogger(ArchivalStorageEmailProcessingService.class);

  private final ArchivalInfoPackageService archivalInfoPackageService;

  protected ArchivalStorageEmailProcessingService(MessageService messageService, EmailService emailService,
          DLCMProperties dlcmProperties, ArchivalInfoPackageService archivalInfoPackageService) {
    super(messageService, emailService, dlcmProperties);
    this.archivalInfoPackageService = archivalInfoPackageService;
  }

  @Override
  protected void processEmailMessage(EmailMessage emailMessage) {
    log.info("Reading message by Storagion module {}", emailMessage);
    try {
      switch (emailMessage.getTemplate()) {
        case NEW_ARCHIVES_COMPLETED:
          this.sendEmailWithArchiveAndCreatorList(emailMessage);
          break;
        case MY_ARCHIVES_COMPLETED:
          this.sendEmailWithArchiveList(emailMessage);
          break;
        default:
          break;
      }
    } catch (MessagingException e) {
      log.error("Cannot send email ({}): {}", emailMessage.getTemplate().getSubject(), e.getMessage(), e);
    }
  }

  private void sendEmailWithArchiveList(EmailMessage emailMessage) throws MessagingException {
    List<ArchivalInfoPackage> aips = new ArrayList<>();
    for (String aipId : emailMessage.getParameters().keySet()) {
      aips.add(this.archivalInfoPackageService.findOne(aipId));
    }

    Map<String, Object> parameters = this.getEmailDefaultParameters();
    parameters.put("aips", aips);

    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(), Collections.singletonList(emailMessage.getTo()),
            emailMessage.getTemplate().getSubject(), parameters);
  }

  private void sendEmailWithArchiveAndCreatorList(EmailMessage emailMessage) throws MessagingException {
    List<ArchivalInfoPackage> aips = new ArrayList<>();
    List<String> creators = new ArrayList<>();
    for (String aipId : emailMessage.getParameters().keySet()) {
      ArchivalInfoPackage aip = this.archivalInfoPackageService.findOne(aipId);
      aips.add(aip);
      creators.add(aip.getCreatorName());
    }

    Map<String, Object> parameters = this.getEmailDefaultParameters();
    parameters.put("aips", aips);
    parameters.put("creators", creators);

    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(), Collections.singletonList(emailMessage.getTo()),
            emailMessage.getTemplate().getSubject(), parameters);
  }
}
