/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - FileStorageService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.storage.file;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.BagItTool;
import ch.unige.solidify.util.ChecksumTool;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.ZipTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.config.DLCMRepositoryDescription;
import ch.dlcm.controller.ArchivalStorageController;
import ch.dlcm.exception.FixityCheckLevelNotImplementedException;
import ch.dlcm.model.FixityCheckLevel;
import ch.dlcm.model.Package;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.preservation.StoredAIP;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveThumbnailRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.storage.StorageService;

@Profile("arc-file")
@Service
@ConditionalOnBean(ArchivalStorageController.class)
public class FileStorageService extends StorageService {

  private static final Logger log = LoggerFactory.getLogger(FileStorageService.class);

  private final Path storagePath;

  public FileStorageService(
          DLCMProperties dlcmProperties,
          DLCMRepositoryDescription repositoryDescription,
          MessageService messageService,
          MetadataService metadataService,
          FallbackOrganizationalUnitRemoteResourceService orgUnitResourceService,
          FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService,
          FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService) {
    super(dlcmProperties, repositoryDescription, messageService, metadataService, orgUnitResourceService,
            archiveThumbnailRemoteResourceService, archiveMetadataRemoteResourceService);
    this.storagePath = Paths.get(URI.create(dlcmProperties.getStorage().getUrl()));
    if (!FileTool.checkFile(this.storagePath)) {
      throw new SolidifyRuntimeException("Storage location (" + this.storagePath + ") does not exist");
    }
  }

  @Override
  public List<String> checkFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate, FixityCheckLevel fixityCheckLevel)
          throws NoSuchAlgorithmException, IOException {

    if (fixityCheckLevel == FixityCheckLevel.DEEP) {
      return this.checkDeepFixity(aip, algorithmsToRecalculate);
    }

    throw new FixityCheckLevelNotImplementedException(
            "FixityCheckLevel '" + fixityCheckLevel.name() + "' not implemented for storage of type " + this.getClass().getSimpleName());
  }

  @Override
  public long getAIPFileNumber(ArchiveContainer aipContainer, String aipId) {
    return this.getPackageFileNumber(aipContainer, this.getArchiveLocation(aipId));
  }

  @Override
  public Map<String, String> getAIPInfo(String aipId) throws IOException, JAXBException {
    final Map<String, String> infoList = new HashMap<>();
    final List<String> filesToExtract = new ArrayList<>();
    String metadataRelativePath = null;
    String researchDataRelativePath = null;
    // load zip archive
    final ZipTool archive = new ZipTool(this.getArchiveLocation(aipId));
    // Determine archive container
    if (archive.contains(BagItTool.BAGIT_FILE)) {
      infoList.put(DLCMConstants.AIP_CONTAINER, ArchiveContainer.BAG_IT.toString());
      // file to extract from archive
      filesToExtract.add(BagItTool.BAGIT_INFO_FILE);
      metadataRelativePath = this.getMetadataRelativePath(ArchiveContainer.BAG_IT);
      researchDataRelativePath = this.getRelativeFolder(ArchiveContainer.BAG_IT, Package.RESEARCH);
    } else {
      infoList.put(DLCMConstants.AIP_CONTAINER, ArchiveContainer.ZIP.toString());
      // file to extract from archive
      metadataRelativePath = this.getMetadataRelativePath(ArchiveContainer.ZIP);
      researchDataRelativePath = this.getRelativeFolder(ArchiveContainer.ZIP, Package.RESEARCH);
    }
    filesToExtract.add(metadataRelativePath);
    // Extract file from archive
    final Path aipPackage = this.getAIPFolder(aipId).resolve(DLCMConstants.PACKAGE);
    if (archive.unzipFiles(aipPackage, filesToExtract)) {
      if (FileTool.checkFile(aipPackage.resolve(BagItTool.BAGIT_INFO_FILE))) {
        this.completeAipInfoWithBagitInfo(infoList, aipPackage.resolve(BagItTool.BAGIT_INFO_FILE));
      }
      if (infoList.containsKey(DLCMConstants.METADATA_VERSION_FIELD)) {
        this.completeAipInfoWithMetadata(infoList, DLCMMetadataVersion.fromVersion(infoList.get(DLCMConstants.METADATA_VERSION_FIELD)),
                aipPackage.resolve(metadataRelativePath));
      }
    }
    // For units, it contains also a folder "researchdata" with the data files.
    if (infoList.containsKey(DLCMConstants.AIP_TOMBSTONE_SIZE) || archive.containsFolder(researchDataRelativePath)) {
      infoList.put(DLCMConstants.AIP_UNIT, String.valueOf(true));
    } else {
      infoList.put(DLCMConstants.AIP_UNIT, String.valueOf(false));
    }
    return infoList;
  }

  @Override
  public long getAIPSize(String aipId) {
    return FileTool.getSize(this.getArchiveLocation(aipId));
  }

  @Override
  public InputStream getAIPStream(String aipId) throws IOException {
    return new BufferedInputStream(Files.newInputStream(Paths.get(this.getArchiveLocation(aipId))));
  }

  @Override
  public URI getArchiveLocation(String aipId) {
    return this.storagePath.resolve(aipId).toUri();
  }

  @Override
  public FixityCheckLevel getDefaultFixityCheckLevel() {
    return FixityCheckLevel.DEEP;
  }

  @Override
  public PackageStatus getStatus(ArchivalInfoPackage aip) {
    // Check AIP
    final Path aipFile = Paths.get(aip.getDataFiles().get(0).getFinalData());
    if (!FileTool.checkFile(aipFile)) {
      return PackageStatus.IN_ERROR;
    }
    // Check file system storage
    final Path aipArchive = Paths.get(this.getArchiveLocation(this.getArchiveId(aip)));
    if (FileTool.checkFile(aipArchive)) {
      if (FileTool.getSize(aipFile) == FileTool.getSize(aipArchive)) {
        return PackageStatus.STORED;
      }
      return PackageStatus.IN_PROGRESS;
    }
    return PackageStatus.READY;
  }

  @Override
  public List<StoredAIP> listStoredAip() {
    final List<StoredAIP> list = new ArrayList<>();
    for (final Path orgUnit : FileTool.scanFolder(this.storagePath, Files::isDirectory)) {
      for (final Path aip : FileTool.scanFolder(orgUnit, Files::isRegularFile)) {
        final String id = orgUnit.getFileName() + SolidifyConstants.URL_SEP + aip.getFileName();
        try {
          list.add(this.getStoredAip(id));
        } catch (final URISyntaxException e) {
          log.error("Cannot find stored AIP ({})", id, e);
        }
      }
    }

    return list;
  }

  @Override
  public String storeAIP(ArchivalInfoPackage aip) {
    try {
      FileTool.copyFile(aip.getDataFiles().get(0).getFinalData(), this.getArchiveLocation(this.getArchiveId(aip)));
      return aip.getResId();
    } catch (final IOException e) {
      log.error("Cannot store AIP file '{}'", aip.getResId(), e);
      return null;
    }
  }

  @Override
  public void replaceAIP(ArchivalInfoPackage aip, Path newPackage) {
    try {
      final Path storagePath = Paths.get(this.getArchiveLocation(this.getArchiveId(aip)));
      // Delete archive
      FileTool.deleteFile(storagePath);
      // Replace by new package
      FileTool.moveFile(newPackage, storagePath);
    } catch (final IOException e) {
      throw new SolidifyProcessingException("Cannot store new AIP file '" + aip.getResId() + "'", e);
    }
  }

  @Override
  public String submitAIP(ArchivalInfoPackage aip) {
    final Path aipFile = Paths.get(aip.getDataFiles().get(0).getFinalData());
    final Path aipArchive = Paths.get(this.getArchiveLocation(this.getArchiveId(aip)));
    if (FileTool.checkFile(aipFile) && FileTool.checkFile(aipArchive) && FileTool.getSize(aipFile) == FileTool.getSize(aipArchive)) {
      return aipArchive.toUri().toString();
    }
    return null;
  }

  /**
   * Recalculate files hashes by reading the file entirely
   *
   * @param aip
   * @param algorithmsToRecalculate
   * @return
   * @throws NoSuchAlgorithmException
   * @throws IOException
   */
  private List<String> checkDeepFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate)
          throws NoSuchAlgorithmException, IOException {

    return Arrays.asList(ChecksumTool.computeChecksum(this.getAIPStream(this.getArchiveId(aip)), algorithmsToRecalculate));
  }

}
