/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - SwiftStorageService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.storage.swift;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import org.javaswift.joss.client.factory.AccountConfig;
import org.javaswift.joss.client.factory.AccountFactory;
import org.javaswift.joss.model.Account;
import org.javaswift.joss.model.Container;
import org.javaswift.joss.model.StoredObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.ChecksumTool;
import ch.unige.solidify.util.FileTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.config.DLCMRepositoryDescription;
import ch.dlcm.controller.ArchivalStorageController;
import ch.dlcm.exception.FixityCheckLevelNotImplementedException;
import ch.dlcm.model.FixityCheckLevel;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.preservation.StoredAIP;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveThumbnailRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.storage.StorageService;

@Profile("arc-swift")
@Service
@ConditionalOnBean(ArchivalStorageController.class)
public class SwiftStorageService extends StorageService {
  private static final Logger log = LoggerFactory.getLogger(SwiftStorageService.class);

  private final Account client;
  private final String container;

  public SwiftStorageService(
          DLCMProperties dlcmProperties,
          DLCMRepositoryDescription repositoryDescription,
          MessageService messageService,
          MetadataService metadataService,
          FallbackOrganizationalUnitRemoteResourceService orgUnitResourceService,
          FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService,
          FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService) {
    super(dlcmProperties, repositoryDescription, messageService, metadataService, orgUnitResourceService,
            archiveThumbnailRemoteResourceService, archiveMetadataRemoteResourceService);

    final AccountConfig accountConfig = new AccountConfig();
    accountConfig.setUsername(dlcmProperties.getStorage().getAccessKey());
    accountConfig.setPassword(dlcmProperties.getStorage().getSecretKey());
    accountConfig.setAuthUrl(dlcmProperties.getStorage().getUrl());
    accountConfig.setTenantName(dlcmProperties.getStorage().getSpace());
    this.client = new AccountFactory(accountConfig).createAccount();
    this.container = dlcmProperties.getStorage().getSpace();
  }

  @Override
  public List<String> checkFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate, FixityCheckLevel fixityCheckLevel)
          throws NoSuchAlgorithmException, IOException {

    if (fixityCheckLevel == FixityCheckLevel.SIMPLE) {
      return this.checkSimpleFixity(aip, algorithmsToRecalculate);
    } else if (fixityCheckLevel == FixityCheckLevel.DEEP) {
      return this.checkDeepFixity(aip, algorithmsToRecalculate);
    }

    throw new FixityCheckLevelNotImplementedException(
            "FixityCheckLevel '" + fixityCheckLevel.name() + "' not implemented for storage of type " + this.getClass().getSimpleName());
  }

  @Override
  public long getAIPFileNumber(ArchiveContainer aipContainer, String aipId) {
    return Long.parseLong(this.getObject(aipId).getMetadata(DLCMConstants.AIP_FILE_NUMBER).toString());
  }

  @Override
  public Map<String, String> getAIPInfo(String aipId) {
    final Map<String, Object> md = this.getObject(aipId).getMetadata();
    final Map<String, String> newsMd = new HashMap<>();
    for (final Map.Entry<String, Object> entry : md.entrySet()) {
      if (entry.getValue() instanceof String) {
        newsMd.put(entry.getKey(), (String) entry.getValue());
      } else {
        newsMd.put(entry.getKey(), entry.getValue().toString());
      }
    }
    return newsMd;
  }

  @Override
  public long getAIPSize(String aipId) {
    return this.getObject(aipId).getContentLength();
  }

  @Override
  public InputStream getAIPStream(String aipId) {
    final StoredObject storedObject = this.getObject(aipId);
    return storedObject.downloadObjectAsInputStream();
  }

  @Override
  public URI getArchiveLocation(String aipId) throws URISyntaxException {
    return new URI(this.getObject(aipId).getURL());
  }

  @Override
  public FixityCheckLevel getDefaultFixityCheckLevel() {
    return FixityCheckLevel.SIMPLE;
  }

  @Override
  public PackageStatus getStatus(ArchivalInfoPackage aip) {
    // Check AIP
    final Path aipFile = Paths.get(aip.getDataFiles().get(0).getFinalData());
    if (!FileTool.checkFile(aipFile)) {
      return PackageStatus.IN_ERROR;
    }
    // Check Storage
    if (this.getObject(this.getArchiveId(aip)).exists()) {
      return PackageStatus.STORED;
    }
    return PackageStatus.READY;
  }

  @Override
  public List<StoredAIP> listStoredAip() {
    final List<StoredAIP> list = new ArrayList<>();

    final Collection<StoredObject> storedObjectList = this.getSwiftContainer().list();
    for (final StoredObject storedObject : storedObjectList) {
      if (!storedObject.getName().endsWith(SolidifyConstants.URL_SEP)) {
        try {
          list.add(this.getStoredAip(storedObject.getName()));
        } catch (final URISyntaxException e) {
          log.error("Cannot find stored AIP (" + storedObject.getName() + ")", e);
        }
      }
    }
    return list;
  }

  @Override
  public String storeAIP(ArchivalInfoPackage aip) {
    try (BufferedInputStream in = new BufferedInputStream(
            Files.newInputStream(Paths.get(aip.getDataFiles().get(0).getFinalData())))) {
      // Set object metadata
      final Map<String, Object> md = new TreeMap<>();
      for (final Entry<String, Object> metadata : aip.getAIPMetadata().entrySet()) {
        md.put(metadata.getKey(), metadata.getValue());
      }
      // Upload data
      final StoredObject storedObject = this.getSwiftContainer().getObject(this.getArchiveId(aip));
      storedObject.setContentType(SolidifyConstants.ZIP_MIME_TYPE);
      storedObject.uploadObject(in);
      storedObject.setMetadata(md);
      return aip.getResId();
    } catch (final IOException e) {
      log.error("Error in storing AIP in container (" + aip.getResId() + ")", e);
    }
    return null;
  }

  @Override
  public void replaceAIP(ArchivalInfoPackage aip, Path newPackage) {
    // TODO to be implemented
  }

  @Override
  public String submitAIP(ArchivalInfoPackage aip) throws URISyntaxException {
    final Path aipFile = Paths.get(aip.getDataFiles().get(0).getFinalData());
    final StoredObject storedObject = this.getObject(this.getArchiveId(aip));
    if (FileTool.checkFile(aipFile) && FileTool.getSize(aipFile) == storedObject.getContentLength()) {
      return this.getArchiveLocation(this.getArchiveId(aip)).toString();
    }
    return null;
  }

  /**
   * Fetch the file from Swift storage and really recalculates the file hashes
   *
   * @param aip
   * @param algorithmsToRecalculate
   * @return
   * @throws NoSuchAlgorithmException
   * @throws IOException
   */
  private List<String> checkDeepFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToCalculate)
          throws NoSuchAlgorithmException, IOException {

    try (InputStream stream = this.getAIPStream(this.getArchiveId(aip))) {
      return ChecksumTool.getChecksums(stream, algorithmsToCalculate);
    }

  }

  /**
   * Get the requested available file hashes by using the one provided by Switf
   *
   * @param aip
   * @param algorithmsToRecalculate
   * @return the available hash values provided by Swift storage, if they are available (currently
   * only MD5 is returned)
   * @throws NoSuchAlgorithmException
   * @throws IOException
   */
  private List<String> checkSimpleFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate) {

    final List<String> hashes = new ArrayList<>();

    for (final ChecksumAlgorithm algo : algorithmsToRecalculate) {
      if (algo.name().equals("MD5")) {

        final String md5 = this.getObject(this.getArchiveId(aip)).getEtag();

        if (!md5.contains("-")) {

          /*
           * Case of MD5 of the whole file
           */
          hashes.add(md5);

        } else {

          /*
           * The file has been uploaded using multi part --> the returned MD5 is not the checksum of the
           * original file
           */
          // TODO: support fixity check for multi part files
          hashes.add(null);
        }

      } else {
        hashes.add(null);
      }
    }

    return hashes;
  }

  private StoredObject getObject(String archiveId) {
    return this.getSwiftContainer().getObject(archiveId);
  }

  private Container getSwiftContainer() {
    final Container swiftContainer = this.client.getContainer(this.container);
    if (!swiftContainer.exists()) {
      swiftContainer.create();
    }
    return swiftContainer;
  }

}
