/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - S3Test.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.storage;

import static org.junit.jupiter.api.Assertions.assertEquals;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.UUID;

import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import com.amazonaws.ClientConfiguration;
import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSStaticCredentialsProvider;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.client.builder.AwsClientBuilder;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.model.CreateBucketRequest;
import com.amazonaws.services.s3.model.ListObjectsRequest;
import com.amazonaws.services.s3.model.ObjectListing;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectInputStream;
import com.amazonaws.services.s3.model.S3ObjectSummary;

import ch.unige.solidify.SolidifyConstants;

@ExtendWith(SpringExtension.class)
@Disabled("Run these tests manually if you have an S3 account")
class S3Test {
  private static final String ACCESS_KEY = "change-it"; // May need to be base64 encoded value of your access key

  private static final String BUCKET = "name of an existing bucket from your s3 storage";
  private static final String DATA_DIRECTORY = "/example/path/";
  private static final String DATA_FILE = DATA_DIRECTORY + "example_file.txt";
  private static final boolean IS_PATH_STYLE_ACCESS = false; // false by default. Set to true if s3 storage requires path-style access
  private static final Logger log = LoggerFactory.getLogger(S3Test.class);
  private static final String OBJECT = "11111";
  private static final String SECRET_KEY = "change-it"; // May need to be MD5 hash value of your secret key
  private static final String TENANT = "s3 storage endpoint";

  @Test
  void createS3ObjectTest() {
    final AmazonS3 s3Client = this.getS3Client(TENANT, ACCESS_KEY, SECRET_KEY);
    s3Client.putObject(BUCKET, this.getObjectId(), new File(DATA_FILE));
  }

  @Test
  void dataLengthTest() throws IOException {
    final String objectId = this.getObjectId();
    final AmazonS3 s3Client = this.getS3Client(TENANT, ACCESS_KEY, SECRET_KEY);
    final File dataFile = new File(DATA_FILE);
    s3Client.putObject(BUCKET, objectId, dataFile);
    try (S3Object s3Object = s3Client.getObject(BUCKET, objectId)) {
      final long dataFromS3Length = s3Object.getObjectMetadata().getContentLength();
      assertEquals(dataFile.length(), dataFromS3Length);
    }
  }

  @Test
  void deleteS3ObjectTest() {
    final AmazonS3 s3Client = this.getS3Client(TENANT, ACCESS_KEY, SECRET_KEY);
    final String objectId = this.getObjectId();
    if (!s3Client.doesBucketExistV2(BUCKET)) {
      s3Client.createBucket(new CreateBucketRequest(BUCKET));
    }
    s3Client.putObject(BUCKET, objectId, new File(DATA_FILE));
    s3Client.deleteObject(BUCKET, objectId);
  }

  @Test
  void getS3ObjectTest() throws IOException {
    final AmazonS3 s3Client = this.getS3Client(TENANT, ACCESS_KEY, SECRET_KEY);
    final String objectId = this.getObjectId();
    s3Client.putObject(BUCKET, objectId, new File(DATA_FILE));
    try (BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(new File(DATA_DIRECTORY + objectId)),
            SolidifyConstants.BUFFER_SIZE)) {
      final S3Object o = s3Client.getObject(BUCKET, objectId);
      final S3ObjectInputStream s3is = o.getObjectContent();
      final byte[] read_buf = new byte[SolidifyConstants.BUFFER_SIZE];
      int read_len = 0;
      while ((read_len = s3is.read(read_buf)) > 0) {
        bos.write(read_buf, 0, read_len);
      }
      s3is.close();
    }
  }

  @Test
  void listS3ObjectTest() {
    final AmazonS3 s3Client = this.getS3Client(TENANT, ACCESS_KEY, SECRET_KEY);
    final ObjectListing objectListing = s3Client.listObjects(new ListObjectsRequest().withBucketName(BUCKET));
    for (final S3ObjectSummary objectSummary : objectListing.getObjectSummaries()) {
      log.info(" - " + objectSummary.getKey() + " " + "(size = " + objectSummary.getSize() + ")");
    }
  }

  private String getObjectId() {
    return OBJECT + "-" + UUID.randomUUID();
  }

  private AmazonS3 getS3Client(String url, String accessKey, String secretKey) {
    final AWSCredentials awsCredentials = new BasicAWSCredentials(accessKey, secretKey);

    final AwsClientBuilder.EndpointConfiguration awsConfig = new AwsClientBuilder.EndpointConfiguration(url, "");
    final ClientConfiguration awsClientConfig = new ClientConfiguration().withSignerOverride("S3SignerType");
    return AmazonS3ClientBuilder.standard()
            .withCredentials(new AWSStaticCredentialsProvider(awsCredentials))
            .withClientConfiguration(awsClientConfig)
            .withEndpointConfiguration(awsConfig)
            .withPathStyleAccessEnabled(IS_PATH_STYLE_ACCESS)
            .build();
  }

}
