/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - ResourceBaseClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.env.Environment;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.util.AbstractRestClientTool;
import ch.unige.solidify.util.ReflectionTool;

import ch.dlcm.rest.ModuleName;

public class ResourceBaseClientService<T extends Resource> {
  private static final Logger logger = LoggerFactory.getLogger(ResourceBaseClientService.class);

  protected final String resourcePath;

  @Autowired
  protected AbstractRestClientTool restClientTool;

  @Autowired
  private Environment env;

  private boolean isResourceServer = false;
  private final ObjectMapper mapper = new ObjectMapper();
  private final String module;
  private final Class<T> resourceClass;

  public ResourceBaseClientService(String resourcePath, Class<T> resourceClass, String module) {
    this.module = module;
    this.resourcePath = resourcePath;
    this.resourceClass = resourceClass;
    this.mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    this.mapper.registerModule(new JavaTimeModule());
  }

  public AbstractRestClientTool getRestClientTool() {
    return this.restClientTool;
  }

  public void setResourceServer(boolean resourceServer) {
    this.isResourceServer = resourceServer;
  }

  protected String getBaseUrl() {
    return this.getBaseUrl(this.module, 0);
  }

  protected String getBaseUrl(String module, int moduleIndex) {
    String propertyName = "dlcm.module." + module + ".url";
    if (module.equals(ModuleName.ARCHIVALSTORAGE)) {
      propertyName += "s";
    }
    String url = this.env.getProperty(propertyName);
    if (url == null) {
      throw new IllegalStateException("Property dlcm.module." + module + ".url has to be defined.");
    }
    // Manage if there is a list of URLs
    if (url.contains(",")) {
      url = url.split(",")[moduleIndex];
    }
    if (this.isResourceServer) {
      final int index = url.lastIndexOf('/');
      url = url.substring(0, index + 1).concat(ModuleName.RES_SRV);
    }
    return url;
  }

  protected Environment getEnv() {
    return this.env;
  }

  protected String getModule() {
    return this.module;
  }

  protected T getObjectFromUrl(String url, Map<String, String> params) {
    T result = null;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    try {
      final String jsonString = restTemplate.getForObject(url, String.class, params);
      if (jsonString == null) {
        return null;
      }
      result = this.mapper.readValue(jsonString, this.resourceClass);
    } catch (final IOException e) {
      logger.error("Error getting object from " + url, e);
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        // No item found return null
        return null;
      } else {
        throw e;
      }
    }
    return result;
  }

  protected Class<T> getResourceClass() {
    return this.resourceClass;
  }

  protected Map<String, Object> objectToMap(Object object, List<String> propertiesToMap) {
    if (object != null) {
      try {
        Map<String, Object> propertiesMap = new HashMap<>();
        for (String propertyName : propertiesToMap) {
          if (propertyName.contains(".")) {
            String parentPropertyName = propertyName.substring(0, propertyName.indexOf("."));
            String childPropertyName = propertyName.substring(propertyName.indexOf(".") + 1);

            Field field = ReflectionTool.getField(object, parentPropertyName);
            field.setAccessible(true);
            Object value = field.get(object);
            Map<String, Object> onePropertyMap = this.objectToMap(value, Arrays.asList(childPropertyName));

            if (!propertiesMap.containsKey(parentPropertyName)) {
              propertiesMap.put(parentPropertyName, new HashMap<String, Object>());
            }
            HashMap<String, Object> propertyMap = (HashMap<String, Object>) propertiesMap.get(parentPropertyName);
            for (Map.Entry<String, Object> entry : onePropertyMap.entrySet()) {
              propertyMap.put(entry.getKey(), entry.getValue());
            }
          } else {
            Field field = ReflectionTool.getField(object, propertyName);
            field.setAccessible(true);
            Object value = field.get(object);
            propertiesMap.put(propertyName, value);
          }
        }
        return propertiesMap;
      } catch (NoSuchFieldException | IllegalAccessException e) {
        throw new SolidifyRuntimeException("Cannot get a Map from the given object", e);
      }
    } else {
      throw new SolidifyRuntimeException("Cannot get a Map from a null object");
    }
  }
}
