/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - ResourceClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.lang.reflect.Array;
import java.net.URI;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.security.TokenUsage;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.DataFileChecksum;
import ch.dlcm.model.IdentifierType;
import ch.dlcm.model.dto.FileUploadDto;
import ch.dlcm.model.preingest.DepositDataFile;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

public abstract class ResourceClientService<T extends Resource> extends ResourceBaseClientService<T> implements
        ResourceClientServiceInterface<T> {
  protected static final String DLCM_ID_LABEL = DLCMConstants.DB_RES_ID;

  protected ResourceClientService(String resourcePath, Class<T> resourceClass, String module) {
    super(resourcePath, resourceClass, module);
  }

  protected ResourceClientService(String resourcePath, Class<T> resourceClass, String module, boolean isResourceServer) {
    super(resourcePath, resourceClass, module);
    this.setResourceServer(isResourceServer);
  }

  @Override
  public T create(T resource) {
    final String url = this.getBaseUrl() + this.resourcePath;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    return restTemplate.postForObject(url, resource, this.getResourceClass());
  }

  @Override
  public void delete(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}";
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    restTemplate.delete(url, params);
  }

  @Override
  public void deleteList(String[] ids) {
    final String url = this.getBaseUrl() + this.resourcePath;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.exchange(url, HttpMethod.DELETE, new HttpEntity<>(ids), String.class);
  }

  @Override
  public T findOne(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}";
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    return this.getObjectFromUrl(url, params);
  }

  public T findByIdentifierType(IdentifierType identifierType, String identifierValue) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}?" + DLCMConstants.IDENTIFIER_TYPE_PARAM
            + "={" + DLCMConstants.IDENTIFIER_TYPE_PARAM + "}";
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, identifierValue);
    params.put(DLCMConstants.IDENTIFIER_TYPE_PARAM, identifierType.name());
    return this.getObjectFromUrl(url, params);
  }

  public String[] getArchivalStorageBaseUrls() {
    final String propertyName = "dlcm.module." + ModuleName.ARCHIVALSTORAGE + ".urls";
    final String url = this.getEnv().getProperty(propertyName);
    if (url == null) {
      throw new IllegalStateException("Property dlcm.module." + ModuleName.ARCHIVALSTORAGE + ".url has to be defined.");
    }
    return url.split(",");
  }

  @Override
  public T update(String resId, T resource) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}";
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    return restTemplate.exchange(url, HttpMethod.PATCH, new HttpEntity<>(resource), this.getResourceClass(), params).getBody();
  }

  @Override
  public T update(String resId, T resource, List<String> propertiesToUpdate) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}";
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);

    // send given properties only
    Map<String, Object> newProperties = this.objectToMap(resource, propertiesToUpdate);

    return restTemplate.exchange(url, HttpMethod.PATCH, new HttpEntity<>(newProperties), this.getResourceClass(), params).getBody();
  }

  protected <U extends Resource> void addItem(String resId, String itemTypeLabel, Class<U> itemClass, String... ids) {
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), ids, itemClass, params);
  }

  protected <U extends Resource> void addItem(String resId, String itemTypeLabel, String itemId, Class<U> itemClass) {
    @SuppressWarnings("unchecked")
    final Class<? extends Resource[]> arrayType = (Class<? extends Resource[]>) Array.newInstance(itemClass, 0).getClass();
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), new String[] { itemId }, arrayType, params);
  }

  protected <U extends Resource> U addItem(String resId, String itemTypeLabel, U item, Class<U> itemClass) {
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    return restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), item, itemClass, resId);
  }

  protected <U extends Resource> void addRelationItem(String resId, String itemTypeLabel, Class<U> itemClass, String... ids) {
    @SuppressWarnings("unchecked")
    final Class<? extends Resource[]> arrayType = (Class<? extends Resource[]>) Array.newInstance(itemClass, 0).getClass();
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), ids, arrayType, params);
  }

  protected T doAction(String action) {
    final String url = this.getActionUrl(action);
    return this.getObjectFromUrl(url, new HashMap<>());
  }

  protected List<T> doActionOnList(String action, Class<T> itemClass) {
    final String url = this.getActionUrl(action) + "?size=" + RestCollectionPage.MAX_SIZE_PAGE;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected List<T> docActionOnListWithParameter(String action, Class<T> itemClass, String parameterName, String parameterValue) {
    final String url = this.getActionUrl(action) + "?size=" + RestCollectionPage.MAX_SIZE_PAGE + "&" + parameterName + "=" + parameterValue;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected T doActionOnId(String resId, String action) {
    final String url = this.getAssociationUrl(action);
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    return this.getObjectFromUrl(url, params);
  }

  protected T doActionWithId(String resId, String action) {
    final String url = this.getActionUrlWithId(action);
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    return this.getObjectFromUrl(url, params);
  }

  protected void downloadByUrl(String urlString, Path path) {
    final String url = this.getBaseUrl() + urlString;
    this.downloadWithToken(url, path);
  }

  protected void downloadDataFile(String resId, String itemId, Path path) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + ResourceName.DATAFILE
            + SolidifyConstants.URL_SEP + itemId + "/download";
    this.downloadWithToken(url, path);
  }

  protected void downloadDeposit(String resId, String folder, Path path) {
    String url = null;
    if (folder == null) {
      url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD;
    } else {
      url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD
              + "?folder="
              + folder;
    }
    this.downloadWithToken(url, path);
  }

  protected List<T> findAll(Class<T> itemClass) {
    final String url = this.getBaseUrl() + this.resourcePath + "?size=" + RestCollectionPage.MAX_SIZE_PAGE;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected List<T> findAllWithParameters(Class<T> itemClass, MultiValueMap<String, String> parameters) {
    final UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(this.getBaseUrl() + this.resourcePath);
    if (!parameters.containsKey("size")) {
      parameters.add("size", String.valueOf(RestCollectionPage.MAX_SIZE_PAGE));
    }
    uriBuilder.queryParams(parameters);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(uriBuilder.toUriString(), String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected List<T> findAllWithPagination(Class<T> itemClass) {
    final String url = this.getBaseUrl() + this.resourcePath + "?size=" + RestCollectionPage.MAX_SIZE_PAGE;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    RestCollectionPage page = CollectionTool.getPage(jsonString);
    List<T> listOfAllItems = CollectionTool.getList(jsonString, itemClass);
    for (int i = 1; i <= page.getTotalPages() - 1; i++) {
      String urlWithPage = this.getBaseUrl() + this.resourcePath + "?size=" + RestCollectionPage.MAX_SIZE_PAGE + "&page=" + i;
      String forObject = restTemplate.getForObject(urlWithPage, String.class);
      listOfAllItems.addAll(CollectionTool.getList(forObject, itemClass));
    }
    return listOfAllItems;
  }

  protected <U> List<U> findAllLinkedResources(String resId, String itemTypeLabel, Class<U> itemClass) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}" + SolidifyConstants.URL_SEP + itemTypeLabel
            + "?size=" + RestCollectionPage.MAX_SIZE_PAGE;
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class, params);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected void removeItem(String parentId, String itemTypeLabel, String itemId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + parentId + SolidifyConstants.URL_SEP + itemTypeLabel
            + SolidifyConstants.URL_SEP + itemId;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.delete(url);
  }

  protected void removeItem(String parentId, String itemTypeLabel, String itemId, String... subItemIds) {
    if (subItemIds.length == 0) {
      this.removeItem(parentId, itemTypeLabel, itemId);
    } else {
      final Map<String, String> params = new HashMap<>();
      params.put(DLCM_ID_LABEL, parentId);
      final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
      restTemplate
              .exchange(this.getAssociationUrl(itemTypeLabel + SolidifyConstants.URL_SEP + itemId), HttpMethod.DELETE,
                      new HttpEntity<>(subItemIds),
                      String.class, params);
    }
  }

  protected List<T> searchByProperties(Map<String, String> params, Class<T> itemClass) {
    String url = this.getBaseUrl() + this.resourcePath + "?size=" + RestCollectionPage.MAX_SIZE_PAGE;
    for (final Map.Entry<String, String> entry : params.entrySet()) {
      url += "&" + entry.getKey() + "=" + entry.getValue();
    }
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected <U extends Resource> U updateSubItem(String resId, String subItemTypeLabel, Class<U> subItemClass, String subItemId,
          Map<String, Object> mapProperties) {
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, resId);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    return restTemplate.exchange(
            this.getAssociationUrl(subItemTypeLabel) + SolidifyConstants.URL_SEP + subItemId,
            HttpMethod.PATCH,
            new HttpEntity<>(mapProperties),
            subItemClass,
            params)
            .getBody();
  }

  public <U> U[] uploadArchiveFile(String resId, ParameterizedTypeReference<U[]> responseType, org.springframework.core.io.Resource resource,
          FileUploadDto<DepositDataFile> fileUploadDto) {

    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + ActionName.UPLOAD_ARCHIVE;

    final MultiValueMap<String, Object> map = new LinkedMultiValueMap<>();

    map.add("file", resource);

    if (fileUploadDto.getDataCategory() != null) {
      map.add("category", fileUploadDto.getDataCategory().name());
    }

    if (fileUploadDto.getDataType() != null) {
      map.add("type", fileUploadDto.getDataType().name());
    }

    if (fileUploadDto.getMetadataType() != null) {
      map.add("metadataType", fileUploadDto.getMetadataType());
    }

    final HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.MULTIPART_FORM_DATA);

    final HttpEntity<MultiValueMap<String, Object>> requestEntity = new HttpEntity<>(map, headers);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final ResponseEntity<U[]> res = restTemplate.exchange(url, HttpMethod.POST, requestEntity, responseType);

    return res.getBody();
  }

  public <U> U uploadFile(String resId, ParameterizedTypeReference<U> responseType,
          org.springframework.core.io.Resource resource, FileUploadDto<?> fileUploadDto) {

    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + "/upload";

    final MultiValueMap<String, Object> map = new LinkedMultiValueMap<>();
    map.add("file", resource);

    if (fileUploadDto.getDataCategory() != null) {
      map.add("category", fileUploadDto.getDataCategory().name());
    }
    if (fileUploadDto.getDataType() != null) {
      map.add("type", fileUploadDto.getDataType().name());
    }
    if (fileUploadDto.getRelLocation() != null) {
      map.add("folder", fileUploadDto.getRelLocation());
    }
    if (fileUploadDto.getMetadataType() != null) {
      map.add("metadataType", fileUploadDto.getMetadataType());
    }

    if (fileUploadDto.getChecksums() != null) {
      Set<Map.Entry<ChecksumAlgorithm, String>> checksumsSet = fileUploadDto.getChecksums().entrySet();
      for (Map.Entry<ChecksumAlgorithm, String> checksum : checksumsSet) {
        if (!StringTool.isNullOrEmpty(checksum.getValue())) {
          map.add(this.getChecksumParamFromAlgo(checksum.getKey()), checksum.getValue());
        }
      }
    }

    if (fileUploadDto.getChecksumsOrigin() != null) {
      Set<Map.Entry<ChecksumAlgorithm, DataFileChecksum.ChecksumOrigin>> checksumsOriginSet = fileUploadDto.getChecksumsOrigin().entrySet();
      for (Map.Entry<ChecksumAlgorithm, DataFileChecksum.ChecksumOrigin> checksumOrigin : checksumsOriginSet) {
        map.add(this.getChecksumOriginParamFromAlgo(checksumOrigin.getKey()), checksumOrigin.getValue().name());
      }
    }

    final HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.MULTIPART_FORM_DATA);

    final HttpEntity<MultiValueMap<String, Object>> requestEntity = new HttpEntity<>(map, headers);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final ResponseEntity<U> res = restTemplate.exchange(url, HttpMethod.POST, requestEntity, responseType);

    return res.getBody();
  }

  private String getChecksumParamFromAlgo(ChecksumAlgorithm algo) {
    switch (algo) {
      case MD5:
        return DLCMConstants.CHECKSUM_MD5;
      case SHA1:
        return DLCMConstants.CHECKSUM_SHA1;
      case SHA256:
        return DLCMConstants.CHECKSUM_SHA256;
      case CRC32:
        return DLCMConstants.CHECKSUM_CRC32;
    }
    throw new SolidifyRuntimeException("Unknow checksum algo");
  }

  private String getChecksumOriginParamFromAlgo(ChecksumAlgorithm algo) {
    switch (algo) {
      case MD5:
        return DLCMConstants.CHECKSUM_MD5_ORIGIN;
      case SHA1:
        return DLCMConstants.CHECKSUM_SHA1_ORIGIN;
      case SHA256:
        return DLCMConstants.CHECKSUM_SHA256_ORIGIN;
      case CRC32:
        return DLCMConstants.CHECKSUM_CRC32_ORIGIN;
    }
    throw new SolidifyRuntimeException("Unknow checksum algo");
  }

  private void downloadWithToken(String url, Path path) {
    this.getRestClientTool().downloadContent(URI.create(url), path, TokenUsage.WITH_TOKEN);
  }

  private String getActionUrl(String actionLabel) {
    return this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + actionLabel;
  }

  private String getActionUrlWithId(String actionLabel) {
    return this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + actionLabel + SolidifyConstants.URL_SEP + "{" + DLCM_ID_LABEL
            + "}";
  }

  private String getAssociationUrl(String resourceLabel) {
    return this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}" + SolidifyConstants.URL_SEP + resourceLabel;
  }

  protected T uploadFileWithUrl(String url, org.springframework.core.io.Resource file, Class<T> itemClass) {
    final HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.MULTIPART_FORM_DATA);

    final MultiValueMap<String, Object> map = new LinkedMultiValueMap<>();
    map.add("file", file);

    final HttpEntity<MultiValueMap<String, Object>> requestEntity = new HttpEntity<>(map, headers);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());

    final ResponseEntity<T> res = restTemplate.exchange(url, HttpMethod.POST, requestEntity, itemClass);

    return res.getBody();
  }

  protected List<T> search(String search, String matchType, Class<T> itemClass) {
    String url = this.getBaseUrl() + this.resourcePath + "/search?search=" + search;
    if (!StringTool.isNullOrEmpty(matchType)) {
      url += "&match=" + matchType;
    }
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }
}
