/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - MetadataClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.access;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.env.Environment;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpMethod;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.model.Citation;
import ch.unige.solidify.model.dto.CitationDto;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.FacetResult;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.util.AbstractRestClientTool;
import ch.unige.solidify.util.CollectionTool;

import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.NoSqlClientService;

@Service
public class MetadataClientService extends NoSqlClientService<ArchiveMetadata> {

  public static final String ALL_METADATA = SolidifyConstants.URL_SEP + ResourceName.ALL_METADATA;

  public static final String PRIVATE_METADATA = SolidifyConstants.URL_SEP + ResourceName.PRIVATE_METADATA;

  public static final String PUBLIC_METADATA = SolidifyConstants.URL_SEP + ResourceName.PUBLIC_METADATA;

  public static final String SEARCH = SolidifyConstants.URL_SEP + ActionName.SEARCH;

  @Value("${dlcm.module.access.url}")
  private String accessUrl;

  private RestTemplate restTemplate;

  public MetadataClientService(AbstractRestClientTool restClientTool, Environment env) {
    super(ArchiveMetadata.class, ModuleName.ACCESS, restClientTool, env);
    this.restTemplate = this.restClientTool.getClient(this.getBaseUrl());
  }

  public void downloadArchive(String dipId, Path path) {
    final String url = PUBLIC_METADATA + SolidifyConstants.URL_SEP + dipId + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD;
    this.downloadWithToken(url, path);
  }

  public String getDownloadStatus(String resId) {
    final String url = PUBLIC_METADATA + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.DOWNLOAD_STATUS;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    return restTemplate.exchange(url, HttpMethod.GET, null, String.class, resId).getBody();
  }

  public RestCollection<ArchiveMetadata> listAllPrivate() {
    final Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
    final String jsonResult = this.getResources(PRIVATE_METADATA + "/" + ActionName.SEARCH, pageable);
    return this.getJson(jsonResult);
  }

  public RestCollection<ArchiveMetadata> listAllPrivateWithoutSecurity() {

    final String url = this.accessUrl + PRIVATE_METADATA;
    final String jsonResult = this.restTemplate.exchange(url, HttpMethod.GET, null, String.class).getBody();
    return this.getJson(jsonResult);
  }

  public RestCollection<ArchiveMetadata> listAllPublic() {
    final Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
    final String jsonResult = this.getResources(PUBLIC_METADATA, pageable);
    return this.getJson(jsonResult);
  }

  public RestCollection<ArchiveMetadata> listAllPublicWithoutSecurity() {
    final String url = this.accessUrl + PUBLIC_METADATA + "?size=" + RestCollectionPage.MAX_SIZE_PAGE;
    final String jsonResult = this.restTemplate.exchange(url, HttpMethod.GET, null, String.class).getBody();
    return this.getJson(jsonResult);
  }

  public RestCollection<ArchiveMetadata> listPublicAndPrivateMetadata() {
    final Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
    final String jsonResult = this.getResources(ALL_METADATA + "/" + ActionName.SEARCH, pageable);
    return this.getJson(jsonResult);
  }

  public RestCollection<ArchiveMetadata> listPublicAndPrivateMetadataWithoutSecurity() {
    final String url = this.accessUrl + ALL_METADATA;
    final String jsonResult = this.restTemplate.exchange(url, HttpMethod.GET, null, String.class).getBody();
    return this.getJson(jsonResult);
  }

  public ArchiveMetadata findOne(String resId) {
    final String url = this.accessUrl + PUBLIC_METADATA + SolidifyConstants.URL_SEP + resId;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final String jsonResult = restTemplate.getForObject(url, String.class);
    return this.getObjectFromUrl(url, jsonResult);
  }

  public ArchiveMetadata searchByDoi(String doi) {
    final String url = this.accessUrl + PUBLIC_METADATA + SolidifyConstants.URL_SEP + DLCMActionName.SEARCH_DOI + "?doi=" + doi;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final String jsonResult = restTemplate.getForObject(url, String.class);
    return this.getObjectFromUrl(url, jsonResult);
  }

  public RestCollection<ArchiveMetadata> searchPublicIndex(String query) {
    return this.searchIndex(query, PUBLIC_METADATA);
  }

  public RestCollection<ArchiveMetadata> searchPrivateIndex(String query) {
    return this.searchIndex(query, PRIVATE_METADATA);
  }

  public RestCollection<ArchiveMetadata> searchAllIndex(String query) {
    return this.searchIndex(query, ALL_METADATA);
  }

  public RestCollection<ArchiveMetadata> searchIndex(String query, String metadataUrl) {
    final String url = this.accessUrl + metadataUrl + SEARCH + "?query=" + query;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final String jsonResult = restTemplate.exchange(url, HttpMethod.GET, null, String.class, new Object[0]).getBody();
    return this.getJson(jsonResult);
  }

  public String prepareDownload(String resId) {
    final String url = PUBLIC_METADATA + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.PREPARE_DOWNLOAD;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    return restTemplate.exchange(url, HttpMethod.POST, null, String.class, resId).getBody();
  }

  public Map getAccessSystemProperties() {
    final String url = this.accessUrl + SolidifyConstants.URL_SEP + DLCMActionName.SYSTEM_PROPERTIES;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    return restTemplate.getForObject(url, Map.class);
  }

  public Map<String, List<String>> getCitationsConfig() {
    Map accessSystemProperties = this.getAccessSystemProperties();
    return (Map<String, List<String>>) accessSystemProperties.get("citations");
  }

  public Map<String, List<String>> getBibliographiesConfig() {
    Map accessSystemProperties = this.getAccessSystemProperties();
    return (Map<String, List<String>>) accessSystemProperties.get("bibliographies");
  }

  public List<CitationDto> getBibliographies(String resId) {
    return this.getCitationDtoList(
            this.accessUrl + PUBLIC_METADATA + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.BIBLIOGRAPHIES);
  }

  public List<CitationDto> getCitations(String resId) {
    return this.getCitationDtoList(
            this.accessUrl + PUBLIC_METADATA + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.CITATIONS);
  }

  private List<CitationDto> getCitationDtoList(String url) {
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    Map[] citationDtoMaps = restTemplate.getForObject(url, Map[].class);
    List<CitationDto> citationDtos = new ArrayList<>();
    for (Map citationMap : citationDtoMaps) {
      Citation citation = new Citation();
      citation.setStyle(citationMap.get("style").toString());
      citation.setLanguage(citationMap.get("language").toString());
      Citation.OutputFormat outputFormat = Citation.OutputFormat.valueOf(citationMap.get("outputFormat").toString().toUpperCase());
      citation.setOutputFormat(outputFormat);
      citation.setText(citationMap.get("text").toString());
      citationDtos.add(new CitationDto(citation));
    }
    return citationDtos;
  }

  private RestCollection<ArchiveMetadata> getJson(String jsonResult) {
    final List<ArchiveMetadata> list = CollectionTool.getList(jsonResult, ArchiveMetadata.class);
    final RestCollectionPage page = CollectionTool.getPage(jsonResult);
    final List<FacetResult> facets = CollectionTool.getFacetResults(jsonResult);
    return new RestCollection<>(list, page, facets);
  }

}
