/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - OrganizationalUnitClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.admin;

import java.io.IOException;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Result;

import ch.dlcm.model.display.OrganizationalUnitDisseminationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitPreservationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitSubmissionPolicyDisplay;
import ch.dlcm.model.display.PersonRole;
import ch.dlcm.model.policies.DisseminationPolicy;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.AdditionalFieldsForm;
import ch.dlcm.model.settings.FundingAgency;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.model.settings.ResearchDomain;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.ResourceClientService;

@Service
public class OrganizationalUnitClientService extends ResourceClientService<OrganizationalUnit> {

  private static final Logger logger = LoggerFactory.getLogger(OrganizationalUnitClientService.class);

  public OrganizationalUnitClientService() {
    super(SolidifyConstants.URL_SEP + ResourceName.ORG_UNIT, OrganizationalUnit.class, ModuleName.ADMIN);
  }

  public void addDisseminationPolicy(String orgUnitId, String policyId) {
    this.addItem(orgUnitId, ResourceName.DIS_POLICY, policyId, DisseminationPolicy.class);
  }

  public void addFundingAgency(String orgUnitId, String fundingAgencyId) {
    this.addItem(orgUnitId, ResourceName.FUNDING_AGENCY, fundingAgencyId, FundingAgency.class);
  }

  public void addInstitution(String orgUnitId, String institutionId) {
    this.addItem(orgUnitId, ResourceName.INSTITUTION, institutionId, Institution.class);
  }

  public void addResearchDomain(String orgUnitId, String researchDomain) {
    this.addItem(orgUnitId, ResourceName.RESEARCH_DOMAIN, researchDomain, ResearchDomain.class);
  }

  public void addPerson(String orgUnitId, String personId) {
    this.addItem(orgUnitId, ResourceName.PERSON, personId, Person.class);
  }

  public void addPerson(String orgUnitId, String personId, String roleName) {
    this.addRelationItem(orgUnitId, ResourceName.PERSON + SolidifyConstants.URL_SEP + personId, Role.class, roleName);
  }

  public void addPreservationPolicy(String orgUnitId, String policyId) {
    this.addItem(orgUnitId, ResourceName.PRES_POLICY, policyId, PreservationPolicy.class);
  }

  public void addSubmissionPolicy(String orgUnitId, String policyId) {
    this.addItem(orgUnitId, ResourceName.SUB_POLICY, policyId, SubmissionPolicy.class);
  }

  public void setSubmissionPolicyDefaultValue(String orgUnitId, String policyId, boolean defaultPolicy) {
    Map<String,Object> properties = Map.of("defaultPolicy", defaultPolicy);
    this.updateSubItem(orgUnitId, ResourceName.SUB_POLICY, SubmissionPolicy.class, policyId, properties);
  }

  public void setPreservationPolicyDefaultValue(String orgUnitId, String policyId, boolean defaultPolicy) {
    Map<String,Object> properties = Map.of("defaultPolicy", defaultPolicy);
    this.updateSubItem(orgUnitId, ResourceName.PRES_POLICY, PreservationPolicy.class, policyId, properties);
  }

  public AdditionalFieldsForm addAdditionalFieldsForm(String orgUnitId, AdditionalFieldsForm form) {
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}" + SolidifyConstants.URL_SEP
            + ResourceName.ADDITIONAL_FIELDS_FORMS;
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, orgUnitId);
    return restTemplate.postForObject(url, form, AdditionalFieldsForm.class, params);
  }

  public AdditionalFieldsForm getCurrentAdditionalFieldsForm(String orgUnitId) {
    String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + "{resId}" + SolidifyConstants.URL_SEP
            + ResourceName.ADDITIONAL_FIELDS_FORMS + SolidifyConstants.URL_SEP + ResourceName.CURRENT_VERSION;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, orgUnitId);

    try {
      final String jsonString = restTemplate.getForObject(url, String.class, params);
      if (jsonString == null) {
        return null;
      } else {
        ObjectMapper mapper = new ObjectMapper();
        mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        mapper.registerModule(new JavaTimeModule());
        return mapper.readValue(jsonString, AdditionalFieldsForm.class);
      }
    } catch (final IOException e) {
      logger.error("Error getting object from " + url, e);
      return null;
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        // No item found return null
        return null;
      } else {
        throw e;
      }
    }
  }

  public void close(String orgUnitId, String date) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + orgUnitId + SolidifyConstants.URL_SEP + ActionName.CLOSE
            + "?"
            + "closingDate=" + date;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());

    restTemplate.postForObject(url, null, Result.class);

  }

  @Override
  public List<OrganizationalUnit> findAll() {
    return this.findAll(OrganizationalUnit.class);
  }

  public List<OrganizationalUnitDisseminationPolicyDisplay> getDisseminationPolicies(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.DIS_POLICY, OrganizationalUnitDisseminationPolicyDisplay.class);
  }

  public List<FundingAgency> getFundingAgencies(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.FUNDING_AGENCY, FundingAgency.class);
  }

  public List<Institution> getInstitutions(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.INSTITUTION, Institution.class);
  }

  public List<ResearchDomain> getResearchDomains(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.RESEARCH_DOMAIN, ResearchDomain.class);
  }

  public List<Person> getPeople(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.PERSON, Person.class);
  }

  public PersonRole getPersonRole(String organizationalorgUnitId) {
    List<PersonRole> personRoleList = this.findAllLinkedResources(organizationalorgUnitId, ResourceName.PERSON, PersonRole.class);
    if (!personRoleList.isEmpty()) {
      return personRoleList.get(0);
    } else {
      return null;
    }
  }

  public List<OrganizationalUnitPreservationPolicyDisplay> getPreservationPolicies(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.PRES_POLICY, OrganizationalUnitPreservationPolicyDisplay.class);
  }

  public List<OrganizationalUnitSubmissionPolicyDisplay> getSubmissionPolicies(String orgUnitId) {
    return this.findAllLinkedResources(orgUnitId, ResourceName.SUB_POLICY, OrganizationalUnitSubmissionPolicyDisplay.class);
  }

  public void removeDisseminationPolicy(String orgUnitId, String policyId) {
    this.removeItem(orgUnitId, ResourceName.DIS_POLICY, policyId);
  }

  public void removeFundingAgency(String orgUnitId, String fundingAgencyId) {
    this.removeItem(orgUnitId, ResourceName.FUNDING_AGENCY, fundingAgencyId);
  }

  public void removeInstitution(String orgUnitId, String institutionId) {
    this.removeItem(orgUnitId, ResourceName.INSTITUTION, institutionId);
  }

  public void removeResearchDomain(String orgUnitId, String researchDomainId) {
    this.removeItem(orgUnitId, ResourceName.RESEARCH_DOMAIN, researchDomainId);
  }

  public void removePerson(String orgUnitId, String personId) {
    this.removeItem(orgUnitId, ResourceName.PERSON, personId);
  }

  public void removePerson(String orgUnitId, String personId, String roleName) {
    this.removeItem(orgUnitId, ResourceName.PERSON, personId, roleName);
  }

  public void removePreservationPolicy(String orgUnitId, String policyId) {
    this.removeItem(orgUnitId, ResourceName.PRES_POLICY, policyId);
  }

  public void removeSubmissionPolicy(String orgUnitId, String policyId) {
    this.removeItem(orgUnitId, ResourceName.SUB_POLICY, policyId);
  }

  @Override
  public List<OrganizationalUnit> searchByProperties(Map<String, String> properties) {
    return this.searchByProperties(properties, OrganizationalUnit.class);
  }

  public OrganizationalUnit uploadLogo(String resId, Resource file) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.UPLOAD_LOGO;
    return this.uploadFileWithUrl(url, file, OrganizationalUnit.class);
  }

  public void downloadLogo(String resId, Path path) {
    final String url = this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.DOWNLOAD_LOGO;
    this.downloadByUrl(url, path);
  }

  public void deleteLogo(String resId) {
    final String url = this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.DELETE_LOGO;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.delete(url);
  }

}
