/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - PersonClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.admin;

import java.nio.file.Path;
import java.util.List;
import java.util.Map;

import org.springframework.core.io.Resource;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.util.CollectionTool;

import ch.dlcm.model.IdentifierType;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.ResourceClientService;

@Service
public class PersonClientService extends ResourceClientService<Person> {
  private static final String PERSON_PATH = SolidifyConstants.URL_SEP + ResourceName.PERSON;

  public PersonClientService() {
    super(PERSON_PATH, Person.class, ModuleName.ADMIN);
  }

  public void addInstitution(String personId, String institutionId) {
    this.addItem(personId, ResourceName.INSTITUTION, institutionId, Institution.class);
  }

  public void addOrganizationalUnit(String personResId, String organizationalUnitResId) {
    this.addItem(personResId, ResourceName.ORG_UNIT, organizationalUnitResId, OrganizationalUnit.class);
  }

  public void addOrganizationalUnit(String personId, String organizationalUnitId, String... roleNames) {
    this.addRelationItem(personId, ResourceName.ORG_UNIT + SolidifyConstants.URL_SEP + organizationalUnitId, Role.class, roleNames);
  }

  @Override
  public List<Person> findAll() {
    return this.findAll(Person.class);
  }

  public List<Institution> getInstitutions(String personId) {
    return this.findAllLinkedResources(personId, ResourceName.INSTITUTION, Institution.class);
  }

  public List<OrganizationalUnit> getOrganizationalUnits(String personId) {
    return this.findAllLinkedResources(personId, ResourceName.ORG_UNIT, OrganizationalUnit.class);
  }

  public void removeOrganizationalUnit(String personId, String organizationalUnitId) {
    this.removeItem(personId, ResourceName.ORG_UNIT, organizationalUnitId);
  }

  public void removeOrganizationalUnit(String personId, String organizationalUnitId, String roleName) {
    this.removeItem(personId, ResourceName.ORG_UNIT, organizationalUnitId, roleName);
  }

  public List<Person> searchWithUser(String search) {
    String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + DLCMActionName.SEARCH_WITH_USER + "?search=" + search;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, Person.class);
  }

  public List<Person> search(String search, String matchType) {
    return this.search(search, matchType, Person.class);
  }

  @Override
  public List<Person> searchByProperties(Map<String, String> properties) {
    return this.searchByProperties(properties, Person.class);
  }

  public Person uploadAvatar(String resId, Resource file) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.UPLOAD_AVATAR;
    return this.uploadFileWithUrl(url, file, Person.class);
  }

  public void downloadAvatar(String resId, Path path) {
    final String url = this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + DLCMActionName.DOWNLOAD_AVATAR;
    this.downloadByUrl(url, path);
  }

  public void deleteAvatar(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.DELETE_AVATAR;
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    restTemplate.delete(url);
  }

  public Person findByOrcid(String orcid) {
    return this.findByIdentifierType(IdentifierType.ORCID, orcid);
  }
}
