/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - DepositClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.preingest;

import java.nio.file.Path;
import java.util.List;
import java.util.Map;

import org.springframework.core.ParameterizedTypeReference;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.model.dto.FileUploadDto;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.DepositDataFile;
import ch.dlcm.model.settings.Person;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.ResourceClientService;

@Service
public class DepositClientService extends ResourceClientService<Deposit> {

  private static final String DEPOSIT_PATH = SolidifyConstants.URL_SEP + ResourceName.DEPOSIT;

  public DepositClientService() {
    super(DEPOSIT_PATH, Deposit.class, ModuleName.PREINGEST);
  }

  public void addAip(String depositId, String aipId) {
    this.addItem(depositId, ResourceName.AIP, aipId, ArchivalInfoPackage.class);
  }

  public void addContributor(String depositId, String contributorId) {
    this.addItem(depositId, ResourceName.CONTRIBUTOR, contributorId, Person.class);
  }

  public void addDepositDataFile(String depositId, DepositDataFile df) {
    this.addItem(depositId, ResourceName.DATAFILE, df, DepositDataFile.class);
  }

  public void submit(String depositId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + depositId + SolidifyConstants.URL_SEP
            + DLCMActionName.SUBMIT_FOR_APPROVAL;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    restTemplate.exchange(url, HttpMethod.POST, null, String.class);
  }

  public void approve(String depositId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + depositId + SolidifyConstants.URL_SEP
            + DLCMActionName.APPROVE;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    restTemplate.exchange(url, HttpMethod.POST, null, String.class);
  }

  public Deposit createFromJson(String depositJson) {

    final HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.APPLICATION_JSON);
    final HttpEntity<String> request = new HttpEntity<>(depositJson, headers);

    final String url = this.getBaseUrl() + this.resourcePath;
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    return restTemplate.postForObject(url, request, Deposit.class);
  }

  @Override
  public void downloadDataFile(String depositId, String dataFileId, Path path) {
    super.downloadDataFile(depositId, dataFileId, path);
  }

  @Override
  public void downloadDeposit(String depositId, String folderName, Path path) {
    super.downloadDeposit(depositId, folderName, path);
  }

  @Override
  public List<Deposit> findAll() {
    return this.findAll(Deposit.class);
  }

  public List<Person> getContributors(String depositId) {
    return this.findAllLinkedResources(depositId, ResourceName.CONTRIBUTOR, Person.class);
  }

  public List<ArchivalInfoPackage> getDepositAip(String depositId) {
    return this.findAllLinkedResources(depositId, ResourceName.AIP, ArchivalInfoPackage.class);
  }

  public List<DepositDataFile> getDepositDataFile(String depositId) {
    return this.findAllLinkedResources(depositId, ResourceName.DATAFILE, DepositDataFile.class);
  }

  public void removeAip(String depositId, String aipId) {
    this.removeItem(depositId, ResourceName.AIP, aipId);
  }

  public void removeDepositDataFile(String depositId, String dataFileId) {
    this.removeItem(depositId, ResourceName.DATAFILE, dataFileId);
  }

  public void removeContributor(String depositId, String contributorId) {
    this.removeItem(depositId, ResourceName.CONTRIBUTOR, contributorId);
  }

  public String reserveDOI(String depositId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + depositId + SolidifyConstants.URL_SEP
            + ResourceName.RESERVE_DOI;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final Deposit deposit = restTemplate.exchange(url, HttpMethod.POST, null, Deposit.class).getBody();

    if (deposit != null) {
      return deposit.getDoi();
    } else {
      throw new SolidifyRuntimeException("Response to POSTing deposit (" + depositId + ") doesn't contain a body");
    }
  }

  public List<DepositDataFile> search(String depositId, String search, String matchType) {
    String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + depositId + SolidifyConstants.URL_SEP
            + ResourceName.DATAFILE
            + SolidifyConstants.URL_SEP + "search?search=" + search;
    if (!StringTool.isNullOrEmpty(matchType)) {
      url += "&match=" + matchType;
    }
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, DepositDataFile.class);
  }

  @Override
  public List<Deposit> searchByProperties(Map<String, String> properties) {
    return this.searchByProperties(properties, Deposit.class);
  }

  public DepositDataFile updateDepositDataFile(String depositId, String dataFileId, Map<String, Object> mapProperties) {
    return this.updateSubItem(depositId, ResourceName.DATAFILE, DepositDataFile.class, dataFileId, mapProperties);
  }

  public DepositDataFile uploadDepositDataFile(String resId, Resource resource) {
    return this.uploadFile(resId, new ParameterizedTypeReference<>() {
    }, resource, new FileUploadDto<DepositDataFile>());
  }

  public DepositDataFile uploadDepositDataFile(String resId, Resource resource, FileUploadDto<DepositDataFile> fileUploadDto) {
    return this.uploadFile(resId, new ParameterizedTypeReference<>() {
    }, resource, fileUploadDto);
  }

  public DepositDataFile[] uploadDepositDataFileArchive(String resId, Resource resource) {
    return this.uploadArchiveFile(resId, new ParameterizedTypeReference<>() {
    }, resource, new FileUploadDto<>());
  }

  public DepositDataFile[] uploadDepositDataFileArchive(String resId, Resource resource, FileUploadDto<DepositDataFile> fileUploadDto) {
    return this.uploadArchiveFile(resId, new ParameterizedTypeReference<>() {
    }, resource, fileUploadDto);
  }

  public void startEditingMetadata(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.START_METADATA_EDITING;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    restTemplate.exchange(url, HttpMethod.POST, null, String.class);
  }

  public void cancelEditingMetadata(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.CANCEL_METADATA_EDITING;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    restTemplate.exchange(url, HttpMethod.POST, null, String.class);
  }

  public void putDepositInError(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.PUT_IN_ERROR;

    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    restTemplate.exchange(url, HttpMethod.POST, null, String.class);
  }

  public <U> U uploadMetadataFile(String resId, ParameterizedTypeReference<U> responseType, org.springframework.core.io.Resource resource) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP
            + DLCMActionName.UPLOAD_METADATA;

    final HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.MULTIPART_FORM_DATA);

    final MultiValueMap<String, Object> body = new LinkedMultiValueMap<>();
    body.add("file", resource);
    final HttpEntity<MultiValueMap<String, Object>> requestEntity = new HttpEntity<>(body, headers);
    final RestTemplate restTemplate = this.getRestClientTool().getClient(this.getBaseUrl());
    final ResponseEntity<U> res = restTemplate.exchange(url, HttpMethod.POST, requestEntity, responseType);

    return res.getBody();
  }

}
