/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Data Management - IndexResourceServiceTest.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.indexing;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

import javax.xml.bind.JAXBException;

import org.mockito.Mock;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;

import ch.unige.solidify.IndexProperties;
import ch.unige.solidify.index.settings.IndexingSettingsService;
import ch.unige.solidify.model.index.IndexSettings;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.service.GitInfoProperties;
import ch.unige.solidify.service.IndexResourceService;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.SolidifyTime;

import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.config.DLCMRepositoryDescription;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.service.HistoryService;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.rest.fallback.FallbackArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveTypeRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackLanguageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackLicenseRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackMetadataTypeRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackPersonRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackPreservationPolicyRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackSubmissionInfoPackageRemoteResourceService;

public abstract class IndexResourceServiceTest {

  private static final Path DLCM_METADATA = Paths.get("src", "test", "resources").resolve("dlcm.xml");
  private static final String METADATA_ID = "metadataId";

  protected IndexProperties indexProperties;
  protected DLCMProperties dlcmProperties;
  protected DLCMRepositoryDescription repositoryDescription;
  protected IndexResourceService<String, ArchiveMetadata> indexResourceService;
  protected MetadataService metadataService;
  protected IndexingSettingsService<IndexSettings> settingsService;

  @Mock
  protected MessageService messageService;
  @Mock
  protected GitInfoProperties gitInfoProperties;
  @Mock
  protected HistoryService historyService;

  @Mock
  protected FallbackLanguageRemoteResourceService fallbackLanguageResourceService;
  @Mock
  protected FallbackLicenseRemoteResourceService fallbackLicenseRemoteResourceService;
  @Mock
  protected FallbackArchivalInfoPackageRemoteResourceService fallbackAipResourceService;
  @Mock
  protected FallbackMetadataTypeRemoteResourceService fallbackMetadataTypeResourceService;
  @Mock
  protected FallbackOrganizationalUnitRemoteResourceService fallbackOrgUnitResourceService;
  @Mock
  protected FallbackPersonRemoteResourceService fallbackPersonRemoteResourceService;
  @Mock
  protected FallbackArchiveMetadataRemoteResourceService fallbackArchiveMetadataResourceService;
  @Mock
  protected FallbackSubmissionInfoPackageRemoteResourceService fallbackSubmissionInfoPackageRemoteResourceService;
  @Mock
  protected FallbackPreservationPolicyRemoteResourceService fallbackPreservationPolicyRemoteService;
  @Mock
  protected FallbackArchiveTypeRemoteResourceService fallbackArchiveTypeResourceService;

  public void runIndexingTest() throws IOException, JAXBException {
    // Initialize indexes
    this.settingsService.init(this.dlcmProperties.getIndexing().getIndexDefinitionList());
    final List<IndexSettings> list = this.settingsService.findAll(null);
    assertEquals(3, list.size(), "Check index number");
    // Check public index
    this.testIndex(this.dlcmProperties.getIndexing().getIndexName());
    // Check private index
    this.testIndex(this.dlcmProperties.getIndexing().getPrivateIndexName());
    // Check registry index
    this.testIndex(this.dlcmProperties.getIndexing().getRegistryName());
    // Delete indexes
    for (final IndexSettings index : list) {
      assertTrue(this.settingsService.delete(index), "Delete index " + index.getResId());
    }
  }

  public void runSearchingTest() throws IOException, JAXBException {
    // Initialize indexes
    this.settingsService.init(this.dlcmProperties.getIndexing().getIndexDefinitionList());
    // Check public index
    this.testIndex(this.dlcmProperties.getIndexing().getIndexName());
    // Check private index
    this.testIndex(this.dlcmProperties.getIndexing().getPrivateIndexName());
    // Check registry index
    this.testIndex(this.dlcmProperties.getIndexing().getRegistryName());
    // Delete indexes
    for (final IndexSettings index : this.settingsService.findAll(null)) {
      assertTrue(this.settingsService.delete(index), "Delete index " + index.getResId());
    }
  }

  private void checkMetadata(ArchiveMetadata original, ArchiveMetadata metadata) {
    assertEquals(original.getResId(), metadata.getResId(), "[" + original.getIndex() + "] Check ID");
    assertEquals(original.getIndex(), metadata.getIndex(), "[" + original.getIndex() + "] Check index");
    assertEquals(original.getMetadata().size(), metadata.getMetadata().size(), "[" + original.getIndex() + "] Check metadata");
  }

  private ArchiveMetadata getMetadata(String indexName) throws IOException, JAXBException {
    final ArchiveMetadata am = new ArchiveMetadata();
    am.setResId(METADATA_ID);
    am.setIndex(indexName);
    return this.metadataService.completeArchiveIndex(DLCMMetadataVersion.getDefaultVersion(), am, DLCM_METADATA);
  }

  private void testIndex(String indexName) throws IOException, JAXBException {
    // Test index if empty
    final List<ArchiveMetadata> metadataList = this.indexResourceService.findAll(indexName, null);
    assertEquals(0, metadataList.size(), "Check '" + indexName + "' index number");
    // Add metadata
    final ArchiveMetadata metadata0 = this.getMetadata(indexName);
    final ArchiveMetadata metadata1 = this.indexResourceService.save(indexName, metadata0);
    SolidifyTime.waitOneSecond(); // wait for commit in elasticsearch
    this.checkMetadata(metadata0, metadata1);
    // Check if metadata exist
    final ArchiveMetadata metadata2 = this.indexResourceService.findOne(indexName, metadata0.getResId());
    this.checkMetadata(metadata0, metadata2);
    // Delete metadata
    assertTrue(this.indexResourceService.delete(indexName, metadata2), "[" + indexName + "] Delete metadata");
    SolidifyTime.waitOneSecond(); // wait for commit in elasticsearch
    assertNull(this.indexResourceService.findOne(indexName, metadata0.getResId()), "[" + indexName + "] Check if metadata deleted");
  }

  private void testSearch(String indexName) throws IOException, JAXBException {
    // Add metadata
    final ArchiveMetadata metadata0 = this.getMetadata(indexName);
    final ArchiveMetadata metadata1 = this.indexResourceService.save(indexName, metadata0);
    SolidifyTime.waitOneSecond(); // wait for commit in elasticsearch
    this.checkMetadata(metadata0, metadata1);
    final String query = "dlcm";
    final Page<ArchiveMetadata> result = this.indexResourceService.search(indexName, query, PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE));
    assertEquals(1, result.getTotalElements(), "Check result");
    this.checkMetadata(metadata0, result.getContent().get(0));
  }

}
