/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - MetadataIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.access;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.collection.IsIterableContainingInAnyOrder.containsInAnyOrder;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;

import ch.unige.solidify.model.dto.CitationDto;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.DisseminationInfoPackage;

class MetadataIT extends AbstractAccessIT {
  private static final Logger logger = LoggerFactory.getLogger(AbstractAccessIT.class);

  @Test
  void downloadArchiveTest() throws IOException {
    final RestCollection<ArchiveMetadata> archives = this.metadataService.listAllPublic();
    if (archives.getPage().getTotalItems() > 0) {
      // get the aip from collection created by init test
      final List<ArchivalInfoPackage> aipList = this.aipService.searchByProperties(Map.of("info.name", DLCMConstants.ARCHIVAL_UNIT_LABEL + " 1"));
      // check-status and prepare download
      if (!aipList.isEmpty()) {
        final String aipId = aipList.get(0).getResId();
        // check if order has been already created
        try {
          this.metadataService.getDownloadStatus(aipId);
        } catch (final HttpClientErrorException e) {
          this.metadataService.prepareDownload(aipId);
          this.waitOrderIsReadyOrInError(aipId);
        }

        Path path1 = Paths.get(System.getProperty("java.io.tmpdir"), "download1.tmp");
        this.metadataService.downloadArchive(aipId, path1);
        assertNotNull(path1.toFile());

        // compare the archive with the one return by dip
        Path path2 = Paths.get(System.getProperty("java.io.tmpdir"), "download2.tmp");
        final DisseminationInfoPackage dip = this.dipService.searchByProperties(Map.of("info.name", aipId)).get(0);
        this.dipService.downloadArchive(dip.getResId(), path2);

        assertNotNull(path2.toFile());
        assertArrayEquals(Files.readAllBytes(path1), Files.readAllBytes(path2));
      }
    }
  }

  @Test
  void bibliographiesTest() {
    final List<ArchivalInfoPackage> aipList = this.aipService.searchByProperties(Map.of("info.name", DLCMConstants.ARCHIVAL_UNIT_LABEL + " 1"));
    if (!aipList.isEmpty()) {
      Map<String, List<String>> bibliographiesConfig = this.metadataService.getBibliographiesConfig();

      final String aipId = aipList.get(0).getResId();
      List<CitationDto> citations = this.metadataService.getBibliographies(aipId);
      assertNotNull(citations);

      // checks that all combinations of configured properties have a match
      for (String style : bibliographiesConfig.get("styles")) {
        for (String language : bibliographiesConfig.get("languages")) {
          for (String outputFormat : bibliographiesConfig.get("outputFormats")) {
            CitationDto citationDto = this.getCitation(citations, style, language, outputFormat);
            assertNotNull(citationDto);
            assertFalse(StringTool.isNullOrEmpty(citationDto.getText()));
          }
        }
      }
    }
  }

  @Test
  void citationsTest() {
    final List<ArchivalInfoPackage> aipList = this.aipService.searchByProperties(Map.of("info.name", DLCMConstants.ARCHIVAL_UNIT_LABEL + " 1"));
    if (!aipList.isEmpty()) {
      Map<String, List<String>> citationsConfig = this.metadataService.getCitationsConfig();

      final String aipId = aipList.get(0).getResId();
      List<CitationDto> citations = this.metadataService.getCitations(aipId);
      assertNotNull(citations);

      // checks that all combinations of configured properties have a match
      for (String style : citationsConfig.get("styles")) {
        for (String language : citationsConfig.get("languages")) {
          for (String outputFormat : citationsConfig.get("outputFormats")) {
            CitationDto citationDto = this.getCitation(citations, style, language, outputFormat);
            assertNotNull(citationDto);
            assertFalse(StringTool.isNullOrEmpty(citationDto.getText()));
          }
        }
      }
    }
  }

  private CitationDto getCitation(List<CitationDto> citationDtos, String style, String language, String outputFormat) {
    Optional<CitationDto> citationDtoOpt = citationDtos.stream()
            .filter(c -> c.getStyle().equalsIgnoreCase(style)
                    && c.getLanguage().equalsIgnoreCase(language.toLowerCase())
                    && c.getOutputFormat().equalsIgnoreCase(outputFormat))
            .findFirst();
    return citationDtoOpt.orElse(null);
  }

  @Test
  void listPublicAndPrivateMetadataTest() {
    final List<ArchiveMetadata> metadataArchivesList = this.metadataService.listPublicAndPrivateMetadata().getData();
    final List<ArchiveMetadata> dataMgmtArchivesList = this.dataMgmtService.listAllPublicAndPrivateAIP().getData();
    this.verifyMetadataList(dataMgmtArchivesList, metadataArchivesList);
  }

  @Test
  void listPublicAndPrivateMetadataWithoutSecurityTest() {
    assertThrows(HttpClientErrorException.Forbidden.class, () -> this.metadataService.listPublicAndPrivateMetadataWithoutSecurity());
  }

  @Test
  void listAllPrivateTest() {
    final List<ArchiveMetadata> metadataArchivesList = this.metadataService.listAllPrivate().getData();
    final List<ArchiveMetadata> dataMgmtArchivesList = this.dataMgmtService.listAllPrivateAIP().getData();
    this.verifyMetadataList(dataMgmtArchivesList, metadataArchivesList);
  }

  @Test
  void listAllPrivateWithoutSecurityTest() {
    assertThrows(HttpClientErrorException.Forbidden.class, () -> this.metadataService.listAllPrivateWithoutSecurity());
  }

  @Test
  void listAllPublicTest() {
    final List<ArchiveMetadata> metadataArchivesList = this.metadataService.listAllPublic().getData();
    final List<ArchiveMetadata> dataMgmtArchivesList = this.dataMgmtService.listAllPublicAIP().getData();
    this.verifyMetadataList(dataMgmtArchivesList, metadataArchivesList);
  }

  @Test
  void listAllPublicWithoutSecurityTest() {
    final List<ArchiveMetadata> metadataArchivesList = this.metadataService.listAllPublicWithoutSecurity().getData();
    final List<ArchiveMetadata> dataMgmtArchivesList = this.dataMgmtService.listAllPublicAIP().getData();
    this.verifyMetadataList(dataMgmtArchivesList, metadataArchivesList);
  }

  @Test
  void searchPublicIndexContainsFacet() {
    RestCollection<ArchiveMetadata> results = this.metadataService.searchPublicIndex("*");
    assertFalse(results.getFacets().isEmpty());
  }

  @Test
  void searchPrivateIndexContainsFacet() {
    RestCollection<ArchiveMetadata> results = this.metadataService.searchPrivateIndex("*");
    assertFalse(results.getFacets().isEmpty());
  }

  @Test
  void searchAllIndexContainsFacet() {
    RestCollection<ArchiveMetadata> results = this.metadataService.searchAllIndex("*");
    assertFalse(results.getFacets().isEmpty());
  }

  @Test
  void searchWithSpecialCharacters() {
    try {
      this.metadataService.searchPublicIndex("?=)");
      fail("A BAD_REQUEST Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
      assertNotNull(e.getMessage());
      assertTrue(e.getMessage().contains(DLCMConstants.BAD_SEARCH_REQUEST));
    }
  }

  @Test
  void searchWithoutSpecialCharacters() {
    RestCollection<ArchiveMetadata> results = this.metadataService.searchPublicIndex("abc");
    assertNotNull(results.getPage());
  }

  private void verifyMetadataList(List<ArchiveMetadata> dataMgmtArchivesList, List<ArchiveMetadata> metadataArchivesList) {
    final int nbDataMgmtArchives = dataMgmtArchivesList.size();
    final int nbMetadataArchives = metadataArchivesList.size();
    assertEquals(nbDataMgmtArchives, nbMetadataArchives);
    if (nbDataMgmtArchives <= RestCollectionPage.MAX_SIZE_PAGE) {
      assertThat(dataMgmtArchivesList, containsInAnyOrder(metadataArchivesList.toArray()));
    } else {
      logger.warn("ArchiveMetadata list contains " + nbDataMgmtArchives + " elements which is bigger then MAX_SIZE_PAGE (" +
              RestCollectionPage.MAX_SIZE_PAGE + "), only the size is checked");
    }
  }

}
