/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - OrganizationalUnitIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;

import org.junit.jupiter.api.Test;
import org.springframework.core.io.ClassPathResource;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.HttpServerErrorException;

import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.display.OrganizationalUnitDisseminationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitPreservationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitSubmissionPolicyDisplay;
import ch.dlcm.model.display.PersonRole;
import ch.dlcm.model.policies.DisseminationPolicy;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.FundingAgency;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.License;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.model.settings.ResearchDomain;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

class OrganizationalUnitIT extends AbstractAdminIT {

  @Override
  protected void setUser() {
    this.restClientTool.sudoAdmin();
  }

  @Test
  void addAndRemoveDisseminationPolicyTest() {
    // Get test orgUnit
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #DISSEMINATION POLICY");

    List<OrganizationalUnitDisseminationPolicyDisplay> policies = this.organizationalUnitService
            .getDisseminationPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    final DisseminationPolicy policy = this.createRemoteDisseminationPolicy();

    // Add policy to orgUnit
    this.organizationalUnitService.addDisseminationPolicy(organizationalUnit.getResId(), policy.getResId());
    policies = this.organizationalUnitService.getDisseminationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test add
    for (final OrganizationalUnitDisseminationPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getType(), policy2.getType());
        assertEquals(policy.getDestinationServer(), policy2.getDestinationServer());
      }
    }

    // Remove policy
    this.organizationalUnitService.removeDisseminationPolicy(organizationalUnit.getResId(), policy.getResId());

    // Test remove
    policies = this.organizationalUnitService.getDisseminationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState);
  }

  @Test
  void addAndRemoveFundingAgency() {
    /*
     * Get test orgUnit
     */
    final OrganizationalUnit orgUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #FUNDING AGENCY");
    List<FundingAgency> fundingAgencies = this.organizationalUnitService.getFundingAgencies(orgUnit.getResId());
    final int initialState = fundingAgencies.size();

    /*
     * Create a new FundingAgency
     */
    final String fundingAgencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String fundingAgencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(fundingAgencyAcronym);
    new_agency.setName(fundingAgencyName);
    new_agency = this.fundingAgencyService.create(new_agency);

    /*
     * Add the FundingAgency to the OrganizationalUnit
     */
    this.organizationalUnitService.addFundingAgency(orgUnit.getResId(), new_agency.getResId());

    /*
     * Check it has been added
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(orgUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 1, fundingAgencies.size());
    final Optional<FundingAgency> optionalFA = fundingAgencies.stream().filter(fa -> fa.getName().equals(fundingAgencyName)).findFirst();
    assertTrue(optionalFA.isPresent());
    assertEquals(fundingAgencyName, optionalFA.get().getName());

    /*
     * Remove the FundingAgency
     */
    this.organizationalUnitService.removeFundingAgency(orgUnit.getResId(), new_agency.getResId());

    /*
     * Check it has been removed
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(orgUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState, fundingAgencies.size());

    /*
     * Create 3 FundingAgencies and link them to the OrganizationalUnit
     */
    final FundingAgency subFA1 = new FundingAgency();
    subFA1.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 1"));
    subFA1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 1"));

    final FundingAgency subFA2 = new FundingAgency();
    subFA2.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 2"));
    subFA2.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 2"));

    final FundingAgency subFA3 = new FundingAgency();
    subFA3.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 3"));
    subFA3.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 3"));

    final FundingAgency[] subAgencies = new FundingAgency[] { subFA1, subFA2, subFA3 };

    String agencyIdToDelete = null;

    int i = 0;
    for (FundingAgency subAgency : subAgencies) {
      subAgency = this.fundingAgencyService.create(subAgency);
      this.organizationalUnitService.addFundingAgency(orgUnit.getResId(), subAgency.getResId());

      if (i == 1) {
        agencyIdToDelete = subAgency.getResId();
      }

      i++;
    }

    /*
     * Check they are all linked correctly
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(orgUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 3, fundingAgencies.size());

    /*
     * Delete one specifically and check it is not in the list anymore
     */
    this.organizationalUnitService.removeFundingAgency(orgUnit.getResId(), agencyIdToDelete);
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(orgUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 2, fundingAgencies.size());
    for (final FundingAgency fetched_agency : fundingAgencies) {
      assertNotEquals(agencyIdToDelete, fetched_agency.getResId());
    }

  }

  @Test
  void addAndRemoveInstitutionTest() {
    // Get test orgUnit
    final OrganizationalUnit orgUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #INSTITUTION");
    List<Institution> list = this.organizationalUnitService.getInstitutions(orgUnit.getResId());
    final int initialState = list.size();

    // Create an institution
    Institution institution1 = new Institution();
    institution1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("institution "));
    institution1.setDescription("description" + ThreadLocalRandom.current().nextInt());
    institution1 = this.institutionService.create(institution1);

    // Add the institution to the OrganizationalUnit
    this.organizationalUnitService.addInstitution(orgUnit.getResId(), institution1.getResId());
    list = this.organizationalUnitService.getInstitutions(orgUnit.getResId());
    assertEquals(initialState + 1, list.size());
    final String institutionName = institution1.getName();
    final Optional<Institution> optionalInstitution = list.stream().filter(institution -> institution.getName().equals(institutionName))
            .findFirst();
    assertTrue(optionalInstitution.isPresent());

    // Test the institution
    assertEquals(institution1.getResId(), optionalInstitution.get().getResId());
    assertEquals(institution1.getName(), optionalInstitution.get().getName());
    assertEquals(institution1.getDescription(), optionalInstitution.get().getDescription());

    this.organizationalUnitService.removeInstitution(orgUnit.getResId(), institution1.getResId());
    list = this.organizationalUnitService.getInstitutions(orgUnit.getResId());
    assertEquals(initialState, list.size());
  }

  @Test
  void addAndRemoveResearchDomainTest() {
    // Get test orgUnit
    final OrganizationalUnit orgUnit = this.createRemoteOrgUnit(
            "Swiss Research Data Preservation #RESEARCHDOMAIN, Test Method : OrganizationalUnitIT.addAndRemoveResearchDomainTest");
    List<ResearchDomain> list = this.organizationalUnitService.getResearchDomains(orgUnit.getResId());
    final int initialState = list.size();

    // Create an institution
    ResearchDomain researchDomain1 = new ResearchDomain();
    researchDomain1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("researchDomain "));
    researchDomain1.setCode("Z0000");
    researchDomain1.setSource("test1");
    researchDomain1 = this.researchDomainService.create(researchDomain1);

    // Add the researchDomain1 to the OrganizationalUnit
    this.organizationalUnitService.addResearchDomain(orgUnit.getResId(), researchDomain1.getResId());
    list = this.organizationalUnitService.getResearchDomains(orgUnit.getResId());
    assertEquals(initialState + 1, list.size());
    final String researchDomainName = researchDomain1.getName();
    final Optional<ResearchDomain> optionalResearchDomain = list.stream()
            .filter(researchDomain -> researchDomain.getName().equals(researchDomainName))
            .findFirst();
    assertTrue(optionalResearchDomain.isPresent());

    // Test the institution
    assertEquals(researchDomain1.getResId(), optionalResearchDomain.get().getResId());
    assertEquals(researchDomain1.getName(), optionalResearchDomain.get().getName());
    assertEquals(researchDomain1.getCode(), optionalResearchDomain.get().getCode());
    assertEquals(researchDomain1.getSource(), optionalResearchDomain.get().getSource());

    this.organizationalUnitService.removeResearchDomain(orgUnit.getResId(), researchDomain1.getResId());
    list = this.organizationalUnitService.getResearchDomains(orgUnit.getResId());
    assertEquals(initialState, list.size());
  }

  @Test
  void addAndRemovePreservationPolicyTest() {
    // Get test orgUnit
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #PRESERVATION POLICY");
    List<OrganizationalUnitPreservationPolicyDisplay> policies = this.organizationalUnitService
            .getPreservationPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    final PreservationPolicy policy = this.createRemotePreservationPolicy();

    // Add policy to orgUnit
    this.organizationalUnitService.addPreservationPolicy(organizationalUnit.getResId(), policy.getResId());
    policies = this.organizationalUnitService.getPreservationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test add
    for (final OrganizationalUnitPreservationPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getRetention(), policy2.getRetention());
        assertEquals(policy.getDispositionApproval(), policy2.getDispositionApproval());
      }
    }

    // Remove policy
    this.organizationalUnitService.removePreservationPolicy(organizationalUnit.getResId(), policy.getResId());

    // Test remove
    policies = this.organizationalUnitService.getPreservationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState);
  }

  @Test
  void addAndRemoveSubmissionPolicyTest() {
    // Get test orgUnit
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #SUBMISSION POLICY");
    List<OrganizationalUnitSubmissionPolicyDisplay> policies = this.organizationalUnitService
            .getSubmissionPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    final SubmissionPolicy policy = this.createRemoteSubmissionPolicy();

    // Add policy to orgUnit
    this.organizationalUnitService.addSubmissionPolicy(organizationalUnit.getResId(), policy.getResId());
    policies = this.organizationalUnitService.getSubmissionPolicies(organizationalUnit.getResId());

    assertEquals(policies.size(), initialState + 1);

    // Test added person on the OrganizationalUnit
    for (final OrganizationalUnitSubmissionPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getSubmissionApproval(), policy2.getSubmissionApproval());
      }
    }

    // Remove policy
    this.organizationalUnitService.removeSubmissionPolicy(organizationalUnit.getResId(), policy.getResId());

    // Test remove
    policies = this.organizationalUnitService.getSubmissionPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState);
  }

  @Test
  void addPersonToTheOrganizationalUnitTest() {
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #PERSON");
    final Person person = this.createRemotePerson();
    final int peopleBeforeAdding = this.organizationalUnitService.getPeople(organizationalUnit.getResId()).size();

    // Add the person to the OrganizationalUnit with default role
    this.organizationalUnitService.addPerson(organizationalUnit.getResId(), person.getResId());
    final List<Person> people = this.organizationalUnitService.getPeople(organizationalUnit.getResId());

    // If a person is linked to the basic auth account who created the organizational unit
    // then this person is automatically added to the organizational unit.
    // So one or two people should be present in the list
    assertTrue(people.size() == peopleBeforeAdding + 1 || people.size() == peopleBeforeAdding + 2);

    // Test added person on the OrganizationalUnit
    for (final Person person2 : people) {
      if (person2.getResId().equals(person.getResId())) {
        assertNotNull(person2);
        assertEquals(person.getFirstName(), person2.getFirstName());
        assertEquals(person.getLastName(), person2.getLastName());
        assertEquals(person.getOrcid(), person2.getOrcid());
      }
    }
  }

  @Test
  void changeRoleOfAPersonInsideAnOrganizationalUnitTest() {
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #ROLE Person");
    final Person person = this.createRemotePerson();

    // Add the person to the OrganizationalUnit with default role (VISITOR)
    this.organizationalUnitService.addPerson(organizationalUnit.getResId(), person.getResId());
    this.organizationalUnitService.removePerson(organizationalUnit.getResId(), person.getResId());

    // Add the person to the OrganizationalUnit with other roles
    this.organizationalUnitService.addPerson(organizationalUnit.getResId(), person.getResId(), Role.MANAGER_ID);
    final PersonRole personRole = this.organizationalUnitService.getPersonRole(organizationalUnit.getResId());

    if (personRole.getResId().equals(person.getResId())) {
      final List<Role> roleList = personRole.getItems();
      assertTrue(roleList.stream().anyMatch(item -> Role.MANAGER_ID.equalsIgnoreCase(item.getName())));
    }

  }

  @Test
  void closeTest() {
    OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #CLOSE");

    // Close
    final String resId = organizationalUnit.getResId();
    final DateTimeFormatter formatter = DateTimeFormatter.ofPattern(StringTool.DATE_FORMAT);
    final String closingDate = LocalDate.now().plus(1, ChronoUnit.DAYS).format(formatter);

    this.organizationalUnitService.close(resId, closingDate);
    organizationalUnit = this.organizationalUnitService.findOne(resId);

    // Test close
    assertEquals(organizationalUnit.getClosingDate().toString(), closingDate);

    this.organizationalUnitService.delete(organizationalUnit.getResId());
  }

  @Test
  void creationTest() {
    // Create an OrganizationalUnit
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #CREATE");

    final OrganizationalUnit organizationalUnit2 = this.organizationalUnitService.findOne(organizationalUnit.getResId());

    // Test the creation
    assertEquals(organizationalUnit.getName(), organizationalUnit2.getName());
    assertEquals(organizationalUnit.getDescription(), organizationalUnit2.getDescription());
    assertThat(organizationalUnit.getKeywords(), is(organizationalUnit2.getKeywords()));

    this.organizationalUnitService.delete(organizationalUnit.getResId());
  }

  @Test
  void creationWithOpeningClosingDatesTest() {
    // Create an OrganizationalUnit
    OrganizationalUnit organizationalUnit1 = new OrganizationalUnit();
    organizationalUnit1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("DLCM project"));
    this.setRandomUrl(organizationalUnit1);
    organizationalUnit1.setDescription("Swiss Research Data Preservation #CREATE");

    final LocalDate openingDate = LocalDate.now();
    final LocalDate closingDate = openingDate.plusDays(7);

    organizationalUnit1.setOpeningDate(openingDate);
    organizationalUnit1.setClosingDate(closingDate);

    organizationalUnit1 = this.organizationalUnitService.create(organizationalUnit1);
    final OrganizationalUnit organizationalUnit2 = this.organizationalUnitService.findOne(organizationalUnit1.getResId());

    // Test the creation
    assertEquals(organizationalUnit1.getName(), organizationalUnit2.getName(), "Names must match");
    assertEquals(organizationalUnit1.getDescription(), organizationalUnit2.getDescription(), "Descriptions must match");
    this.assertEqualsWithoutNanoSeconds("Creation times must match", organizationalUnit1.getCreationTime(),
            organizationalUnit2.getCreationTime());
    assertEquals(openingDate, organizationalUnit2.getOpeningDate(), "Opening dates must match");
    assertEquals(closingDate, organizationalUnit2.getClosingDate(), "Closing dates must match");

    this.organizationalUnitService.delete(organizationalUnit1.getResId());

  }

  @Test
  void creationWithWrongOpeningClosingDatesTest() {

    // Create an OrganizationalUnit
    OrganizationalUnit organizationalUnit1 = new OrganizationalUnit();
    organizationalUnit1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("DLCM project"));
    this.setRandomUrl(organizationalUnit1);
    organizationalUnit1.setDescription("Swiss Research Data Preservation #CREATE");

    final LocalDate closingDate = LocalDate.now();
    final LocalDate openingDate = closingDate.plusDays(7);
    organizationalUnit1.setOpeningDate(openingDate);
    organizationalUnit1.setClosingDate(closingDate);

    boolean success = true;
    try {
      organizationalUnit1 = this.organizationalUnitService.create(organizationalUnit1);
    } catch (final Exception e) {
      success = false;
    }
    assertFalse(success);

    final OrganizationalUnit organizationalUnit2 = this.organizationalUnitService.findOne(organizationalUnit1.getResId());

    // Test the creation failed
    assertNull(organizationalUnit2);
  }

  @Test
  void defaultLicenseTest() {
    final String orgUnitDefaultLicenseId = "CC-BY-NC-ND-4.0";
    final String orgUnitDefaultLicenseTitle = "Creative Commons Attribution Non Commercial No Derivatives 4.0 International";

    // Test the presence of the system wide default license on a newly created organizational unit
    OrganizationalUnit orgUnit = this.getOrCreatePermanentOrganizationalUnit(OrganizationalUnitStatus.OPEN);
    final String orgUnitId = orgUnit.getResId();

    // Test the setting of a organizational unit specific license
    License newDefaultLicense = new License();
    newDefaultLicense.setResId(orgUnitDefaultLicenseId);
    orgUnit.setDefaultLicense(newDefaultLicense);
    this.organizationalUnitService.update(orgUnitId, orgUnit);
    orgUnit = this.organizationalUnitService.findOne(orgUnitId);
    assertEquals(orgUnitDefaultLicenseId, orgUnit.getDefaultLicense().getResId());
    assertEquals(orgUnitDefaultLicenseTitle, orgUnit.getDefaultLicense().getTitle());

    // Test that setting a garbage license title implies 500 internal error
    OrganizationalUnit orgUnitWithLicenseTitle = new OrganizationalUnit();
    newDefaultLicense.setTitle("KKKKKKKKKK");
    orgUnitWithLicenseTitle.setDefaultLicense(newDefaultLicense);
    assertThrows(HttpServerErrorException.InternalServerError.class, () -> this.organizationalUnitService.update(orgUnitId, orgUnitWithLicenseTitle,
            Arrays.asList("defaultLicense.title")));

    orgUnit = this.organizationalUnitService.findOne(orgUnitId);
    assertEquals(orgUnitDefaultLicenseId, orgUnit.getDefaultLicense().getResId());
    assertEquals(orgUnitDefaultLicenseTitle, orgUnit.getDefaultLicense().getTitle());

    // Try to reset the default license
    orgUnit.setDefaultLicense(null);
    this.organizationalUnitService.update(orgUnitId, orgUnit);
    orgUnit = this.organizationalUnitService.findOne(orgUnitId);
    assertNull(orgUnit.getDefaultLicense());

    // Try to add an non-existing license
    newDefaultLicense = new License();
    newDefaultLicense.setResId("KKKKKKKKKK");
    newDefaultLicense.setTitle("KKKKKKKKKKK");
    orgUnit.setDefaultLicense(newDefaultLicense);
    try {
      this.organizationalUnitService.update(orgUnitId, orgUnit);
      fail("A NOT_FOUND Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.NOT_FOUND, e.getStatusCode());
    }
  }

  @Test
  void deleteTest() {
    // Create a new OrganizationalUnit
    OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation");

    // Delete the organizationalUnit
    final String resId = organizationalUnit.getResId();
    this.organizationalUnitService.delete(resId);
    organizationalUnit = this.organizationalUnitService.findOne(resId);

    // Test deletion
    assertNull(organizationalUnit);
  }

  @Test
  void removePersonFromAnOrganizationalUnitTest() {
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnit("Swiss Research Data Preservation #REMOVE PERSON");
    final Person person = this.createRemotePerson();

    // Add the person to the OrganizationalUnit with default role
    this.organizationalUnitService.addPerson(organizationalUnit.getResId(), person.getResId());
    List<Person> people = this.organizationalUnitService.getPeople(organizationalUnit.getResId());

    final int nbPepole = people.size();
    this.organizationalUnitService.removePerson(organizationalUnit.getResId(), person.getResId());
    people = this.organizationalUnitService.getPeople(organizationalUnit.getResId());
    assertEquals(nbPepole - 1, people.size());
  }

  @Test
  void unicityTest() {
    // Creating a new Organizational Unit
    OrganizationalUnit organizationalUnit1 = this.createRemoteOrgUnit("Swiss Research Data Preservation #CREATE");
    organizationalUnit1 = this.organizationalUnitService.findOne(organizationalUnit1.getResId());

    // Creating a second Organizational unit with the same name as the first's
    OrganizationalUnit organizationalUnit2 = new OrganizationalUnit();
    organizationalUnit2.setName(organizationalUnit1.getName());
    this.setRandomUrl(organizationalUnit2);
    organizationalUnit2.setDescription("Dufourian Research Data Preservation #CREATE");

    // Testing the unicity of name/**/
    try {
      this.organizationalUnitService.create(organizationalUnit2);
      fail("A BAD_REQUEST Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode(), "Unicity constraint not satisfied");
    }
  }

  @Test
  void updateTest() {
    // Create an OrganizationalUnit
    final OrganizationalUnit organizationalUnit1 = this.createRemoteOrgUnit("Swiss Research Data Preservation #UPDATE");

    // Update the person
    final OrganizationalUnit organizationalUnit2 = new OrganizationalUnit();
    organizationalUnit2.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("DLCM project 2.0"));
    this.setRandomUrl(organizationalUnit2);
    organizationalUnit2.setDescription("Swiss Research Data Preservation (v2) #UPDATE");

    // Update the keywords
    organizationalUnit2.setKeywords(Arrays.asList("astrochemistry ", "immunohistochemistry ", "phytochemistry"));
    final OrganizationalUnit organizationalUnit3 = this.organizationalUnitService.update(organizationalUnit1.getResId(), organizationalUnit2);

    // Test the update
    assertEquals(organizationalUnit1.getResId(), organizationalUnit3.getResId());
    assertEquals(organizationalUnit2.getDescription(), organizationalUnit3.getDescription());
    assertThat(organizationalUnit3.getKeywords(), is(organizationalUnit2.getKeywords()));

  }

  @Test
  void uploadLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    /*
     * Create an org unit
     */
    final OrganizationalUnit organizationalUnit1 = this.createRemoteOrgUnit("Swiss Research Data Preservation #LOGO");

    /*
     * Add logo
     */
    OrganizationalUnit fetchedOrgUnit = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);

    /*
     * Check logo has been uploaded
     */
    OrganizationalUnit refetchedOrgUnit = this.organizationalUnitService.findOne(fetchedOrgUnit.getResId());
    assertEquals(refetchedOrgUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrgUnit.getImage().getFileSize(), sizeLogo);

    /*
     * Update org unit again
     */
    String newName = DLCMTestConstants.getRandomNameWithTemporaryLabel("DLCM project");
    refetchedOrgUnit.setName(newName);
    refetchedOrgUnit = this.organizationalUnitService.update(refetchedOrgUnit.getResId(), refetchedOrgUnit);
    assertEquals(refetchedOrgUnit.getName(), newName);

  }

  @Test
  void downloadLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    /*
     * Create an org unit
     */
    final OrganizationalUnit organizationalUnit1 = this.createRemoteOrgUnit("Swiss Research Data Preservation #LOGO");

    /*
     * Add logo
     */
    OrganizationalUnit organizationalUnit2 = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);
    /*
     * Check logo has been uploaded
     */
    final OrganizationalUnit refetchedOrgUnit = this.organizationalUnitService.findOne(organizationalUnit2.getResId());
    assertEquals(refetchedOrgUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrgUnit.getImage().getFileSize(), sizeLogo);

    Path path = Paths.get(System.getProperty("java.io.tmpdir"), "download.tmp");
    this.organizationalUnitService.downloadLogo(refetchedOrgUnit.getResId(), path);
    assertNotNull(path.toFile());
    assertEquals(Long.valueOf(path.toFile().length()), refetchedOrgUnit.getImage().getFileSize());
  }

  @Test
  void deleteLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    /*
     * Create an org unit
     */
    final OrganizationalUnit organizationalUnit1 = this.createRemoteOrgUnit("Swiss Research Data Preservation #LOGO");

    /*
     * Add logo
     */
    OrganizationalUnit fetchedOrgUnit = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);

    /*
     * Check logo has been uploaded
     */
    OrganizationalUnit refetchedOrgUnit = this.organizationalUnitService.findOne(fetchedOrgUnit.getResId());
    assertEquals(refetchedOrgUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrgUnit.getImage().getFileSize(), sizeLogo);

    this.organizationalUnitService.deleteLogo(organizationalUnit1.getResId());
    refetchedOrgUnit = this.organizationalUnitService.findOne(fetchedOrgUnit.getResId());
    assertNull(refetchedOrgUnit.getImage());
  }

  /*************************************************/

  @Override
  protected void deleteFixtures() {
    this.clearOrganizationalUnitFixtures();
    this.clearFundingAgencyFixtures();
    this.clearInstitutionsFixtures();
    this.clearPeopleFixtures();
  }

  private DisseminationPolicy createRemoteDisseminationPolicy() {
    final DisseminationPolicy policy = new DisseminationPolicy();
    policy.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("dissemination policy"));
    policy.setType(DisseminationPolicy.DisseminationPolicyType.WEB);
    policy.setDestinationServer("http://server.com");
    return this.disseminationPolicyService.create(policy);
  }

  private OrganizationalUnit createRemoteOrgUnit(String description) {
    // Create a new OrganizationalUnit
    OrganizationalUnit organizationalUnit = new OrganizationalUnit();
    organizationalUnit.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("DLCM project"));
    this.setRandomUrl(organizationalUnit);
    organizationalUnit.setDescription(description);
    organizationalUnit.setKeywords(Arrays.asList("magnetostatics", "photonics", "oceanography"));
    organizationalUnit = this.organizationalUnitService.create(organizationalUnit);
    return organizationalUnit;
  }

  private Person createRemotePerson() {
    final Person person = new Person();
    person.setFirstName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Any First name"));
    person.setLastName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Any Last name"));
    person.setOrcid(DLCMTestConstants.getRandomOrcId());
    return this.personService.create(person);
  }

  private PreservationPolicy createRemotePreservationPolicy() {
    final PreservationPolicy policy = new PreservationPolicy();
    policy.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Keep It For xx Years"));
    policy.setRetention(10 * DLCMConstants.DAYS_BY_YEAR);
    return this.preservationPolicyService.create(policy);
  }

  private SubmissionPolicy createRemoteSubmissionPolicy() {
    final SubmissionPolicy policy = new SubmissionPolicy();
    policy.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("submission policy "));
    policy.setSubmissionApproval(true);
    return this.submissionPolicyService.create(policy);
  }
}
