/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - OrganizationalUnitManagerIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.List;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;

import org.junit.jupiter.api.Test;

import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.FundingAgency;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;

public class OrganizationalUnitManagerIT extends AbstractAdminIT {

  @Override
  protected void setUser() {
    this.restClientTool.sudoUser();
  }

  @Test
  public void creationOrgUnitAndAddInstitutionsTest() {
    /*
     * Get test orgUnit
     */
    final OrganizationalUnit organizationalUnit = this.getOrCreatePermanentOrganizationalUnit(DLCMTestConstants.OrganizationalUnitStatus.OPEN,
            Role.MANAGER_ID);
    List<Institution> list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    final int initialState = list.size();

    /*
     * Create a new Institution
     */
    this.restClientTool.sudoAdmin();
    Institution institution = new Institution();
    institution.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution"));
    institution.setDescription("description" + ThreadLocalRandom.current().nextInt());
    institution = this.institutionService.create(institution);
    this.restClientTool.exitSudo();

    /*
     * Add the Institution to the OrganizationalUnit
     */
    this.organizationalUnitService.addInstitution(organizationalUnit.getResId(), institution.getResId());
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertEquals(initialState + 1, list.size());
    final String institutionName = institution.getName();
    final Optional<Institution> optionalInstitution = list.stream().filter(ins -> ins.getName().equals(institutionName))
            .findFirst();
    assertTrue(optionalInstitution.isPresent());

    /*
     * Check it has been added
     */
    assertEquals(institution.getResId(), optionalInstitution.get().getResId());
    assertEquals(institution.getName(), optionalInstitution.get().getName());
    assertEquals(institution.getDescription(), optionalInstitution.get().getDescription());

    /*
     * Remove the Institution
     */
    this.organizationalUnitService.removeInstitution(organizationalUnit.getResId(), institution.getResId());
    /*
     * Check it has been removed
     */
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertEquals(initialState, list.size());

    /*
     * Create 3 Institutions and link them to the OrganizationalUnit
     */
    Institution institution1 = new Institution();
    institution1.setDescription(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 1"));
    institution1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 1"));

    Institution institution2 = new Institution();
    institution2.setDescription(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 2"));
    institution2.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 2"));

    Institution institution3 = new Institution();
    institution3.setDescription(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 3"));
    institution3.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 3"));

    final Institution[] institutions = new Institution[] { institution1, institution2, institution3 };

    String institutionsToDelete = null;

    int i = 0;
    for (Institution ins : institutions) {
      this.restClientTool.sudoAdmin();
      ins = this.institutionService.create(ins);
      this.restClientTool.exitSudo();
      this.organizationalUnitService.addInstitution(organizationalUnit.getResId(), ins.getResId());

      if (i == 1) {
        institutionsToDelete = ins.getResId();
      }

      i++;
    }

    /*
     * Check they are all linked correctly
     */
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertNotNull(list);
    assertEquals(initialState + 3, list.size());

    /*
     * Delete one specifically and check it is not in the list anymore
     */
    this.organizationalUnitService.removeInstitution(organizationalUnit.getResId(), institutionsToDelete);
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertNotNull(list);
    assertEquals(initialState + 2, list.size());
    for (final Institution fetched_institution : list) {
      assertNotEquals(institutionsToDelete, fetched_institution.getResId());
    }
  }

  @Test
  void addAndRemoveFundingAgency() {
    /*
     * Get test orgUnit
     */
    final OrganizationalUnit organizationalUnit = this.getOrCreatePermanentOrganizationalUnit(DLCMTestConstants.OrganizationalUnitStatus.OPEN,
            Role.MANAGER_ID);
    List<FundingAgency> fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    final int initialState = fundingAgencies.size();

    /*
     * Create a new FundingAgency
     */
    this.restClientTool.sudoAdmin();
    final String fundingAgencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String fundingAgencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(fundingAgencyAcronym);
    new_agency.setName(fundingAgencyName);
    new_agency = this.fundingAgencyService.create(new_agency);
    this.restClientTool.exitSudo();

    /*
     * Add the FundingAgency to the OrganizationalUnit
     */
    this.organizationalUnitService.addFundingAgency(organizationalUnit.getResId(), new_agency.getResId());

    /*
     * Check it has been added
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 1, fundingAgencies.size());
    final Optional<FundingAgency> optionalFA = fundingAgencies.stream().filter(fa -> fa.getName().equals(fundingAgencyName)).findFirst();
    assertTrue(optionalFA.isPresent());
    assertEquals(fundingAgencyName, optionalFA.get().getName());

    /*
     * Remove the FundingAgency
     */
    this.organizationalUnitService.removeFundingAgency(organizationalUnit.getResId(), new_agency.getResId());

    /*
     * Check it has been removed
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState, fundingAgencies.size());

    /*
     * Create 3 FundingAgencies and link them to the OrganizationalUnit
     */
    final FundingAgency subFA1 = new FundingAgency();
    subFA1.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 1"));
    subFA1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 1"));

    final FundingAgency subFA2 = new FundingAgency();
    subFA2.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 2"));
    subFA2.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 2"));

    final FundingAgency subFA3 = new FundingAgency();
    subFA3.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 3"));
    subFA3.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 3"));

    final FundingAgency[] subAgencies = new FundingAgency[] { subFA1, subFA2, subFA3 };

    String agencyIdToDelete = null;

    int i = 0;
    for (FundingAgency subAgency : subAgencies) {
      this.restClientTool.sudoAdmin();
      subAgency = this.fundingAgencyService.create(subAgency);
      this.restClientTool.exitSudo();
      this.organizationalUnitService.addFundingAgency(organizationalUnit.getResId(), subAgency.getResId());

      if (i == 1) {
        agencyIdToDelete = subAgency.getResId();
      }

      i++;
    }

    /*
     * Check they are all linked correctly
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 3, fundingAgencies.size());

    /*
     * Delete one specifically and check it is not in the list anymore
     */
    this.organizationalUnitService.removeFundingAgency(organizationalUnit.getResId(), agencyIdToDelete);
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 2, fundingAgencies.size());
    for (final FundingAgency fetched_agency : fundingAgencies) {
      assertNotEquals(agencyIdToDelete, fetched_agency.getResId());
    }

  }

  @Override
  protected void deleteFixtures() {
    this.restClientTool.sudoAdmin();
    this.clearOrganizationalUnitFixtures();
    this.clearFundingAgencyFixtures();
    this.clearInstitutionsFixtures();
    this.restClientTool.exitSudo();
  }

}
