/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - AbstractPreingestDemoIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.demo;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.time.OffsetDateTime;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.web.client.HttpClientErrorException;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.util.SolidifyTime;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.model.Access;
import ch.dlcm.model.ComplianceLevel;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.DataTag;
import ch.dlcm.model.EmbargoInfo;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.dto.FileUploadDto;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.oais.AipDataFile;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.RepresentationInfo;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.preingest.DepositDataFile;
import ch.dlcm.model.settings.License;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.service.archivalstorage.AIPClientService;
import ch.dlcm.service.ingest.SIPClientService;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.preingest.AbstractPreingestIT;

public abstract class AbstractPreingestDemoIT extends AbstractPreingestIT {

  private static final Logger log = LoggerFactory.getLogger(AbstractPreingestDemoIT.class);

  protected final String BODMER_DEPOSIT_FOLDER = "deposit/bodmer";
  protected final String DLCM_DEPOSIT_FOLDER = "deposit/dlcm";
  protected final String UPDATED_BY = " => ";

  @Autowired
  private SIPClientService sipService;

  @Override
  protected void setUser() {
    this.restClientTool.sudoRoot();
  }

  @Override
  protected void deleteFixtures() {
    // do nothing
  }

  protected Deposit submitAndApproveDeposit(Deposit deposit) {
    this.submitDeposit(deposit);
    SolidifyTime.waitOneSecond();
    return this.approveDeposit(deposit);
  }

  protected String getVersionLabel(DLCMMetadataVersion version) {
    return " (" + version.getVersion() + ")";
  }

  protected OrganizationalUnit getOrgUnit(String orgUnitName) {
    List<OrganizationalUnit> orgUnits = this.organizationalUnitService.findAll();
    Optional<OrganizationalUnit> orgUnit = orgUnits.stream().filter(ou -> ou.getName().equals(orgUnitName)).findFirst();
    if (orgUnit.isPresent()) {
      return orgUnit.get();
    }
    return null;
  }

  protected int getRandomUpdate() {
    return this.getRandomNumberUsingInts(0, UpdatableField.values().length);
  }

  protected boolean getRandomBoolean() {
    Random random = new Random();
    return random.nextBoolean();
  }

  private int getRandomNumberUsingInts(int min, int max) {
    Random random = new Random();
    return random.ints(min, max)
            .findFirst()
            .getAsInt();
  }

  protected String getFilename(Resource file) {
    return file.getFilename().substring(0, file.getFilename().lastIndexOf("."));
  }

  protected void addThumbnail(Deposit deposit, Resource file) {
    FileUploadDto<DepositDataFile> fileInfo = new FileUploadDto<>();
    fileInfo.setDataCategory(DataCategory.Internal);
    fileInfo.setDataType(DataCategory.DatasetThumbnail);
    this.depositService.uploadDepositDataFile(deposit.getResId(), file, fileInfo);
  }

  protected boolean checkIfThumbnailSupported(DLCMMetadataVersion version) {
    if (version == DLCMMetadataVersion.V2_1 || version == DLCMMetadataVersion.V3_0) {
      return true;
    }
    return false;
  }

  protected boolean checkCompatibleVersion(DLCMMetadataVersion version1, DLCMMetadataVersion version2) {
    switch (version1) {
      case V1_0:
        if (version2 == DLCMMetadataVersion.V1_0) {
          return true;
        }
        break;
      case V1_1:
        if (version2 == DLCMMetadataVersion.V1_0
                || version2 == DLCMMetadataVersion.V1_1) {
          return true;
        }
        break;
      case V2_0:
        if (version2 == DLCMMetadataVersion.V1_0
                || version2 == DLCMMetadataVersion.V1_1
                || version2 == DLCMMetadataVersion.V2_0) {
          return true;
        }
        break;
      case V2_1:
        if (version2 != DLCMMetadataVersion.V3_0) {
          return true;
        }
        break;
      case V3_0:
        return true;
      default:
        break;
    }
    return false;
  }

  protected Access getNewAccessLevelToUpdate(Access access) {
    switch (access) {
      case PUBLIC:
        return Access.RESTRICTED;
      case RESTRICTED:
        return Access.CLOSED;
      case CLOSED:
      default:
        return Access.PUBLIC;
    }
  }

  protected String getNewLicenseToUpdate(String licenseId) {
    for (License license : this.licenseService.findAll()) {
      if (license.getResId().equals(licenseId)) {
        continue;
      }
      return license.getResId();
    }
    return null;
  }

  protected EmbargoInfo getNewEmbargoToUpdate(EmbargoInfo embargo, Access accessLevel) {
    if (embargo == null) {
      embargo = new EmbargoInfo();
      embargo.setAccess(this.getNewAccessLevelToUpdate(accessLevel));
      embargo.setMonths(6);
    } else {
      embargo.setAccess(this.getNewAccessLevelToUpdate(embargo.getAccess()));
      embargo.setMonths(embargo.getMonths() + 1);
      embargo.setStartDate(null);
    }
    if (this.getRandomBoolean() || embargo.getAccess() == Access.PUBLIC) {
      return null;
    }
    return embargo;
  }

  protected DataTag getNewDataTagToUpdate(DataTag dataSensitivity, Access accessLevel) {
    DataTag dataTag = null;
    switch (dataSensitivity) {
      case UNDEFINED:
        dataTag = DataTag.BLUE;
        break;
      case BLUE:
        dataTag = DataTag.GREEN;
        break;
      case GREEN:
        dataTag = DataTag.YELLOW;
        break;
      case YELLOW:
        dataTag = DataTag.ORANGE;
        break;
      case ORANGE:
        dataTag = DataTag.RED;
        break;
      case RED:
        dataTag = DataTag.CRIMSON;
        break;
      case CRIMSON:
        dataTag = DataTag.UNDEFINED;
        break;
    }
    if (dataTag.isAccessLevelAllowed(accessLevel)) {
      return dataTag;
    }
    return this.getNewDataTagToUpdate(dataTag, accessLevel);
  }

  private List<String> getNewKeywords(List<String> keywords) {
    if (keywords.isEmpty() || this.getRandomBoolean()) {
      // Add keyword
      keywords.add("keyword" + keywords.size() + 1);
    } else {
      // Remove keyword
      keywords.remove(keywords.size() - 1);
    }
    return keywords;
  }

  protected void updateDeposit(
          OrganizationalUnit orgUnit,
          DLCMMetadataVersion version,
          String title,
          UpdatableField updateField,
          boolean cancelUpdate) {
    // Find deposit
    Deposit originalDeposit = this.findDeposit(orgUnit, version, title);
    assertNotNull(originalDeposit, "Cannot find deposit: [" + version.getVersion() + "] " + title);
    // Start metadata edition
    log.info("Updating deposit {} [{}]", originalDeposit.getTitle(), originalDeposit.getResId());
    Deposit workingDeposit = this.startEditingMetadata(originalDeposit);
    // Update deposit
    String changedValues;
    workingDeposit.setTitle(this.getNewArchiveToUpdateTitle(workingDeposit.getTitle()));
    switch (updateField) {
      case ACCESS_LEVEL:
        changedValues = workingDeposit.getAccess().toString() + UPDATED_BY;
        workingDeposit.setAccess(this.getNewAccessLevelToUpdate(workingDeposit.getAccess()));
        changedValues += workingDeposit.getAccess().toString();
        workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate of access level " + changedValues);
        break;
      case DATA_TAG:
        changedValues = workingDeposit.getDataSensitivity().toString() + UPDATED_BY;
        workingDeposit.setDataSensitivity(this.getNewDataTagToUpdate(workingDeposit.getDataSensitivity(), workingDeposit.getAccess()));
        changedValues += workingDeposit.getDataSensitivity().toString();
        workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate of data tag " + changedValues);
        break;
      case DESCRIPTION:
        workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate of description");
        break;
      case EMBARGO:
        changedValues = this.embargoToString(workingDeposit.getEmbargo()) + UPDATED_BY;
        workingDeposit.setEmbargo(this.getNewEmbargoToUpdate(workingDeposit.getEmbargo(), workingDeposit.getAccess()));
        changedValues += this.embargoToString(workingDeposit.getEmbargo());
        workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate of embargo " + changedValues);
        break;
      case LICENSE:
        changedValues = workingDeposit.getLicenseId() + UPDATED_BY;
        workingDeposit.setLicenseId(this.getNewLicenseToUpdate(workingDeposit.getLicenseId()));
        changedValues += workingDeposit.getLicenseId();
        workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate of license " + changedValues);
        break;
      case KEYWORDS:
        changedValues = workingDeposit.getKeywords().toString() + UPDATED_BY;
        workingDeposit.setKeywords(this.getNewKeywords(workingDeposit.getKeywords()));
        changedValues += workingDeposit.getKeywords().toString();
        workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate of keywords " +
                changedValues);
        break;
      case TITLE:
      default:
        break;
    }
    // Save & submit deposit
    this.saveAndSubmitDeposit(originalDeposit, workingDeposit, cancelUpdate);
  }

  private String embargoToString(EmbargoInfo embargo) {
    if (embargo == null) {
      return "null";
    }
    return "[access=" + embargo.getAccess().toString()
            + " month=" + embargo.getMonths().toString()
            + " startDate=" + embargo.getStartDate()
            + "]";
  }

  protected void updateDepositContributors(OrganizationalUnit orgUnit, DLCMMetadataVersion version, String title) {
    // Find deposit
    Deposit originalDeposit = this.findDeposit(orgUnit, version, title);
    assertNotNull(originalDeposit, "Cannot find deposit: [" + version.getVersion() + "] " + title);
    // Start metadata edition
    log.info("Updating deposit {} [{}]", originalDeposit.getTitle(), originalDeposit.getResId());
    Deposit workingDeposit = this.startEditingMetadata(originalDeposit);
    // Update deposit title
    workingDeposit.setTitle(this.getNewArchiveToUpdateTitle(workingDeposit.getTitle()));
    // Deposit contributors
    List<Person> contributors = this.depositService.getContributors(originalDeposit.getResId());
    switch (contributors.size()) {
      case 1:
        // Add contributor
        for (String orcid : DLCMDemoConstants.ORCID_LIST) {
          if (this.getRandomBoolean()
                  && !contributors.get(0).getOrcid().equals(orcid)) {
            this.addContributor(originalDeposit, orcid);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nAdd contributor " + orcid);
            break;
          }
        }
        break;
      case 2:
        // Remove contributor
        if (this.getRandomBoolean()) {
          this.removeContributor(workingDeposit, contributors.get(1).getResId());
          workingDeposit.setDescription(workingDeposit.getDescription() + "\nRemove contributor " + contributors.get(1).getOrcid());
        } else {
          // Add contributor
          for (String orcid : DLCMDemoConstants.ORCID_LIST) {
            if (this.getRandomBoolean()
                    && !contributors.get(0).getOrcid().equals(orcid)
                    && !contributors.get(1).getOrcid().equals(orcid)) {
              this.addContributor(originalDeposit, orcid);
              workingDeposit.setDescription(workingDeposit.getDescription() + "\nAdd contributor " + orcid);
              break;
            }
          }
        }
        break;
      case 3:
        // Remove contributor
        if (this.getRandomBoolean()) {
          this.removeContributor(workingDeposit, contributors.get(2).getResId());
          workingDeposit.setDescription(workingDeposit.getDescription() + "\nRemove contributor " + contributors.get(2).getOrcid());
        }
        break;
      default:
        break;
    }
    List<Person> newContributors = this.depositService.getContributors(originalDeposit.getResId());
    // Save & submit deposit
    boolean cancelUpdate = contributors.size() == newContributors.size();
    this.saveAndSubmitDeposit(originalDeposit, workingDeposit, cancelUpdate);
  }

  protected void updateDepositDoi(OrganizationalUnit orgUnit, DLCMMetadataVersion version, String title) {
    boolean cancelUpdate = false;
    // Find deposit
    Deposit originalDeposit = this.findDeposit(orgUnit, version, title);
    assertNotNull(originalDeposit, "Cannot find deposit: [" + version.getVersion() + "] " + title);
    // Start metadata edition
    log.info("Updating deposit {} [{}]", originalDeposit.getTitle(), originalDeposit.getResId());
    Deposit workingDeposit = this.startEditingMetadata(originalDeposit);
    // Update deposit title
    workingDeposit.setTitle(this.getNewArchiveToUpdateTitle(workingDeposit.getTitle()));
    String newDoi = DLCMDemoConstants.DOI_LIST[this.getRandomNumberUsingInts(0, DLCMDemoConstants.DOI_LIST.length)];
    String oldDoi = null;
    // Deposit contributors
    switch (this.getRandomNumberUsingInts(0, 2)) {
      case 0:
        // 'isIdenticalTo' DOI
        oldDoi = workingDeposit.getIsIdenticalTo();
        if (this.getRandomBoolean()) {
          if (oldDoi == null) {
            workingDeposit.setIsIdenticalTo(newDoi);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate 'IdenticalTo' DOI " + newDoi);
          } else {
            workingDeposit.setIsIdenticalTo(null);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nRemove 'IdenticalTo' DOI");
          }
        }
        cancelUpdate = (workingDeposit.getIsIdenticalTo() == null && oldDoi == null)
                || (workingDeposit.getIsIdenticalTo() != null && workingDeposit.getIsIdenticalTo().equals(oldDoi));
        break;
      case 1:
        // 'isReferencedBy' DOI
        oldDoi = workingDeposit.getIsReferencedBy();
        if (this.getRandomBoolean()) {
          if (oldDoi == null) {
            workingDeposit.setIsReferencedBy(newDoi);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate 'ReferencedBy' DOI " + newDoi);
          } else {
            workingDeposit.setIsReferencedBy(null);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nRemove 'ReferencedBy' DOI");
          }
        }
        cancelUpdate = (workingDeposit.getIsReferencedBy() == null && oldDoi == null)
                || (workingDeposit.getIsReferencedBy() != null && workingDeposit.getIsReferencedBy().equals(oldDoi));
        break;
      case 2:
        // 'isObsoletedBy' DOI
        oldDoi = workingDeposit.getIsObsoletedBy();
        if (this.getRandomBoolean()
                && version != DLCMMetadataVersion.V1_0
                && version != DLCMMetadataVersion.V1_1) {
          if (oldDoi == null) {
            workingDeposit.setIsObsoletedBy(newDoi);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nUpdate 'ObsoletedBy' DOI " + newDoi);
          } else {
            workingDeposit.setIsObsoletedBy(null);
            workingDeposit.setDescription(workingDeposit.getDescription() + "\nRemove 'ObsoletedBy' DOI");
          }
        }
        cancelUpdate = (workingDeposit.getIsObsoletedBy() == null && oldDoi == null)
                || (workingDeposit.getIsObsoletedBy() != null && workingDeposit.getIsObsoletedBy().equals(oldDoi));
        break;
      default:
        break;
    }
    // Save & submit deposit
    this.saveAndSubmitDeposit(originalDeposit, workingDeposit, cancelUpdate);
  }

  protected List<String> getDepositList() {
    return Arrays.asList(DLCMDemoConstants.BODMER_DEPOSIT1, DLCMDemoConstants.BODMER_DEPOSIT2);
  }

  protected void addAipCollection(OrganizationalUnit orgUnit, DLCMMetadataVersion version, String collectionTitle, String depositTitle,
          boolean cancelUpdate) {
    // Find collection
    Deposit collection = this.findDeposit(orgUnit, version, collectionTitle);
    assertNotNull(collection, "Cannot find deposit: [" + version.getVersion() + "] " + collectionTitle);
    // Start metadata edition
    Deposit workingcollection = this.startEditingMetadata(collection);
    // Update collection
    for (final ArchivalInfoPackage aip : this.aipService.searchByProperties(Map.of("info.organizationalUnitId", orgUnit.getResId(),
            "info.name", depositTitle))) {
      // Add AIP
      if (!aip.isCollection() && this.checkCompatibleVersion(version, aip.getMetadataVersion())) {
        this.depositService.addAip(workingcollection.getResId(), aip.getResId());
        break;
      }
    }
    // Save & submit deposit
    this.saveAndSubmitDeposit(collection, workingcollection, cancelUpdate);
  }

  protected void removeAipCollection(OrganizationalUnit orgUnit, DLCMMetadataVersion version, String collectionTitle, boolean cancelUpdate) {
    // Find collection
    Deposit collection = this.findDeposit(orgUnit, version, collectionTitle);
    assertNotNull(collection, "Cannot find deposit: [" + version.getVersion() + "] " + collectionTitle);
    // Start metadata edition
    Deposit workingcollection = this.startEditingMetadata(collection);
    // Update collection
    for (final ArchivalInfoPackage aip : this.depositService.getDepositAip(workingcollection.getResId())) {
      // Add AIP
      if (!aip.isCollection() && this.checkCompatibleVersion(version, aip.getMetadataVersion())) {
        this.depositService.removeAip(workingcollection.getResId(), aip.getResId());
        break;
      }
    }
    // Save & submit deposit
    this.saveAndSubmitDeposit(collection, workingcollection, cancelUpdate);
  }

  protected boolean waitAIPIsProcessed(String depositId) {
    int count = 0;
    Deposit deposit = this.depositService.findOne(depositId);
    SubmissionInfoPackage sip = this.sipService.findOne(deposit.getSipId());
    ArchivalInfoPackage aip;

    do {
      aip = this.aipService.findOne(sip.getAipId());
      DLCMTestConstants
              .failAfterNAttempt(count,
                      "AIP " + sip.getAipId() + " should have status Completed or InError instead of " + aip.getInfo().getStatus());
      count++;
    } while (aip.getInfo().getStatus() != PackageStatus.COMPLETED && aip.getInfo().getStatus() != PackageStatus.IN_ERROR);
    return (PackageStatus.COMPLETED == aip.getInfo().getStatus());
  }

  protected void saveAndSubmitDeposit(Deposit originalDeposit, Deposit workingDeposit, boolean cancelUpdate) {
    Deposit updatedDeposit = null;
    try {
      updatedDeposit = this.depositService.update(workingDeposit.getResId(), workingDeposit);
      assertEquals(workingDeposit.getTitle(), updatedDeposit.getTitle(), "Deposit title not updated");
      assertEquals(workingDeposit.getAccess(), updatedDeposit.getAccess(), "Deposit access level not updated");
      assertEquals(workingDeposit.getDataSensitivity(), updatedDeposit.getDataSensitivity(), "Deposit data tag not updated");
    } catch (HttpClientErrorException.BadRequest e) {
      log.error(e.getResponseBodyAsString());
      cancelUpdate = true;
    }
    if (cancelUpdate) {
      log.info("Cancelling edition of deposit {} [{}]", originalDeposit.getTitle(), originalDeposit.getResId());
      this.depositService.cancelEditingMetadata(originalDeposit.getResId());
    } else {
      log.info("Submitting deposit {} [{}]", originalDeposit.getTitle(), originalDeposit.getResId());
      this.submitAndApproveDeposit(updatedDeposit);
    }
    assertTrue(this.waitDepositIsCompleted(workingDeposit.getResId()), "Deposit not completed: " + workingDeposit.getResId());
    Deposit latestDeposit = this.depositService.findOne(workingDeposit.getResId());
    if (cancelUpdate) {
      checkInfo(originalDeposit.getResId(), "Cancel => ", this.getDepositInfo(latestDeposit), this.getDepositInfo(originalDeposit), true);
    } else {
      checkInfo(originalDeposit.getResId(), "Update => ", this.getDepositInfo(latestDeposit), this.getDepositInfo(updatedDeposit), true);
    }
    assertTrue(this.waitAIPIsProcessed(latestDeposit.getResId()), "AIP not completed for deposit " + latestDeposit.getResId());
    this.verifyDeposit(latestDeposit);
  }

  protected Deposit findDeposit(OrganizationalUnit orgUnit, DLCMMetadataVersion version, String title) {
    String depositTitle = title + this.getVersionLabel(version);
    List<Deposit> deposits = this.depositService
            .searchByProperties(Map.of("title", depositTitle, DLCMConstants.ORG_UNIT_ID_FIELD, orgUnit.getResId()));
    assertFalse(deposits.isEmpty(), "Cannot find deposit: " + depositTitle);
    return deposits.get(0);
  }

  protected Resource[] listFiles(String folder, String filePattern) throws IOException {
    final ResourcePatternResolver resolver = new PathMatchingResourcePatternResolver(this.getClass().getClassLoader());
    return resolver.getResources("classpath*:/" + folder + "/" + filePattern);
  }

  protected void createCollection(
          OrganizationalUnit orgUnit,
          DLCMMetadataVersion version,
          String collectionName,
          SubmissionPolicy submissionPolicy,
          PreservationPolicy preservationPolicy,
          String depositName,
          boolean includeUnit,
          boolean includeCollection) throws IOException {
    log.info("Creating new collections {} for {}", depositName, version.getVersion());
    Access collectionAccessLevel = Access.PUBLIC;
    DataTag collectionDataTag = DataTag.BLUE;
    Deposit deposit = this.createDeposit(orgUnit, version, collectionName, collectionName, submissionPolicy, preservationPolicy);
    // Add contributors
    this.addContributors(deposit);
    for (final ArchivalInfoPackage aip : this.aipService.searchByProperties(Map.of("info.organizationalUnitId", orgUnit.getResId(),
            "info.name", depositName))) {
      if (aip.getInfo().getStatus() != PackageStatus.COMPLETED
              || (!aip.isCollection() && !includeUnit)
              || (aip.isCollection() && !includeCollection)
              || ((submissionPolicy != null || preservationPolicy != null) && !aip.getInfo().getName().contains(DLCMDemoConstants.FOR_DEMO))) {
        continue;
      }
      // Add AIP
      if (this.checkCompatibleVersion(version, aip.getMetadataVersion())) {
        this.depositService.addAip(deposit.getResId(), aip.getResId());
        if (aip.getInfo().getAccess().value() > collectionAccessLevel.value()) {
          collectionAccessLevel = aip.getInfo().getAccess();
        }
        if (aip.getInfo().getDataSensitivity().value() > collectionDataTag.value()) {
          collectionDataTag = aip.getInfo().getDataSensitivity();
        }
      }
    }
    // Add thumbnail
    if (this.checkIfThumbnailSupported(version)) {
      this.addThumbnail(deposit, this.listFiles(BODMER_DEPOSIT_FOLDER, "thumbnail/Faust.timg")[0]);
    }
    // Update collection
    deposit.setAccess(collectionAccessLevel);
    deposit.setDataSensitivity(collectionDataTag);
    deposit = this.depositService.update(deposit.getResId(), deposit);
    // Check if there are archives
    List<ArchivalInfoPackage> aipList = this.depositService.getDepositAip(deposit.getResId());
    assertFalse(aipList.isEmpty(), "No archive in deposit: " + deposit.getResId());
    // Submit deposit
    this.waitDepositWithDataIsReady(deposit.getResId());
    this.submitAndApproveDeposit(deposit);
    assertTrue(this.waitDepositIsCompleted(deposit.getResId()), "Deposit not completed: " + deposit.getResId());
  }

  protected void addContributors(Deposit deposit) {
    for (String orcid : DLCMDemoConstants.ORCID_LIST) {
      this.addContributor(deposit, orcid);
    }

  }

  private void addContributor(Deposit deposit, String orcid) {
    Person person = this.personService.findByOrcid(orcid);
    this.depositService.addContributor(deposit.getResId(), person.getResId());
  }

  private void removeContributor(Deposit deposit, String personId) {
    this.depositService.removeContributor(deposit.getResId(), personId);
  }

  protected Deposit createDeposit(
          OrganizationalUnit orgUnit,
          DLCMMetadataVersion version,
          String title,
          String description,
          SubmissionPolicy submissionPolicy,
          PreservationPolicy preservationPolicy) {
    String label = this.getVersionLabel(version);
    if (submissionPolicy != null || preservationPolicy != null) {
      label += DLCMDemoConstants.FOR_DEMO;
    }
    final Deposit deposit = new Deposit();
    deposit.init();
    deposit.setMetadataVersion(version);
    deposit.setTitle(title + label);
    deposit.setDescription(description + label);
    deposit.setStatus(DepositStatus.IN_PROGRESS);
    deposit.setOrganizationalUnitId(orgUnit.getResId());
    if (submissionPolicy == null) {
      deposit.setSubmissionPolicyId(orgUnit.getDefaultSubmissionPolicy().getResId());
    } else {
      deposit.setSubmissionPolicyId(submissionPolicy.getResId());
    }
    if (preservationPolicy == null) {
      deposit.setPreservationPolicyId(orgUnit.getDefaultPreservationPolicy().getResId());
    } else {
      deposit.setPreservationPolicyId(preservationPolicy.getResId());
    }

    return this.depositService.create(deposit);
  }

  protected void verifyDeposit(Deposit deposit) {
    checkDeposit(deposit);
    checkSIP(deposit);
    checkAIP(deposit, this.aipService);
  }

  protected void checkDeposit(Deposit deposit) {
    log.info("Deposit [{}] {}", deposit.getResId(), deposit.getTitle());
    assertTrue(deposit.getStatus() == DepositStatus.COMPLETED
            || deposit.getStatus() == DepositStatus.CLEANED,
            "Deposit not completed: " + deposit.getResId());
  }

  protected void checkSIP(Deposit deposit) {
    RepresentationInfo depositInfo = this.getDepositInfo(deposit);
    SubmissionInfoPackage sip = this.sipService.findOne(deposit.getSipId());
    assertNotNull(sip, "Cannot find SIP for deposit " + deposit.getResId());
    log.info("SIP     [{}] {}", sip.getResId(), sip.getInfo().getName());
    assertTrue(sip.getInfo().getStatus() == PackageStatus.COMPLETED
            || sip.getInfo().getStatus() == PackageStatus.CLEANED,
            "SIP not completed for deposit " + deposit.getResId());
    this.checkInfo(deposit.getResId(), "SIP => ", depositInfo, sip.getInfo(), true);
    List<ArchivalInfoPackage> depositCollection = this.depositService.getDepositAip(deposit.getResId());
    if (!depositCollection.isEmpty()) {
      this.checkCollectionContent(
              deposit.getResId(),
              depositCollection,
              this.sipService.getAip(sip.getResId()));
    }
  }

  private void checkCollectionContent(String depositId, List<ArchivalInfoPackage> aipList1, List<ArchivalInfoPackage> aipList2) {
    assertEquals(aipList1.size(), aipList2.size(), "Different collection content number for deposit " + depositId);
    List<String> ids = aipList1.stream().map(ArchivalInfoPackage::getResId).collect(Collectors.toList());
    ids.removeAll(aipList2.stream().map(ArchivalInfoPackage::getResId).collect(Collectors.toList()));
    assertEquals(0, ids.size(), "Different collection content for deposit " + depositId);
  }

  protected void checkAIP(Deposit deposit, AIPClientService aipService) {
    RepresentationInfo depositInfo = this.getDepositInfo(deposit);
    SubmissionInfoPackage sip = this.sipService.findOne(deposit.getSipId());
    ArchivalInfoPackage aip = aipService.findOne(sip.getAipId());
    assertNotNull(aip, "Cannot find AIP for deposit " + deposit.getResId());
    log.info("AIP     [{}] {}", aip.getResId(), aip.getInfo().getName());
    assertEquals(PackageStatus.COMPLETED, aip.getInfo().getStatus(), "AIP not completed for deposit " + deposit.getResId());
    this.checkInfo(deposit.getResId(), "AIP =>", depositInfo, aip.getInfo(), true);
    List<AipDataFile> aipDataFiles = aipService.getAipDataFile(aip.getResId());
    assertTrue(aipDataFiles.isEmpty(), "AIP contains data files for deposit" + deposit.getResId());
    List<ArchivalInfoPackage> depositCollection = this.depositService.getDepositAip(deposit.getResId());
    if (!depositCollection.isEmpty()) {
      this.checkCollectionContent(
              deposit.getResId(),
              depositCollection,
              this.aipService.getAip(aip.getResId()));
    }
  }

  protected void checkArchive(Deposit deposit) {
    RepresentationInfo depositInfo = this.getDepositInfo(deposit);
    SubmissionInfoPackage sip = this.sipService.findOne(deposit.getSipId());
    ArchiveMetadata metadata = this.metadataService.findOne(sip.getAipId());
    assertNotNull(metadata, "Cannot find archive for deposit " + deposit.getResId());
    RepresentationInfo archiveInfo = this.getArchiveInfo(metadata);
    log.info("Archive [{}] {}", metadata.getResId(), archiveInfo.getName());
    this.checkInfo(deposit.getResId(), "Archive =>", depositInfo, archiveInfo, false);
    this.checkMetadataInfo(deposit.getResId(), deposit, metadata);
  }

  private void checkInfo(String depositId, String prefix, RepresentationInfo info1, RepresentationInfo info2, boolean allFields) {
    assertEquals(info1.getMetadataVersion(), info2.getMetadataVersion(), prefix + "Not same metadata version for deposit [" + depositId + "]");
    assertEquals(info1.getName(), info2.getName(), prefix + "Not same name");
    assertEquals(info1.getOrganizationalUnitId(), info2.getOrganizationalUnitId(), prefix + "Not same orgUnit for deposit [" + depositId + "]");
    assertEquals(info1.getAccess(), info2.getAccess(), prefix + "Not same access level for deposit " + depositId + "]");
    if (info1.getEmbargo() != null && info2.getEmbargo() != null) {
      assertEquals(info1.getEmbargo().getAccess(), info2.getEmbargo().getAccess(),
              prefix + "Not same embargo access level for deposit [" + depositId + "]");
      assertEquals(info1.getEmbargo().getMonths(), info2.getEmbargo().getMonths(),
              prefix + "Not same embargo month number for deposit [" + depositId + "]");
      if (info1.getEmbargo().getStartDate() != null && info2.getEmbargo().getStartDate() != null) {
        assertEquals(info1.getEmbargo().getStartDate().getYear(), info2.getEmbargo().getStartDate().getYear(),
                prefix + "Not same embargo start date (year) for deposit [" + depositId + "]");
        assertEquals(info1.getEmbargo().getStartDate().getMonth(), info2.getEmbargo().getStartDate().getMonth(),
                prefix + "Not same embargo start date (month) for deposit [" + depositId + "]");
        assertEquals(info1.getEmbargo().getStartDate().getDayOfMonth(), info2.getEmbargo().getStartDate().getDayOfMonth(),
                prefix + "Not same embargo start date (day) for deposit [" + depositId + "]");
      }
    } else if ((info1.getEmbargo() != null && info2.getEmbargo() == null)
            || (info1.getEmbargo() == null && info2.getEmbargo() != null)) {
      assertTrue(false, prefix + "Not same embargo for deposit [" + depositId + "]");
    }
    assertEquals(info1.getDataSensitivity(), info2.getDataSensitivity(), prefix + "Not same data tag for deposit [" + depositId + "]");
    assertEquals(info1.getComplianceLevel(), info2.getComplianceLevel(), prefix + "Not same compliance level for deposit [" + depositId + "]");
    if (allFields) {
      assertEquals(info1.getLicenseId(), info2.getLicenseId(), prefix + "Not same license for deposit [" + depositId + "]");
    }
  }

  private RepresentationInfo getDepositInfo(Deposit deposit) {
    RepresentationInfo info = new RepresentationInfo();
    info.setMetadataVersion(deposit.getMetadataVersion());
    info.setName(deposit.getTitle());
    info.setDescription(StringTool.truncateOnSpaceWithEllipsis(
            deposit.getDescription(),
            DLCMConstants.DB_LONG_STRING_LENGTH));
    info.setOrganizationalUnitId(deposit.getOrganizationalUnitId());
    info.setAccess(deposit.getAccess());
    info.setEmbargo(deposit.getEmbargo());
    info.setDataSensitivity(deposit.getDataSensitivity());
    info.setLicenseId(deposit.getLicenseId());
    info.setComplianceLevel(deposit.getComplianceLevel());
    info.setContainsUpdatedMetadata(deposit.getContainsUpdatedMetadata());
    return info;
  }

  private void checkMetadataInfo(String depositId, Deposit deposit, ArchiveMetadata metadata) {
    // DOI
    assertEquals(deposit.getDoi(), metadata.getDoi(), "Not same DOI for deposit [" + depositId + "]");
    // Related DOI
    String identicalToDoi = null;
    String referencedByDoi = null;
    String obsoletedByDoi = null;
    for (Map<String, Object> relationInfo : metadata.getRelations()) {
      if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_ID_TYPE_NAME)).equals(DLCMConstants.DOI)) {
        if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_TYPE_NAME)).equals("IsIdenticalTo")) {
          identicalToDoi = (String) relationInfo.get(DLCMConstants.CONTENT_FIELD);
        } else if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_TYPE_NAME)).equals("IsReferencedBy")) {
          referencedByDoi = (String) relationInfo.get(DLCMConstants.CONTENT_FIELD);
        } else if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_TYPE_NAME)).equals("IsObsoletedBy")) {
          obsoletedByDoi = (String) relationInfo.get(DLCMConstants.CONTENT_FIELD);
        }
      }
    }
    assertEquals(deposit.getIsIdenticalTo(), identicalToDoi, "Not same 'isIdenticalTo' DOI for deposit [" + depositId + "]");
    assertEquals(deposit.getIsReferencedBy(), referencedByDoi, "Not same 'IsReferencedBy' DOI for deposit [" + depositId + "]");
    assertEquals(deposit.getIsObsoletedBy(), obsoletedByDoi, "Not same 'IsObsoletedBy' DOI for deposit [" + depositId + "]");
    assertEquals(deposit.getKeywords(), metadata.getKeywords(), "Not same keywords for deposit [" + depositId + "]");
    // Collection
    List<ArchivalInfoPackage> depositCollection = this.depositService.getDepositAip(deposit.getResId());
    List<ArchivalInfoPackage> metadataCollection = new ArrayList<>();
    for (Map<String, Object> relationInfo : metadata.getRelations()) {
      if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_TYPE_NAME)).equals("HasPart")) {
        if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_ID_TYPE_NAME)).equals(DLCMConstants.DOI)) {
          ArchiveMetadata archive = this.metadataService.searchByDoi((String) relationInfo.get(DLCMConstants.CONTENT_FIELD));
          metadataCollection.add(this.aipService.findOne(archive.getResId()));
        } else if (((String) relationInfo.get(DLCMConstants.DATACITE_RELATION_ID_TYPE_NAME)).equals(DLCMConstants.HANDLE)) {
          metadataCollection.add(this.aipService.findOne((String) relationInfo.get(DLCMConstants.CONTENT_FIELD)));
        }
      }
    }
    if (!depositCollection.isEmpty()) {
      this.checkCollectionContent(deposit.getResId(), depositCollection, metadataCollection);
    }

  }

  private RepresentationInfo getArchiveInfo(ArchiveMetadata archive) {
    RepresentationInfo info = new RepresentationInfo();
    info.setMetadataVersion(DLCMMetadataVersion.fromVersion(archive.getMetadataVersion()));
    info.setName(archive.getTitle());
    info.setDescription(StringTool.truncateOnSpaceWithEllipsis(
            archive.getAbstract(),
            DLCMConstants.DB_LONG_STRING_LENGTH));
    info.setOrganizationalUnitId(archive.getOrganizationalUnitId());
    info.setAccess(archive.getFinalAccessLevel());
    if (archive.getEmbargoAccessLevel().isPresent()
            && !archive.getFinalAccessLevel().equals(archive.getEmbargoAccessLevel().get())) {
      OffsetDateTime startDate = null;
      OffsetDateTime endDate = null;
      EmbargoInfo embargo = new EmbargoInfo();
      embargo.setAccess(archive.getEmbargoAccessLevel().get());
      for (Map<String, Object> dateInfo : (ArrayList<Map<String, Object>>) archive
              .getMetadata(DLCMConstants.DATES_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP))) {
        if (((String) dateInfo.get("dateType")).equals("Created")) {
          startDate = StringTool.getDate((String) dateInfo.get("content"));
        } else if (((String) dateInfo.get("dateType")).equals("Available")) {
          endDate = StringTool.getDate((String) dateInfo.get("content"));
        }
      }
      embargo.setMonths((int) ChronoUnit.MONTHS.between(startDate, endDate));
      embargo.setStartDate(startDate);
      info.setEmbargo(embargo);
    } else {
      info.setEmbargo(null);
    }
    info.setDataSensitivity(DataTag.valueOf((String) archive.getMetadata(DLCMConstants.AIP_DATA_TAG)));
    // info.setLicenseId(archive.getLicenseId());
    info.setComplianceLevel(ComplianceLevel.valueOf((String) archive.getMetadata(DLCMConstants.AIP_COMPLIANCE_LEVEL)));
    return info;
  }

}
