/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - AdminDemoIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.demo;

import static org.junit.jupiter.api.Assertions.assertNotNull;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.MethodOrderer;
import org.junit.jupiter.api.Order;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestMethodOrder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.Resource;

import ch.unige.solidify.util.StringTool;

import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.settings.FundingAgency;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.License;

/*
 * Demo execution order
 * =================================================================================
 * 1) AdminDemoIT                      => To define & update settings
 * 2) OrgUnitDemoIT                    => To define & update orgUnits
 * 3) DepositsDemoIT                   => To create archives
 * 4) CollectionsDemoIT                => To create collections of archives
 * 5) CollectionsOfCollectionsDemoIT   => To create collections of collections (WIP)
 * 6) MetadataEditionDepositsDemoIT    => To update archives
 * 7) MetadataEditionCollectionsDemoIT => To update collections of archives
 * 9) ArchivesDemoIT                   => To check archives 
 */

@Order(1)
@TestMethodOrder(MethodOrderer.OrderAnnotation.class)
@Disabled(DLCMDemoConstants.DEMO_MESSAGE)
class AdminDemoIT extends AbstractAdminDemoIT {
  private static final Logger log = LoggerFactory.getLogger(AdminDemoIT.class);

  private final String LICENSE_FOLDER = "license";
  private final String FUNDING_AGENCY_FOLDER = "fundingAgency";
  private final String INSTITUTION_FOLDER = "institution";

  @Order(10)
  @Test
  void createSubmissionPolicy() {
    final List<SubmissionPolicy> submissionPolicies = this.submissionPolicyService
            .searchByProperties(Map.of("name", DLCMDemoConstants.DEMO_SUBMISSION_POLICY));
    if (submissionPolicies.isEmpty()) {
      // Create a policy
      final SubmissionPolicy policy = new SubmissionPolicy();
      policy.setName(DLCMDemoConstants.DEMO_SUBMISSION_POLICY);
      policy.setSubmissionApproval(true);
      policy.setTimeToKeep(1);
      final SubmissionPolicy createdPolicy = this.submissionPolicyService.create(policy);

      // Test the creation
      assertNotNull(createdPolicy, "In creating submission policy");
    }
  }

  @Order(10)
  @Test
  void createPreservationPolicy() {
    final List<PreservationPolicy> preservationPolicies = this.preservationPolicyService
            .searchByProperties(Map.of("name", DLCMDemoConstants.DEMO_PRESERVATION_POLICY));
    if (preservationPolicies.isEmpty()) {
      // Create a policy
      final PreservationPolicy policy = new PreservationPolicy();
      policy.setName(DLCMDemoConstants.DEMO_PRESERVATION_POLICY);
      policy.setDispositionApproval(true);
      policy.setRetention(1);
      final PreservationPolicy createdPolicy = this.preservationPolicyService.create(policy);

      // Test the creation
      assertNotNull(createdPolicy, "In creating preservation policy");
    }
  }

  @Order(20)
  @Test
  void addLicenseLogo() throws IOException {
    for (Resource file : this.listFiles(LICENSE_FOLDER, DLCMDemoConstants.PNG_FILES)) {
      String spdxId = this.getFilename(file);
      log.info("License '{}'", spdxId);
      // Get License by SPDX ID
      License license = this.licenseService.findBySpdxId(spdxId);
      assertNotNull(license, "License not found: " + spdxId);
      this.licenseService.uploadLogo(license.getResId(), file);
    }
  }

  @Order(30)
  @Test
  void addFundingAgencyRorId() throws IOException {
    for (Entry<String, String> rorInfo : this.getFundingAgencyRorIdList().entrySet()) {
      log.info("Funding agency '{}'", rorInfo.getKey());
      FundingAgency fa = this.fundingAgencyService.findByAcronym(rorInfo.getKey());
      assertNotNull(fa, "Funding agency not found: " + rorInfo.getKey());
      if (StringTool.isNullOrEmpty(fa.getRorId())) {
        fa.setRorId(rorInfo.getValue());
        this.fundingAgencyService.update(fa.getResId(), fa);
      } else {
        log.warn("Funding agency  '{}' not updated", rorInfo.getKey());
      }
    }
  }

  @Order(31)
  @Test
  void addFundingAgencyLogo() throws IOException {
    for (Resource file : this.listFiles(FUNDING_AGENCY_FOLDER, DLCMDemoConstants.PNG_FILES)) {
      String rorId = this.getFilename(file);
      log.info("Funding agency '{}'", rorId);
      FundingAgency fa = this.fundingAgencyService.findByRorId(rorId);
      if (fa == null) {
        fa = this.fundingAgencyService.findByAcronym(rorId);
      }
      assertNotNull(fa, "Funding agency not found: " + rorId);
      this.fundingAgencyService.uploadLogo(fa.getResId(), file);
    }
  }

  @Order(40)
  @Test
  void addInstitutionRorId() throws IOException {
    for (Entry<String, String> rorInfo : this.getInstitutionRorIdList().entrySet()) {
      log.info("Institution '{}'", rorInfo.getKey());
      Institution institution = this.institutionService.findByName(rorInfo.getKey());
      assertNotNull(institution, "Institution not found: " + rorInfo.getKey());
      if (StringTool.isNullOrEmpty(institution.getRorId())) {
        institution.setRorId(rorInfo.getValue());
        this.institutionService.update(institution.getResId(), institution);
      } else {
        log.warn("Institution '{}' not updated", rorInfo.getKey());
      }
    }
  }

  @Order(41)
  @Test
  void addInstitutionLogo() throws IOException {
    for (Resource file : this.listFiles(INSTITUTION_FOLDER, DLCMDemoConstants.PNG_FILES)) {
      String rorId = this.getFilename(file);
      log.info("Institution '{}'", rorId);
      Institution institution = this.institutionService.findByRorId(rorId);
      if (institution == null) {
        institution = this.institutionService.findByName(rorId);
      }
      assertNotNull(institution, "Institution not found: " + rorId);
      this.institutionService.uploadLogo(institution.getResId(), file);
    }
  }

  private Map<String, String> getFundingAgencyRorIdList() {
    return Map.of(
            "SNF", "00yjd3n13",
            "CTI", "012y9zp98",
            "swissuniversities", "02fsagg23");
  }

  private Map<String, String> getInstitutionRorIdList() {
    Map<String, String> list = new HashMap<>();
    list.put("UNIGE", "01swzsf04");
    list.put("EPFL", "02s376052");
    list.put("ETH", "05a28rw58");
    list.put("UZH", "02crff812");
    list.put("HES-SO", "01xkakk17");
    list.put("ZHAW", "05pmsvm27");
    list.put("HUG", "01m1pv723");
    list.put("IHEID", "007ygn379");
    list.put("UNIL", "019whta54s");
    list.put("CERN", "01ggx4157");
    list.put("UNIFR", "022fs9h90");
    list.put("BFH", "02bnkt322");
    return list;
  }
}
