/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - InitDataIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.init;

import static ch.dlcm.DLCMConstants.PERMANENT_DEPOSIT;
import static ch.dlcm.test.DLCMTestConstants.PRESERVATION_POLICY_TEST_NAME;
import static ch.dlcm.test.DLCMTestConstants.SUBMISSION_POLICY_TEST_NAME;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.junit.jupiter.api.Order;
import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;

import ch.unige.solidify.auth.model.AuthApplicationRole;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.dto.FileUploadDto;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.preingest.DepositDataFile;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

@Order(Integer.MIN_VALUE)
class InitDataIT extends AbstractInitDataIT {
  private static final Logger logger = LoggerFactory.getLogger(InitDataIT.class);

  private final static int AIU_NUM = 3;

  @Override
  protected void setUser() {
    this.restClientTool.sudoAdmin();
  }

  public void initCollectionData() {
    // Create Archival Unit
    for (int i = 1; i <= AIU_NUM; i++) {
      final String title = DLCMConstants.ARCHIVAL_UNIT_LABEL + " " + i;
      if (this.isExactlyOneAipInStatus(title, PackageStatus.IN_ERROR)) {
        this.resumeAIP(title);
      } else if (this.isNoAip(title)) {
        this.createAndValidatePermanentDeposit(title, DEFAULT_DESCRIPTION);
      }
    }
    // Create Archival collection
    final String title = DLCMConstants.ARCHIVAL_COLLECTION_LABEL;
    if (this.isExactlyOneAipInStatus(title, PackageStatus.IN_ERROR)) {
      logger.warn("Archival Collection IN_ERROR trying to resume");
      this.resumeAIP(title);
    } else if (this.isNoAip(title)) {
      final Deposit deposit = this.createRemoteDepositWithContributor(false, title, DEFAULT_DESCRIPTION);
      this.createAIC(deposit, AIU_NUM, false);
      this.approveDeposit(deposit);
      this.waitDepositIsCompleted(deposit.getResId());
      this.waitAIPIsCompletedOrError(deposit.getOrganizationalUnitId(), deposit.getTitle());
    }
  }

  @Override
  protected void createFixtures() {
    this.initUsers();
    this.initPreservationPolicy();
    this.initSubmissionPolicy();
    this.initPersonOrgUnitRole();
    this.createOrgUnitForNoOne();
    this.initPermanentDeposit();
    this.initArchiveToUpdate();
    this.initCollectionData();
  }

  public void initUsers() {
    // Make a REST call with test users, in order to present their tokens to the DLCM application and create their users
    this.restClientTool.sudoUser();
    this.submissionPolicyService.findAll();
    this.restClientTool.exitSudo();

    this.restClientTool.sudoThirdPartyUser();
    this.submissionPolicyService.findAll();
    this.restClientTool.exitSudo();

    this.restClientTool.sudoAdmin();
    this.submissionPolicyService.findAll();
    this.restClientTool.exitSudo();

    this.restClientTool.sudoRoot();
    this.submissionPolicyService.findAll();
    this.restClientTool.exitSudo();
  }

  public void initPreservationPolicy() {
    final String preservationPolicyName = DLCMTestConstants.getNameWithPermanentLabel(PRESERVATION_POLICY_TEST_NAME);
    if (this.preservationPolicyService.searchByProperties(Map.of("name", preservationPolicyName)).isEmpty()) {
      final PreservationPolicy localPreservationPolicy = new PreservationPolicy();
      localPreservationPolicy.setName(preservationPolicyName);
      localPreservationPolicy.setRetention(5 * DLCMConstants.DAYS_BY_YEAR);
      this.preservationPolicyService.create(localPreservationPolicy);
    }
  }

  public void initSubmissionPolicy() {
    final String submissionPolicyName = DLCMTestConstants.getNameWithPermanentLabel(SUBMISSION_POLICY_TEST_NAME);
    if (this.submissionPolicyService.searchByProperties(Map.of("name", submissionPolicyName)).isEmpty()) {
      final SubmissionPolicy localSubmissionPolicy = new SubmissionPolicy();
      localSubmissionPolicy.setName(submissionPolicyName);
      localSubmissionPolicy.setSubmissionApproval(false);
      this.submissionPolicyService.create(localSubmissionPolicy);
    }
  }

  public void initPersonOrgUnitRole() {
    // Needed to create org unit for each role as ApplicationRole User
    final List<Role> listRoles = this.roleService.findAll();
    for (final Role role : listRoles) {
      this.getOrCreatePermanentOrganizationalUnit(OrganizationalUnitStatus.OPEN, role.getName().toUpperCase(), AuthApplicationRole.USER_ID);
    }
  }

  public void initPermanentDeposit() {
    List<Deposit> depositList = this.depositService.searchByProperties(Collections.singletonMap("title", PERMANENT_DEPOSIT));
    if (depositList.isEmpty()) {
      final Deposit newDeposit = this.createNewDeposit(false, PERMANENT_DEPOSIT, Role.MANAGER_ID);

      FileUploadDto<DepositDataFile> fileUploadDto = new FileUploadDto<>();
      fileUploadDto.setDataCategory(DataCategory.Primary);
      fileUploadDto.setDataType(DataCategory.Reference);
      this.depositService.uploadDepositDataFile(newDeposit.getResId(), new ClassPathResource("bootstrap.yml"), fileUploadDto);
    } else {
      assertEquals(DepositStatus.IN_PROGRESS, depositList.get(0).getStatus(), "Permanent deposit should be IN_PROGRESS");
    }
  }

  public void createOrgUnitForNoOne() {
    OrganizationalUnit organizationalUnit = this.createLocalOrganizationalUnitWithPermanentLabel(OrganizationalUnitStatus.OPEN, NO_ONE);
    final List<OrganizationalUnit> orgUnits = this.organizationalUnitService.searchByProperties(Map.of("name", organizationalUnit.getName()));
    if (orgUnits.isEmpty()) {
      this.organizationalUnitService.create(organizationalUnit);
    }
  }

  public void initArchiveToUpdate() {
    if (this.isNoAip(DEFAULT_TITLE_TO_UPDATE)) {
      String title = this.getNewArchiveToUpdateTitle(DEFAULT_TITLE_TO_UPDATE);
      this.createAndValidatePermanentDeposit(title, DEFAULT_DESCRIPTION_TO_UPDATE);
    }
  }

  private void createAndValidatePermanentDeposit(String title, String description) {
    Deposit deposit = this.createRemoteDepositWithContributor(false, title, description);
    deposit.addItem(this.createLocalDepositDataFile(deposit));

    deposit = this.createRemoteDepositWithData(deposit);
    this.waitDepositWithDataIsReady(deposit.getResId());
    this.approveDeposit(deposit);
    this.waitDepositIsCompleted(deposit.getResId());
    this.waitAIPIsCompletedOrError(deposit.getOrganizationalUnitId(), deposit.getTitle());
  }

  @Test
  void verifyUserPerson() {
    final List<Person> personList = this.personService.findAll();
    final List<User> userList = this.userService.findAll();
    final List<String> roles = new ArrayList<>(Arrays.asList("ADMIN", "USER", "ROOT"));
    for (String role : roles) {
      List<Person> rolePersons = personList.stream()
              .filter(person -> person.getFirstName().toLowerCase().contains((DLCMConstants.PERMANENT_TEST_DATA_LABEL + role).toLowerCase()))
              .collect(Collectors.toList());
      List<User> roleUsers = userList.stream()
              .filter(user -> user.getFirstName().toLowerCase().contains((DLCMConstants.PERMANENT_TEST_DATA_LABEL + role).toLowerCase()))
              .collect(Collectors.toList());
      assertEquals(1, rolePersons.size(), "There must be exactly one " + role + " person");
      assertEquals(1, roleUsers.size(), "There must be exactly one " + role + " user");
      assertNotNull(roleUsers.get(0).getPerson(), "The role " + role + " must have a person connected to its user");
      assertEquals(role, roleUsers.get(0).getApplicationRole().getResId(), "User for " + role + " has not a " + role + " role");
    }
  }
}
