/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - DepositUserManagerIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.preingest;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.HttpServerErrorException;

import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

class DepositUserManagerIT extends AbstractPreingestIT {

  @Test
  void contributorTest() {
    final Deposit deposit = this.createRemoteDeposit(true, DEFAULT_TITLE, DEFAULT_DESCRIPTION);

    // Create a new Person
    final Person person1 = this.getOrCreateRemotePerson();

    // Add the person as contributor
    this.depositService.addContributor(deposit.getResId(), person1.getResId());
    final List<Person> listContributors = this.depositService.getContributors(deposit.getResId());

    assertEquals(1, listContributors.size());
    final Person person2 = listContributors.get(0);

    // Test the contributor
    assertEquals(person1.getResId(), person2.getResId());
    assertEquals(person1.getFirstName(), person2.getFirstName());
    assertEquals(person1.getLastName(), person2.getLastName());
  }

  @Test
  void creationTest() {
    final OrganizationalUnit organizationalUnit = this.getOrCreatePermanentOrganizationalUnit(OrganizationalUnitStatus.OPEN);
    final String title = DLCMTestConstants.getRandomNameWithTemporaryLabel("deposit created by test");

    // Create a deposit
    final Deposit deposit1 = this.createLocalDeposit(true, title);
    final Deposit deposit2 = this.depositService.create(deposit1);

    // Test the creation
    assertEquals(deposit1.getTitle(), deposit2.getTitle());

    // Check if pending ( InProgress, inValidation, inError )
    assertEquals(DepositStatus.IN_PROGRESS, deposit2.getStatus());

    // Check that the deposit has been created
    final List<Deposit> deposits = this.depositService.searchByProperties(Map.of("title", title,
            DLCMConstants.ORG_UNIT_ID_FIELD, organizationalUnit.getResId()));
    assertEquals(1, deposits.size());
  }

  @Test
  void findAllTest() {
    final OrganizationalUnit organizationalUnit = this.getOrCreatePermanentOrganizationalUnit(OrganizationalUnitStatus.OPEN);

    // Create deposit 1
    final String title1 = "deposit created by findAllTest";
    final Deposit deposit1 = this.createLocalDeposit(true, title1);
    final Deposit storedDeposit1 = this.depositService.create(deposit1);
    final String tmpTitle1 = deposit1.getTitle();

    // Create deposit 2
    final String title2 = "deposit created by findAllTest";
    final Deposit deposit2 = this.createLocalDeposit(true, title2);
    final Deposit storedDeposit2 = this.depositService.create(deposit2);
    final String tmpTitle2 = deposit2.getTitle();

    // Check that the deposits have been created
    List<Deposit> deposits = this.depositService.searchByProperties(Map.of(DLCMConstants.ORG_UNIT_ID_FIELD, organizationalUnit.getResId()));
    assertTrue(deposits.stream().anyMatch(deposit -> deposit.getTitle().equals(tmpTitle1)));
    assertTrue(deposits.stream().anyMatch(deposit -> deposit.getTitle().equals(tmpTitle2)));

    // Filter on title
    deposits = this.depositService
            .searchByProperties(Map.of("title", tmpTitle1, DLCMConstants.ORG_UNIT_ID_FIELD, organizationalUnit.getResId()));
    assertTrue(deposits.stream().anyMatch(deposit -> deposit.getTitle().equals(tmpTitle1)));
    assertFalse(deposits.stream().anyMatch(deposit -> deposit.getTitle().equals(tmpTitle2)));

    // Filter on title + statusList=IN_PROGRESS,COMPLETED
    deposits = this.depositService.searchByProperties(
            Map.of("title", tmpTitle1, "statusList", "IN_PROGRESS,COMPLETED", DLCMConstants.ORG_UNIT_ID_FIELD, organizationalUnit.getResId()));
    assertTrue(deposits.stream().anyMatch(deposit -> deposit.getTitle().equals(tmpTitle1)));
    assertFalse(deposits.stream().anyMatch(deposit -> deposit.getTitle().equals(tmpTitle2)));
  }

  @Test
  void creationWithExistingResIdTest() {
    final OrganizationalUnit organizationalUnit = this.getOrCreatePermanentOrganizationalUnit(OrganizationalUnitStatus.OPEN, Role.MANAGER_ID);

    final Deposit deposit = this.createRemoteDeposit(true, "creationWithExistingResIdTest", DEFAULT_DESCRIPTION);
    final String existingResId = deposit.getResId();

    final Deposit newDeposit = this.createLocalDeposit(true, "should not be saved");
    newDeposit.setResId(existingResId);

    Deposit savedDeposit = null;
    try {
      savedDeposit = this.depositService.create(newDeposit);
      fail("An 500 Http response should be received");
    } catch (final HttpServerErrorException ex) {
      assertEquals(500, ex.getRawStatusCode());
    }

    assertNull(savedDeposit);
    deposit.setCollectionBegin(OffsetDateTime.now(ZoneOffset.UTC).minusYears(5));

    // Check that the deposit has been created
    final List<Deposit> deposits = this.depositService.searchByProperties(Map.of("title", deposit.getTitle(),
            DLCMConstants.ORG_UNIT_ID_FIELD, organizationalUnit.getResId()));
    assertEquals(1, deposits.size());

    // Check that the deposit has been created
    final List<Deposit> deposits2 = this.depositService.searchByProperties(Map.of("title", newDeposit.getTitle(),
            DLCMConstants.ORG_UNIT_ID_FIELD, organizationalUnit.getResId()));
    assertEquals(0, deposits2.size());
  }

  @Test
  void creationWithResIdTest() {

    final Deposit deposit = this.createLocalDeposit(true, DEFAULT_TITLE);
    final String resId = StringTool.generateResId();
    deposit.setResId(resId);
    final Deposit savedDeposit = this.depositService.create(deposit);

    assertNotNull(savedDeposit);
    assertEquals(savedDeposit.getResId(), resId);
  }

  @Test
  void deleteTest() {
    Deposit deposit = this.createRemoteDeposit(true, DEFAULT_TITLE, DEFAULT_DESCRIPTION);

    // Delete the deposit
    final String resId = deposit.getResId();
    this.depositService.delete(resId);
    deposit = this.depositService.findOne(resId);

    // Test deletion
    assertNull(deposit);
  }

  @Test
  void dontUpdateResIdTest() {
    final Deposit deposit = this.createRemoteDeposit(true, DEFAULT_TITLE, DEFAULT_DESCRIPTION);
    final String currentResId = deposit.getResId();

    final String newTitle = DLCMTestConstants.getRandomNameWithTemporaryLabel("B");
    final String newResId = "xyz-xyz-xyz";

    // Update the deposit with a new resId
    deposit.setTitle(newTitle);
    deposit.setResId(newResId);
    final Deposit savedDeposit = this.depositService.update(currentResId, deposit);

    assertNotNull(savedDeposit);
    assertEquals(savedDeposit.getResId(), currentResId); // resId has not changed
    assertEquals(savedDeposit.getTitle(), newTitle); // title has changed
  }

  /**
   * Updating the deposit status with an update is forbidden
   */
  @Test
  void dontUpdateStatusTest() {
    assertThrows(HttpClientErrorException.class, () -> {
      final Deposit deposit = this.createNewDeposit(true, DEFAULT_TITLE, DEFAULT_DESCRIPTION);

      deposit.setStatus(DepositStatus.IN_VALIDATION);
      this.depositService.update(deposit.getResId(), deposit);
    });
  }

  @Test
  void reserveDOITest() {
    final Deposit deposit = this.createRemoteDeposit(true, DEFAULT_TITLE, DEFAULT_DESCRIPTION);
    final String depositId = deposit.getResId();

    final String newDOI = this.depositService.reserveDOI(depositId);

    /*
     * Fetch again Deposit
     */
    final Deposit fetchedDeposit = this.depositService.findOne(depositId);
    assertNotNull(fetchedDeposit.getDoi());
    assertEquals(newDOI, fetchedDeposit.getDoi());
  }

  @Test
  void updateTest() {
    final Deposit deposit = this.createRemoteDeposit(true, DEFAULT_TITLE, DEFAULT_DESCRIPTION);

    // Update the deposit
    deposit.setTitle(DLCMTestConstants.getRandomNameWithTemporaryLabel("B"));
    final Deposit actualDeposit = this.depositService.update(deposit.getResId(), deposit);
    final Deposit deposit4 = this.depositService.findOne(deposit.getResId());

    // Test the update
    assertEquals(deposit.getTitle(), deposit4.getTitle());
    this.assertEqualsWithoutNanoSeconds(deposit.getCreationTime(), deposit4.getCreationTime());
    assertNotNull(actualDeposit);
  }

}
