/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - DLCMMetadataVersion.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonValue;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;

public enum DLCMMetadataVersion {
  V1_0("1.0"), // v1.0: DataCite 4.0 + Premis 3.0 + DLCM Info 1.0
  V1_1("1.1"), // v1.1: DataCite 4.0 + Premis 3.0 + DLCM Info 1.0
  V2_0("2.0"), // v2.0: DataCite 4.3 + Premis 3.0 + DLCM Info 2.0
  V2_1("2.1"), // v2.1: DataCite 4.3 + Premis 3.0 + DLCM Info 2.1
  V3_0("3.0"); // v3.0: DataCite 4.4 + Premis 3.0 + DLCM Info 2.1

  private static final String METS_PROFILE_PREFIX = "dlcm_profile-";
  private static final String METS_SCHEMA_PREFIX = "dlcm_mets-";
  private static final String PREMIS_3 = "premis-3.0.xsd";
  private static final String REP_INFO_SCHEMA_PREFIX = "dlcm_datacite-";
  private static final String INDEXING_XSL = "dlcm4indexing.xsl";

  public static DLCMMetadataVersion fromVersion(String version) {
    switch (version) {
      case "1.0":
        return V1_0;
      case "1.1":
        return V1_1;
      case "2.0":
        return V2_0;
      case "2.1":
        return V2_1;
      case "3.0":
        return V3_0;
      default:
        throw new SolidifyRuntimeException("Wrong metadata version: " + version);
    }
  }

  public static DLCMMetadataVersion getDefaultVersion() {
    return DLCMMetadataVersion.V3_0;
  }

  private String version;

  private DLCMMetadataVersion(String v) {
    this.version = v;
  }

  public String getAdministrativeInfoSchema() {
    return PREMIS_3;
  }

  public String getMetsProfile() {
    return METS_PROFILE_PREFIX + this.getVersion() + SolidifyConstants.XML_EXT;
  }

  public String getMetsSchema() {
    return METS_SCHEMA_PREFIX + this.getVersion() + SolidifyConstants.XSD_EXT;
  }

  public String getRepresentationInfoSchema() {
    switch (this.version) {
      case "1.0":
        return REP_INFO_SCHEMA_PREFIX + this.getVersion() + SolidifyConstants.XSD_EXT;
      case "1.1":
        return REP_INFO_SCHEMA_PREFIX + V1_0.getVersion() + SolidifyConstants.XSD_EXT;
      case "2.0":
        return REP_INFO_SCHEMA_PREFIX + this.getVersion() + SolidifyConstants.XSD_EXT;
      case "2.1":
        return REP_INFO_SCHEMA_PREFIX + V2_0.getVersion() + SolidifyConstants.XSD_EXT;
      case "3.0":
        return REP_INFO_SCHEMA_PREFIX + V3_0.getVersion() + SolidifyConstants.XSD_EXT;
      default:
        throw new SolidifyRuntimeException("Wrong metadata version: " + this.version);
    }
  }

  public String getIndexingTransformation() {
    return INDEXING_XSL;
  }

  @JsonValue
  public String getVersion() {
    return this.version;
  }

  @JsonIgnore
  public String getMajorVersion() {
    return this.version.substring(0, this.version.indexOf('.') + 1);
  }

  @Override
  public String toString() {
    return this.getVersion();
  }
}
