/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - AbstractDataFile.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model;

import static ch.unige.solidify.SolidifyConstants.DB_ID_LENGTH;

import java.net.URI;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Objects;
import javax.persistence.Column;
import javax.persistence.Embedded;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.MappedSuperclass;
import javax.persistence.PrePersist;
import javax.persistence.Transient;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;
import javax.validation.constraints.Size;

import org.apache.commons.validator.routines.UrlValidator;
import org.springframework.data.jpa.domain.Specification;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonProperty.Access;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.SearchableResourceNormalized;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.SearchCriteria;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.oais.MetadataVersionAware;
import ch.dlcm.model.settings.MetadataType;

@MappedSuperclass
public abstract class AbstractDataFile<T extends Resource & MetadataVersionAware, V> extends SearchableResourceNormalized<V> {

  private static final int DB_SOURCE_DATA_LENGTH = 2024;
  private static final int DB_FINAL_DATA_LENGTH = 2024;
  private static final int DB_RELATIVE_LOCATION_LENGTH = 460;

  //@formatter:off
  public enum DataFileStatus {
    // Special processing status
    CHANGE_RELATIVE_LOCATION, CHANGE_DATA_CATEGORY, CLEANED, CLEANING, EXCLUDED_FILE,
    CHECK_COMPLIANCE, CHECK_COMPLIANCE_CLEANED, CHECKED_COMPLIANCE, CHECKED_COMPLIANCE_CLEANED, IGNORED_FILE,
    // File format identification status
    FILE_FORMAT_IDENTIFIED, FILE_FORMAT_SKIPPED, FILE_FORMAT_UNKNOWN,
    IN_ERROR, PROCESSED,
    // Normal processing status
    READY, RECEIVED, TO_PROCESS,
    // Virus check status
    VIRUS_CHECKED, VIRUS_SKIPPED
  }
  //@formatter:on
  private static final String DATA_FILE_PREFIX = "{Data File} ";

  private ComplianceLevel complianceLevel;

  @JsonProperty(required = true)
  @Enumerated(EnumType.STRING)
  private DataCategory dataCategory;

  @JsonProperty(required = true)
  @Enumerated(EnumType.STRING)
  private DataCategory dataType;

  @Embedded
  @Valid
  private FileFormat fileFormat = new FileFormat();

  private Long fileSize;

  @Column(length = DB_FINAL_DATA_LENGTH)
  private URI finalData;

  @NotNull
  private String fileName;

  @Transient
  private MetadataType metadataType;

  @JsonIgnore
  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String metadataTypeId;

  @Size(max = DB_RELATIVE_LOCATION_LENGTH)
  @Pattern(regexp = "(^\\/$)|(^\\/.*(?<!\\/)$)")
  private String relativeLocation;

  @NotNull
  @Column(length = DB_SOURCE_DATA_LENGTH)
  private URI sourceData;

  @Enumerated(EnumType.STRING)
  private DataFileStatus status;

  @Size(max = DLCMConstants.DB_LONG_STRING_LENGTH)
  private String statusMessage;

  @Embedded
  @Valid
  private VirusCheck virusCheck = new VirusCheck();

  /********************************************************************************************************************/

  /**
   * Throws an IllegalArgumentException if data category and data type are not correct
   */
  public void check() {
    // Data category & type consistency
    if (!DataCategory.check(this.getDataCategory(), this.getDataType())) {
      throw new IllegalArgumentException("Wrong data category: " + this.getDataCategory() + "/" + this.getDataType());
    }
    // Custom metadata check
    if (this.getDataCategory().equals(DataCategory.Package) && this.getDataType().equals(DataCategory.CustomMetadata)
            && this.getMetadataTypeId() == null
            && (this.getMetadataType() == null || this.getMetadataType().getResId() == null)) {
      throw new IllegalArgumentException("Missing metadata type for " + this.getDataCategory() + "/" + this.getDataType());
    }
  }

  /**
   * @return the list of data file checksums
   */
  public abstract List<DataFileChecksum> getChecksums();

  /**
   * Return true if the data file is in-progress
   */
  public boolean isInProgress() {
    return (this.getStatus() != DataFileStatus.IN_ERROR &&
            this.getStatus() != DataFileStatus.READY &&
            this.getStatus() != DataFileStatus.CLEANED &&
            this.getStatus() != DataFileStatus.EXCLUDED_FILE &&
            this.getStatus() != DataFileStatus.IGNORED_FILE &&
            this.getStatus() != DataFileStatus.CHECKED_COMPLIANCE_CLEANED &&
            this.getStatus() != DataFileStatus.CHECKED_COMPLIANCE);
  }

  /**
   * @param status the status of the data file
   */
  public void setFileStatus(DataFileStatus status) {
    this.setStatus(status);
    this.setStatusMessage(null);
  }

  public abstract T getInfoPackage();

  public abstract void setInfoPackage(T infoPackage);

  public boolean checkSource() {
    final String[] schemes = { "http", "https", "file" };
    final UrlValidator urlValidator = new UrlValidator(schemes, UrlValidator.ALLOW_LOCAL_URLS);
    return urlValidator.isValid(this.sourceData.toString());
  }

  public Path defineTargetFile(String preparationPath) {
    String target = preparationPath;
    switch (this.dataCategory) {
      case Package:
        switch (this.dataType) {
          case CustomMetadata:
            target += Package.METADATA.getName() + SolidifyConstants.URL_SEP + this.getFileName();
            break;
          case Metadata:
            target += SolidifyConstants.URL_SEP + Package.METADATA_FILE.getName();
            break;
          case InformationPackage:
            target += SolidifyConstants.URL_SEP + this.getFileName();
            break;
          case UpdatedMetadata:
            target += Package.UPDATE.getName() + SolidifyConstants.URL_SEP + DLCMConstants.UPDATED_METADATA_FILE
                    + OffsetDateTime.now().format(DateTimeFormatter.ofPattern(StringTool.DATE_TIME_FORMAT_FOR_FILE)) + SolidifyConstants.XML_EXT;
            break;
          case UpdatePackage:
            target += SolidifyConstants.URL_SEP + DLCMConstants.UPDATE_PACKAGE_FILE + "-"
                    + OffsetDateTime.now().format(DateTimeFormatter.ofPattern(StringTool.DATE_TIME_FORMAT_FOR_FILE));
            break;
          default:
            throw new SolidifyRuntimeException(this.dataType + " is not a data type for the Package data category");
        }
        break;
      case Primary:
        target += Package.RESEARCH.getName() + this.getExtendedRelativeLocation() + SolidifyConstants.URL_SEP + this.getFileName();
        break;
      case Secondary:
        target += Package.DOC.getName() + this.getExtendedRelativeLocation() + SolidifyConstants.URL_SEP + this.getFileName();
        break;
      case Software:
        target += Package.SOFTWARE.getName() + this.getExtendedRelativeLocation() + SolidifyConstants.URL_SEP + this.getFileName();
        break;
      case Internal:
        switch (this.dataType) {
          case DatasetThumbnail:
            target += Package.INTERNAL.getName() + SolidifyConstants.URL_SEP + Package.DATASET_THUMBNAIL.getName();
            break;
          case DatafileThumbnail:
            target += Package.INTERNAL.getName() + this.getExtendedRelativeLocation() + SolidifyConstants.URL_SEP + this.getFileName()
                    + DLCMConstants.THUMBNAIL_EXTENSION;
            break;
          default:
            throw new SolidifyRuntimeException(this.dataType + " is not a data type for the Internal data category");
        }
        break;
      default:
        throw new SolidifyRuntimeException(this.dataCategory + " is not a data category");
    }
    final Path targetPath = Paths.get(target);
    if (FileTool.ensureFolderExists(targetPath.getParent())) {
      return targetPath;
    } else {
      throw new SolidifyRuntimeException(DATA_FILE_PREFIX + " Impossible to create directory : " + targetPath.getParent());
    }
  }

  @Override
  public boolean equals(final Object other) {
    if (this == other) {
      return true;
    }
    if (!(other instanceof AbstractDataFile)) {
      return false;
    }

    final AbstractDataFile<?, ?> castOther = (AbstractDataFile<?, ?>) other;
    return super.equals(other)
            && Objects.equals(this.getSourceData(), castOther.getSourceData())
            && Objects.equals(this.getRelativeLocation(), castOther.getRelativeLocation())
            && Objects.equals(this.getFinalData(), castOther.getFinalData())
            && Objects.equals(this.getFileFormat(), castOther.getFileFormat())
            && Objects.equals(this.getVirusCheck(), castOther.getVirusCheck())
            && Objects.equals(this.getFileSize(), castOther.getFileSize())
            && Objects.equals(this.getStatus(), castOther.getStatus())
            && Objects.equals(this.getDataCategory(), castOther.getDataCategory())
            && Objects.equals(this.getDataType(), castOther.getDataType())
            && Objects.equals(this.getComplianceLevel(), castOther.getComplianceLevel())
            && Objects.equals(this.getInfoPackage(), castOther.getInfoPackage())
            && Objects.equals(this.getChecksums(), castOther.getChecksums());
  }

  public ComplianceLevel getComplianceLevel() {
    return this.complianceLevel;
  }

  public DataCategory getDataCategory() {
    return this.dataCategory;
  }

  public DataCategory getDataType() {
    return this.dataType;
  }

  /*
   * ************************************** Getters and Setters **************************************
   */

  public FileFormat getFileFormat() {
    return this.fileFormat;
  }

  @JsonProperty(access = Access.READ_ONLY)
  public String getFileName() {
    return this.fileName;
  }

  public Long getFileSize() {
    return this.fileSize;
  }

  public URI getFinalData() {
    return this.finalData;
  }

  @JsonProperty(access = Access.READ_ONLY)
  public String getFullFileName() {
    return this.getExtendedRelativeLocation() + SolidifyConstants.URL_SEP + this.getFileName();
  }

  @JsonIgnore
  public String getInitialPath() {
    String categoryFolder = "";
    if (this.dataCategory != null) {
      switch (this.dataCategory) {
        case Package:
          switch (this.dataType) {
            case CustomMetadata:
              categoryFolder = Package.METADATA.getName();
              break;
            case UpdatedMetadata:
              categoryFolder = Package.UPDATE.getName();
              break;
            case Metadata:
            case InformationPackage:
              categoryFolder = SolidifyConstants.URL_SEP;
              break;
            default:
              throw new SolidifyRuntimeException(this.dataType + " is not a data type for the Package data category");
          }
          break;
        case Primary:
          categoryFolder = Package.RESEARCH.getName();
          break;
        case Secondary:
          categoryFolder = Package.DOC.getName();
          break;
        case Software:
          categoryFolder = Package.SOFTWARE.getName();
          break;
        case Internal:
          categoryFolder = Package.INTERNAL.getName();
          break;
        default:
          throw new SolidifyRuntimeException(this.dataCategory + " is not a data category");
      }
    }
    return categoryFolder;

  }

  public MetadataType getMetadataType() {
    return this.metadataType;
  }

  public String getMetadataTypeId() {
    return this.metadataTypeId;
  }

  @JsonIgnore
  public Path getPath() {
    return Paths.get(this.getFinalData());
  }

  public String getRelativeLocation() {
    return this.relativeLocation;
  }

  @JsonProperty(access = Access.READ_ONLY)
  public String getSmartSize() {
    if (this.fileSize == null) {
      return DLCMConstants.UNKNOWN;
    }
    return StringTool.formatSmartSize(this.fileSize);
  }

  public URI getSourceData() {
    return this.sourceData;
  }

  public DataFileStatus getStatus() {
    return this.status;
  }

  public String getStatusMessage() {
    return this.statusMessage;
  }

  public VirusCheck getVirusCheck() {
    return this.virusCheck;
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (this.getSourceData() != null ? this.getSourceData().hashCode() : 0);
    result = 31 * result + (this.getRelativeLocation() != null ? this.getRelativeLocation().hashCode() : 0);
    result = 31 * result + (this.getFinalData() != null ? this.getFinalData().hashCode() : 0);
    result = 31 * result + (this.getFileFormat() != null ? this.getFileFormat().hashCode() : 0);
    result = 31 * result + (this.getVirusCheck() != null ? this.getVirusCheck().hashCode() : 0);
    result = 31 * result + (this.getFileSize().hashCode());
    result = 31 * result + (this.getStatus() != null ? this.getStatus().hashCode() : 0);
    result = 31 * result + (this.getDataCategory() != null ? this.getDataCategory().hashCode() : 0);
    result = 31 * result + (this.getDataType() != null ? this.getDataType().hashCode() : 0);
    result = 31 * result + (this.getComplianceLevel() != null ? this.getComplianceLevel().hashCode() : 0);
    result = 31 * result + (this.getInfoPackage() != null ? this.getInfoPackage().hashCode() : 0);
    result = 31 * result + (this.getChecksums() != null ? this.getChecksums().hashCode() : 0);
    return result;
  }

  @Override
  @PrePersist
  public void init() {
    if (this.status == null) {
      this.status = DataFileStatus.RECEIVED;
    }
    if (this.relativeLocation == null) {
      this.relativeLocation = SolidifyConstants.URL_SEP;
    } else {
      this.checkRelativeLocation();
    }
    if (this.dataCategory == null) {
      this.dataCategory = DataCategory.Primary;
    }
    if (this.dataType == null) {
      this.dataType = DataCategory.Reference;
    }
    if (this.fileFormat == null) {
      this.fileFormat = new FileFormat();
    }
    if (this.fileSize == null) {
      this.fileSize = 0L;
    }
    if (this.complianceLevel == null) {
      this.complianceLevel = ComplianceLevel.NOT_ASSESSED;
    }

    this.check();
  }

  @JsonIgnore
  public boolean isAvailable() {
    return this.getStatus() == DataFileStatus.PROCESSED || this.getStatus() == DataFileStatus.READY;
  }

  @JsonIgnore
  public boolean isInRoot() {
    return (StringTool.isNullOrEmpty(this.getRelativeLocation()) || this.getRelativeLocation().equals(SolidifyConstants.URL_SEP));
  }

  public void setComplianceLevel(ComplianceLevel complianceLevel) {
    this.complianceLevel = complianceLevel;
  }

  public void setDataCategory(DataCategory dataCat) {
    this.dataCategory = dataCat;
  }

  public void setDataType(DataCategory dataType) {
    this.dataType = dataType;
  }

  public void setFileFormat(FileFormat fileFormat) {
    this.fileFormat = fileFormat;
  }

  public void setFileSize(Long size) {
    this.fileSize = size;
  }

  @SuppressWarnings("squid:S2637")
  public void setFinalData(URI finalData) {
    this.finalData = finalData;
    this.fileName = this.computeFileName();
  }

  public void setMetadataType(MetadataType metadataType) {
    this.metadataType = metadataType;
    if (this.metadataType != null && this.metadataType.getResId() != null) {
      this.setMetadataTypeId(this.metadataType.getResId());
    }
  }

  public void setPath(Path file) {
    // Do nothing
  }

  public void setRelativeLocation(String relativeLocation) {
    this.relativeLocation = relativeLocation;
  }

  @SuppressWarnings("squid:S2637")
  public void setSourceData(URI sourceData) {
    this.sourceData = sourceData;
    this.fileName = this.computeFileName();
  }

  /*
   * ************************************* Functional methods
   **************************************/

  public void setStatus(DataFileStatus status) {
    this.status = status;
  }

  public void setStatusMessage(String message) {
    this.statusMessage = StringTool.truncateWithEllipsis(message, DLCMConstants.DB_LONG_STRING_LENGTH);
  }

  public void setStatusWithMessage(DataFileStatus status, String message) {
    this.setStatus(status);
    this.setStatusMessage(message);
  }

  public void setVirusCheck(VirusCheck virusCheck) {
    this.virusCheck = virusCheck;
  }

  private void checkRelativeLocation() {
    // Check if root
    if (!this.relativeLocation.equals(SolidifyConstants.URL_SEP)) {
      // Check starting / & add it if needed
      if (!this.relativeLocation.startsWith(SolidifyConstants.URL_SEP)) {
        this.setRelativeLocation(SolidifyConstants.URL_SEP + this.relativeLocation);
      }
      // Check ending / & remove it if needed
      if (this.relativeLocation.endsWith(SolidifyConstants.URL_SEP)) {
        this.setRelativeLocation(this.relativeLocation.substring(0, this.relativeLocation.length() - 1));
      }
    }
  }

  private String getExtendedRelativeLocation() {
    if (this.isInRoot()) {
      return "";
    }
    return this.getRelativeLocation();
  }

  private void setMetadataTypeId(String metadataTypeId) {
    this.metadataTypeId = metadataTypeId;
  }

  private String computeFileName() {
    URI fileURI = this.finalData;
    if (fileURI == null) {
      fileURI = this.sourceData;
    }

    if (fileURI != null) {
      final int lastIndexOfSlash = fileURI.toString().lastIndexOf('/');
      return StringTool.decodeUrl(fileURI.toString().substring(lastIndexOfSlash + 1));
    }
    return null;
  }

  @Override
  public Specification<V> getSearchSpecification(SearchCriteria criteria) {
    throw new UnsupportedOperationException("Searchable not implemented for datafile of type " + this.getClass().getSimpleName());
  }
}
