/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - FileFormat.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model;

import static ch.dlcm.DLCMConstants.DB_DEFAULT_STRING_LENGTH;

import java.io.StringReader;
import java.util.Objects;
import javax.persistence.AttributeOverride;
import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.Embedded;
import javax.persistence.Lob;
import javax.validation.Valid;
import javax.validation.constraints.Size;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;

@Embeddable
public class FileFormat implements ToolContainer {

  private static final int DB_MD5_LENGTH = 256;
  private static final int DB_VERSION_LENGTH = 50;
  private static final int DB_PUID_LENGTH = 20;

  @Size(max = DLCMConstants.DB_MEDIUM_STRING_LENGTH)
  @Column(length = DLCMConstants.DB_MEDIUM_STRING_LENGTH)
  private String contentType;

  @Lob
  @Column(name = "formatDetails")
  private String details;

  private String format;

  @Size(max = DB_MD5_LENGTH)
  @Column(length = DB_MD5_LENGTH)
  private String md5;

  @Size(max = DB_PUID_LENGTH)
  @Column(length = DB_PUID_LENGTH)
  private String puid;

  @Embedded
  @Valid
  @AttributeOverride(name = "name", column = @Column(name = "formatToolName"))
  @AttributeOverride(name = "description", column = @Column(name = "formatToolDescription"))
  @AttributeOverride(name = "version", column = @Column(name = "formatToolVersion"))
  @AttributeOverride(name = "puid", column = @Column(name = "formatToolPuid"))
  private Tool tool;

  @Size(max = DB_VERSION_LENGTH)
  @Column(name = "formatVersion", length = DB_VERSION_LENGTH)
  private String version;

  @Override
  public boolean equals(final Object other) {
    if (this == other) {
      return true;
    }
    if (!(other instanceof FileFormat)) {
      return false;
    }
    final FileFormat castOther = (FileFormat) other;
    return Objects.equals(this.getContentType(), castOther.getContentType()) && Objects.equals(this.getFormat(), castOther.getFormat())
            && Objects
                    .equals(this.getVersion(), castOther.getVersion())
            && Objects.equals(this.getPuid(), castOther.getPuid()) && Objects
                    .equals(this.getMd5(), castOther.getMd5())
            && Objects.equals(this.getDetails(), castOther.getDetails()) && Objects
                    .equals(this.getTool(), castOther.getTool());
  }

  public String getContentType() {
    return this.contentType;
  }

  public String getDetails() {
    return this.details;
  }

  public String getFormat() {
    return this.format;
  }

  public String getMd5() {
    return this.md5;
  }

  public String getPuid() {
    return this.puid;
  }

  @JsonIgnore
  public Object getSmartDetails(JAXBContext jaxbContext) throws JAXBException {
    final Unmarshaller um = jaxbContext.createUnmarshaller();
    return um.unmarshal(new StringReader(this.getDetails()));
  }

  @Override
  public Tool getTool() {
    return this.tool;
  }

  public String getVersion() {
    return this.version;
  }

  @Override
  public int hashCode() {
    return Objects.hash(this.getContentType(), this.getFormat(), this.getVersion(), this.getPuid(), this.getMd5(), this.getDetails(),
            this.getTool());
  }

  public void setContentType(String contentType) {
    this.contentType = contentType;
  }

  public void setDetails(String details) {
    if (details != null) {
      this.details = details;
    }
  }

  public void setFormat(String format) {
    if (format != null && format.length() > DB_DEFAULT_STRING_LENGTH) {
      this.format = StringTool.truncateWithEllipsis(format, DB_DEFAULT_STRING_LENGTH);
    } else {
      this.format = format;
    }
  }

  public void setMd5(String md5) {
    this.md5 = md5;
  }

  public void setPuid(String puid) {
    this.puid = puid;
  }

  @Override
  public void setTool(Tool tool) {
    this.tool = tool;
  }

  public void setVersion(String version) {
    this.version = version;
  }

}
