/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - Notification.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.notification;

import static ch.unige.solidify.SolidifyConstants.DB_DATE_LENGTH;

import java.time.OffsetDateTime;
import java.util.Objects;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.ManyToOne;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonFormat;

import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.rest.ModuleName;

@Entity
@EntityListeners(HistoryListener.class)
public class Notification extends ResourceNormalized {

  @ManyToOne
  private Person recipient;

  @NotNull
  @ManyToOne
  private User emitter;

  @NotNull
  @Size(min = 1, max = DLCMConstants.DB_LARGE_STRING_LENGTH)
  private String message;

  @NotNull
  @Enumerated(EnumType.STRING)
  private NotificationStatus notificationStatus;

  @NotNull
  @ManyToOne
  private NotificationType notificationType;

  @ManyToOne
  private OrganizationalUnit notifiedOrgUnit;

  private String objectId;

  @Size(max = DLCMConstants.DB_LONG_STRING_LENGTH)
  private String statusMessage;

  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_TIME_FORMAT)
  @Column(length = DB_DATE_LENGTH)
  private OffsetDateTime sentTime;

  @Override
  public boolean equals(final Object o) {
    if (this == o) {
      return true;
    }
    if (!(o instanceof Notification)) {
      return false;
    }
    if (!super.equals(o)) {
      return false;
    }
    final Notification that = (Notification) o;
    return this.notificationType == that.notificationType &&
            this.notificationStatus == that.notificationStatus &&
            Objects.equals(this.emitter, that.emitter) &&
            Objects.equals(this.notifiedOrgUnit, that.notifiedOrgUnit) &&
            Objects.equals(this.message, that.message) &&
            Objects.equals(this.recipient, that.recipient);
  }

  public User getEmitter() {
    return this.emitter;
  }

  public String getMessage() {
    return this.message;
  }

  public NotificationStatus getNotificationStatus() {
    return this.notificationStatus;
  }

  public NotificationType getNotificationType() {
    return this.notificationType;
  }

  public OrganizationalUnit getNotifiedOrgUnit() {
    return this.notifiedOrgUnit;
  }

  public OffsetDateTime getSentTime() {
    return this.sentTime;
  }

  public Person getRecipient() {
    return this.recipient;
  }

  public String getStatusMessage() {
    return this.statusMessage;
  }

  public String getObjectId() {
    return this.objectId;
  }

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), this.notificationType, this.notificationStatus, this.emitter, this.notifiedOrgUnit, this.message);
  }

  @Override
  public void init() {
    if (this.notificationStatus == null) {
      this.notificationStatus = NotificationStatus.PENDING;
      this.statusMessage = null;
    }
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  public void setEmitter(User emitter) {
    this.emitter = emitter;
  }

  public void setRecipient(Person recipient) {
    this.recipient = recipient;
  }

  public void setMessage(String message) {
    this.message = message;
  }

  public void setNotificationStatus(NotificationStatus notificationStatus) {
    this.notificationStatus = notificationStatus;
  }

  public void setNotificationType(NotificationType notificationType) {
    this.notificationType = notificationType;
  }

  public void setNotifiedOrgUnit(OrganizationalUnit notifiedOrgUnit) {
    this.notifiedOrgUnit = notifiedOrgUnit;
  }

  public void setStatusMessage(String message) {
    this.statusMessage = StringTool.truncateWithEllipsis(message, DLCMConstants.DB_LONG_STRING_LENGTH);
  }

  public void setObjectId(String objectId) {
    this.objectId = objectId;
  }

  public void setSentTime(OffsetDateTime sentTime) {
    this.sentTime = sentTime;
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.HISTORY).withRel(ActionName.HISTORY));
    }
  }
}
