/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - DisseminationInfoPackage.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.oais;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import javax.persistence.CascadeType;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.OneToMany;
import javax.persistence.PrePersist;
import javax.persistence.Table;
import javax.persistence.UniqueConstraint;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.SubResourceContainer;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.listener.RegistryListener;
import ch.dlcm.model.AbstractDataFile.DataFileStatus;
import ch.dlcm.model.DataFileOwnerInterface;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.access.Order;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

@Entity
@Table(name = "dip")
@EntityListeners({ HistoryListener.class, RegistryListener.class })
public class DisseminationInfoPackage extends ResourceNormalized
        implements SubResourceContainer, InfoPackageInterface<DipDataFile>, DataFileOwnerInterface<DipDataFile>, OrganizationalUnitAwareResource,
        MetadataVersionAware {

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "dipAip", joinColumns = { @JoinColumn(name = DLCMConstants.DB_PACKAGE_ID) }, inverseJoinColumns = {
          @JoinColumn(name = DLCMConstants.DB_AIP_ID) }, uniqueConstraints = {
          @UniqueConstraint(columnNames = { DLCMConstants.DB_PACKAGE_ID, DLCMConstants.DB_AIP_ID }) })
  private List<ArchivalInfoPackage> aips = new ArrayList<>();

  @JsonIgnore
  @OneToMany(mappedBy = "infoPackage", cascade = CascadeType.ALL, orphanRemoval = true)
  private List<DipDataFile> dataFiles = new ArrayList<>();

  @NotNull
  @Embedded
  @Valid
  private RepresentationInfo info = new RepresentationInfo();

  @ManyToMany(mappedBy = "dipPackages")
  @JsonIgnore
  private List<Order> orders;

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof DipDataFile) {
      return this.addDataFile((DipDataFile) t);
    }
    if (t instanceof ArchivalInfoPackage) {
      return this.addAIP((ArchivalInfoPackage) t);
    }
    return false;
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.DATAFILE).withRel(ResourceName.DATAFILE));
      if (this.getInfo().getStatus() != PackageStatus.IN_PREPARATION) {
        this.add(linkBuilder.slash(this.getResId()).slash(ActionName.DOWNLOAD).withRel(ActionName.DOWNLOAD));
      }
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.HISTORY).withRel(ActionName.HISTORY));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.AIP).withRel(ResourceName.AIP));
    }
  }

  @Override
  public boolean equals(final Object other) {
    if (!(other instanceof DisseminationInfoPackage)) {
      return false;
    }
    final DisseminationInfoPackage castOther = (DisseminationInfoPackage) other;
    return Objects.equals(this.info, castOther.info) && Objects.equals(this.dataFiles, castOther.dataFiles)
            && Objects.equals(this.aips, castOther.aips);
  }

  public List<String> getAipIds() {
    return this.aips.stream().map(ArchivalInfoPackage::getResId).collect(Collectors.toList());
  }

  @SuppressWarnings("unchecked")
  @Override
  public List<DipDataFile> getDataFiles() {
    return this.dataFiles;
  }

  @Override
  public URI getDownloadUri() {
    if (this.getInfo().getStatus() == PackageStatus.IN_PREPARATION) {
      throw new SolidifyRuntimeException("Package not available");
    }
    if (this.getDataFiles().isEmpty()) {
      throw new SolidifyRuntimeException("Package not accessible");
    }
    return this.getDataFiles().get(0).getFinalData();
  }

  @Override
  public RepresentationInfo getInfo() {
    return this.info;
  }

  @Override
  @JsonIgnore
  public String getOrganizationalUnitId() {
    return this.info.getOrganizationalUnitId();
  }

  @Override
  public int hashCode() {
    return Objects.hash(this.info, this.dataFiles, this.aips);
  }

  @Override
  @PrePersist
  public void init() {
    this.info.init();
  }

  @Override
  public boolean isDeletable() {
    if (this.getInfo().getStatus() != PackageStatus.COMPLETED) {
      return false;
    }
    return true;
  }

  @Override
  @JsonIgnore
  public boolean isIndexable() {
    return false;
  }

  @Override
  public boolean isModifiable() {
    return false;
  }

  @Override
  public boolean isReady() {
    if (this.getDataFiles().isEmpty()) {
      return false;
    }
    for (final DipDataFile df : this.getDataFiles()) {
      if (df.getStatus() != DataFileStatus.READY) {
        return false;
      }
    }
    return true;
  }

  @Override
  public String managedBy() {
    return ModuleName.ACCESS;
  }

  @Override
  public <T> boolean removeItem(T t) {
    if (t instanceof DipDataFile) {
      return this.removeDataFile((DipDataFile) t);
    }
    if (t instanceof ArchivalInfoPackage) {
      return this.removeAIP((ArchivalInfoPackage) t);
    }
    return false;
  }

  public void setDataFiles(List<DipDataFile> dataFiles) {
    this.dataFiles = dataFiles;
  }

  public void setInfo(RepresentationInfo info) {
    this.info = info;
  }

  @Override
  public boolean addDataFile(DipDataFile t) {
    return this.dataFiles.add(t);
  }

  private boolean addAIP(ArchivalInfoPackage t) {
    return this.getAips().add(t);
  }

  private boolean removeDataFile(DipDataFile t) {
    return this.dataFiles.remove(t);
  }

  private boolean removeAIP(ArchivalInfoPackage t) {
    return this.getAips().remove(t);
  }

  public List<ArchivalInfoPackage> getAips() {
    return this.aips;
  }

  @JsonIgnore
  public List<String> getAipsIds() {
    return this.aips.stream().map(ArchivalInfoPackage::getResId).collect(Collectors.toList());
  }

  public void setAips(List<ArchivalInfoPackage> aips) {
    this.aips = aips;
  }

  public List<Order> getOrders() {
    return this.orders;
  }

  @JsonIgnore
  @Override
  public DLCMMetadataVersion getMetadataVersion() {
    return this.info.getMetadataVersion();
  }
}
