/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - DisseminationPolicy.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.policies;

import java.util.ArrayList;
import java.util.List;
import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.OneToMany;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import ch.unige.solidify.exception.SolidifyUndeletableException;
import ch.unige.solidify.rest.ResourceNormalized;

import ch.dlcm.DLCMConstants;
import ch.dlcm.rest.ModuleName;

@Entity
public class DisseminationPolicy extends ResourceNormalized {

  public enum DisseminationPolicyType {
    IIIF, WEB
  }

  @JsonIgnore
  @OneToMany(mappedBy = OrganizationalUnitDisseminationPolicy.PATH_TO_DISSEMINATION_POLICY, cascade = CascadeType.ALL, orphanRemoval = true)
  protected List<OrganizationalUnitDisseminationPolicy> organizationalUnits = new ArrayList<>();

  @NotNull
  @Size(min = 1, max = DLCMConstants.DB_LONG_STRING_LENGTH)
  private String destinationServer;

  @NotNull
  @Size(min = 1)
  private String name;

  private String subFolder;

  @Enumerated(EnumType.STRING)
  private DisseminationPolicyType type;

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (!super.equals(obj)) {
      return false;
    }
    if (this.getClass() != obj.getClass()) {
      return false;
    }
    final DisseminationPolicy other = (DisseminationPolicy) obj;
    if (this.getName() != null ? !this.getName().equals(other.getName()) : other.getName() != null) {
      return false;
    }
    if (this.getType() != null ? !this.getType().equals(other.getType()) : other.getType() != null) {
      return false;
    }
    if (this.getDestinationServer() != null ? !this.getDestinationServer().equals(other.getDestinationServer())
            : other.getDestinationServer() != null) {
      return false;
    }
    if (this.getSubFolder() != null ? !this.getSubFolder().equals(other.getSubFolder()) : other.getSubFolder() != null) {
      return false;
    }
    return true;
  }

  public String getDestinationServer() {
    return this.destinationServer;
  }

  public String getName() {
    return this.name;
  }

  public String getSubFolder() {
    return this.subFolder;
  }

  public DisseminationPolicyType getType() {
    return this.type;
  }

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public int getUseNumber() {
    return this.organizationalUnits.size();
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (this.getName() != null ? this.getName().hashCode() : 0);
    result = 31 * result + (this.getType() != null ? this.getType().hashCode() : 0);
    result = 31 * result + (this.getDestinationServer() != null ? this.getDestinationServer().hashCode() : 0);
    result = 31 * result + (this.getSubFolder() != null ? this.getSubFolder().hashCode() : 0);

    return result;
  }

  @PrePersist
  @PreUpdate
  @Override
  public void init() {
    if (this.subFolder == null) {
      this.subFolder = "/";
    }
    if (this.type == null) {
      this.type = DisseminationPolicyType.WEB;
    }
  }

  @Override
  public boolean isDeletable() {
    final int nbOrgUnits = this.getUseNumber();
    if (nbOrgUnits > 0) {
      throw new SolidifyUndeletableException(
              "The dissemination policy " + this.getResId() + " is used in " + nbOrgUnits + " organizational units");
    }
    return true;
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  public void setDestinationServer(String destinationServer) {
    this.destinationServer = destinationServer;
  }

  public void setName(String name) {
    this.name = name;
  }

  public void setSubFolder(String subFolder) {
    this.subFolder = subFolder;
  }

  public void setType(DisseminationPolicyType type) {
    this.type = type;
  }

  @Override
  public String toString() {
    return super.toString() + " (name=" + this.getName() + ")";
  }

}
