/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - PreservationJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.preservation;

import java.beans.Transient;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Objects;
import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.OneToMany;
import javax.persistence.PrePersist;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.SubResourceContainer;

import ch.dlcm.listener.RegistryListener;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

@Entity
@EntityListeners({ RegistryListener.class })
public class PreservationJob extends ResourceNormalized
        implements SubResourceContainer {

  public enum JobStatus {
    COMPLETED, CREATED, IN_ERROR, IN_PROGRESS, READY
  }

  private Boolean enable;

  @JsonIgnore
  @OneToMany(mappedBy = "preservationJob", cascade = CascadeType.ALL, orphanRemoval = true)
  private List<JobExecution> executions = new ArrayList<>();

  @NotNull
  @Enumerated(EnumType.STRING)
  private JobRecurrence jobRecurrence;

  @NotNull
  @Enumerated(EnumType.STRING)
  private JobType jobType;

  @NotNull
  @Size(min = 1)
  private String name;

  private JobScheduling scheduling;

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof JobExecution) {
      return this.addJobExecution((JobExecution) t);
    }
    return false;
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      if (this.isEnable()) {
        if (this.getLastExecutionStatus() == JobStatus.IN_ERROR) {
          this.add(linkBuilder.slash(this.getResId()).slash(ActionName.RESUME)
                  .withRel(ActionName.RESUME));
          if (this.getJobRecurrence() != JobRecurrence.ONCE) {
            this.add(linkBuilder.slash(this.getResId()).slash(ActionName.START)
                    .withRel(ActionName.START));
          }
        } else {
          this.add(linkBuilder.slash(this.getResId()).slash(ActionName.START).withRel(ActionName.START));
        }
      }
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.PRES_JOB_EXECUTION)
              .withRel(ResourceName.PRES_JOB_EXECUTION));
    }
  }

  public Boolean getEnable() {
    return this.enable;
  }

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public int getExecutionNumber() {
    return this.getExecutions().size();
  }

  public List<JobExecution> getExecutions() {
    return this.executions;
  }

  public JobRecurrence getJobRecurrence() {
    return this.jobRecurrence;
  }

  public JobType getJobType() {
    return this.jobType;
  }

  @Transient
  @JsonIgnore
  public JobExecution getLastExecution() {
    if (this.getExecutions().isEmpty()) {
      return null;
    }
    return this.getExecutions().stream().max(Comparator.comparing(JobExecution::getRunNumber))
            .orElseThrow(NoSuchElementException::new);
  }

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public JobStatus getLastExecutionStatus() {
    if (this.getLastExecution() != null && !this.getExecutions().isEmpty()) {
      return this.getLastExecution().getStatus();
    }
    return null;
  }

  public String getName() {
    return this.name;
  }

  public JobScheduling getScheduling() {
    return this.scheduling;
  }

  @PrePersist
  @Override
  public void init() {
    if (this.jobRecurrence == null) {
      this.jobRecurrence = JobRecurrence.ONCE;
    }
    if (this.enable == null) {
      this.enable = true;
    }
  }

  public boolean isEnable() {
    return this.enable;
  }

  @Override
  public String managedBy() {
    return ModuleName.PRES_PLANNING;
  }

  @Override
  public <T> boolean removeItem(T t) {
    if (t instanceof JobExecution) {
      return this.removeJobExecution((JobExecution) t);
    }
    return false;
  }

  public boolean resume() {
    if (this.getExecutions().isEmpty()) {
      return this.addItem(this.newJobExecution());
    }
    if (this.getLastExecutionStatus().equals(JobStatus.IN_ERROR)) {
      this.getLastExecution().setStatus(JobStatus.READY);
      this.getLastExecution().setStatusMessage(null);
      return true;
    }
    return false;
  }

  public void setEnable(Boolean enable) {
    this.enable = enable;
  }

  public void setJobRecurrence(JobRecurrence jobRecurrence) {
    this.jobRecurrence = jobRecurrence;
  }

  public void setJobType(JobType jobType) {
    this.jobType = jobType;
  }

  public void setName(String name) {
    this.name = name;
  }

  public void setScheduling(JobScheduling scheduling) {
    this.scheduling = scheduling;
  }

  public boolean start() {
    if (this.getExecutions().isEmpty()) {
      return this.addItem(this.newJobExecution());
    }
    switch (this.getLastExecutionStatus()) {
      case COMPLETED:
      case IN_ERROR:
        if (this.getJobRecurrence() != JobRecurrence.ONCE && this.getJobType().isRecurring()) {
          return this.addItem(this.newJobExecution());
        }
        break;
      default:
        break;
    }
    return false;
  }

  private boolean addJobExecution(JobExecution e) {
    this.getExecutions().add(e);
    return true;
  }

  private JobExecution newJobExecution() {
    final JobExecution exe = new JobExecution();
    exe.setPreservationJob(this);
    exe.setStatus(JobStatus.READY);
    exe.setRunNumber(this.getExecutionNumber() + 1L);
    return exe;
  }

  private boolean removeJobExecution(JobExecution e) {
    return this.getExecutions().remove(e);
  }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (!(o instanceof PreservationJob))
      return false;
    if (!super.equals(o))
      return false;
    PreservationJob that = (PreservationJob) o;
    return Objects.equals(this.enable, that.enable) &&
            Objects.equals(this.executions, that.executions) &&
            this.jobRecurrence == that.jobRecurrence &&
            this.jobType == that.jobType &&
            Objects.equals(this.name, that.name) &&
            Objects.equals(this.scheduling, that.scheduling);
  }

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), this.enable, this.executions, this.jobRecurrence, this.jobType, this.name, this.scheduling);
  }
}
