/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - Person.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.settings;

import java.util.ArrayList;
import java.util.List;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.UniqueConstraint;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;
import javax.validation.constraints.Size;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.unige.solidify.auth.model.OrcidInfo;
import ch.unige.solidify.model.PersonWithOrcid;
import ch.unige.solidify.rest.SearchableResourceNormalized;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.util.SearchCriteria;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.IdentifierType;
import ch.dlcm.model.Image;
import ch.dlcm.model.ImageInterface;
import ch.dlcm.model.notification.NotificationType;
import ch.dlcm.model.security.Role;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.specification.PersonSearchSpecification;

@Entity
public class Person extends SearchableResourceNormalized<Person> implements PersonWithOrcid, SubResourceContainer, ImageInterface {

  @JsonIgnore
  @OneToMany(mappedBy = OrganizationalUnitPersonRole.PATH_TO_PERSON, cascade = CascadeType.ALL, orphanRemoval = true)
  protected List<OrganizationalUnitPersonRole> organizationalUnitRoles = new ArrayList<>();

  @NotNull
  @Size(min = 1)
  private String firstName;

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "personInstitutions", joinColumns = { @JoinColumn(name = DLCMConstants.DB_PERSON_ID) }, inverseJoinColumns = {
          @JoinColumn(name = DLCMConstants.DB_INSTITUTION_ID) }, uniqueConstraints = {
                  @UniqueConstraint(columnNames = { DLCMConstants.DB_PERSON_ID, DLCMConstants.DB_INSTITUTION_ID }) })
  private List<Institution> institutions = new ArrayList<>();

  @NotNull
  @Size(min = 1)
  private String lastName;

  @Pattern(regexp = OrcidInfo.ORCID_PATTERN, message = OrcidInfo.ORCID_MESSAGE)
  @Column(unique = true)
  private String orcid;

  @NotNull
  private Boolean verifiedOrcid = false;

  @OneToOne(cascade = CascadeType.ALL, orphanRemoval = true)
  @JoinColumn(name = DLCMConstants.DB_AVATAR_ID, referencedColumnName = DLCMConstants.DB_RES_ID)
  private PersonAvatar avatar;

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "personNotificationTypes", joinColumns = { @JoinColumn(name = DLCMConstants.DB_PERSON_ID) }, inverseJoinColumns = {
          @JoinColumn(name = DLCMConstants.DB_NOTIFICATION_TYPE_ID) }, uniqueConstraints = {
                  @UniqueConstraint(columnNames = { DLCMConstants.DB_PERSON_ID, DLCMConstants.DB_NOTIFICATION_TYPE_ID }) })
  private List<NotificationType> subscribedNotificationTypes = new ArrayList<>();

  public boolean addInstitution(Institution institution) {
    final boolean result = this.institutions.add(institution);
    if (!institution.getPeople().contains(this)) {
      institution.addPerson(this);
    }
    return result;
  }

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof Institution) {
      return this.addInstitution((Institution) t);
    }
    if (t instanceof NotificationType) {
      return this.addNotificationType((NotificationType) t);
    }
    return false;
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.INSTITUTION).withRel(ResourceName.INSTITUTION));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.ORG_UNIT).withRel(ResourceName.ORG_UNIT));
      this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.DOWNLOAD_AVATAR).withRel(DLCMActionName.DOWNLOAD_AVATAR));
      this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.UPLOAD_AVATAR).withRel(DLCMActionName.UPLOAD_AVATAR));
      if (!StringTool.isNullOrEmpty(this.getOrcid())) {
        this.add(
                Tool.filter(
                        linkBuilder.slash(this.getOrcid()).toUriComponentsBuilder(),
                        DLCMConstants.IDENTIFIER_TYPE_PARAM,
                        IdentifierType.ORCID.toString())
                        .withRel(IdentifierType.ORCID.toString().toLowerCase()));
      }
    }
  }

  public boolean addOrganizationalUnit(OrganizationalUnit organizationalUnit, List<Role> roles) {
    for (final Role role : roles) {
      if (!this.addOrganizationalUnit(organizationalUnit, role)) {
        return false;
      }
    }
    return true;
  }

  @JsonIgnore
  public String getDefault() {
    return Role.DEFAULT_ROLE_ID;
  }

  /************************************************************/

  public String getFirstName() {
    return this.firstName;
  }

  public String getFullName() {
    return this.getLastName() + ", " + this.getFirstName();
  }

  public List<Institution> getInstitutions() {
    return this.institutions;
  }

  public String getLastName() {
    return this.lastName;
  }

  @Override
  public String getOrcid() {
    return this.orcid;
  }

  public List<OrganizationalUnitPersonRole> getOrganizationalUnitRoles() {
    return this.organizationalUnitRoles;
  }

  public PersonAvatar getAvatar() {
    return this.avatar;
  }

  @JsonIgnore
  public List<NotificationType> getSubscribedNotifications() {
    return this.subscribedNotificationTypes;
  }

  @Override
  public void init() {
    // Do nothing
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  public boolean removeInstitution(Institution institution) {
    final boolean result = this.institutions.remove(institution);
    if (institution.getPeople().contains(this)) {
      institution.removePerson(this);
    }
    return result;
  }

  @Override
  public <T> boolean removeItem(T t) {
    if (t instanceof Institution) {
      return this.removeInstitution((Institution) t);
    }
    if (t instanceof OrganizationalUnit) {
      return this.removeOrganizationalUnit((OrganizationalUnit) t);
    }
    if (t instanceof OrganizationalUnitPersonRole) {
      return this.removeOrganizationalUnitRole((OrganizationalUnitPersonRole) t);
    }
    if (t instanceof NotificationType) {
      return this.removeNotificationType((NotificationType) t);
    }
    return false;
  }

  public void setFirstName(String firstName) {
    this.firstName = firstName;
  }

  public void setLastName(String lastName) {
    this.lastName = lastName;
  }

  @Override
  public void setOrcid(String orcid) {
    this.orcid = orcid;
  }

  public void setAvatar(PersonAvatar avatar) {
    this.avatar = avatar;
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    builder.append("Person [resId=").append(this.getResId())
            .append(", firstName=").append(this.firstName)
            .append(", lastName=").append(this.lastName)
            .append(", orcid=").append(this.orcid).append("]");
    return builder.toString();
  }

  protected boolean removeOrganizationalUnit(OrganizationalUnit unit) {
    final OrganizationalUnit organizationalUnit = unit;
    return this.organizationalUnitRoles.removeIf(ouprPredicate -> ouprPredicate.getCompositeKey()
            .getOrganizationalUnit() == organizationalUnit);
  }

  private boolean addOrganizationalUnit(OrganizationalUnit organizationalUnit, Role role) {
    final OrganizationalUnitPersonRole organizationalUnitPersonRole = new OrganizationalUnitPersonRole();
    organizationalUnitPersonRole.setRole(role);
    organizationalUnitPersonRole.setOrganizationalUnit(organizationalUnit);
    organizationalUnitPersonRole.setPerson(this);
    return this.organizationalUnitRoles.add(organizationalUnitPersonRole);
  }

  private boolean removeOrganizationalUnitRole(OrganizationalUnitPersonRole unitPersonRole) {
    return this.organizationalUnitRoles.removeIf(ouprPredicate -> ouprPredicate.equals(unitPersonRole));
  }

  public boolean addNotificationType(NotificationType notificationType) {
    final boolean result = this.subscribedNotificationTypes.add(notificationType);
    if (!notificationType.getSubscribers().contains(this)) {
      notificationType.addSubscriber(this);
    }
    return result;
  }

  public boolean removeNotificationType(NotificationType notificationType) {
    final boolean result = this.subscribedNotificationTypes.remove(notificationType);
    if (notificationType.getSubscribers().contains(this)) {
      notificationType.removeSubscriber(this);
    }
    return result;
  }

  @Override
  @JsonIgnore
  public Image getNewIfNeededImage() {
    if (this.getAvatar() == null) {
      this.setAvatar(new PersonAvatar());
    }
    return this.getAvatar();
  }

  @Override
  @JsonIgnore
  public Image getImage() {
    return this.getAvatar();
  }

  @Override
  public void setImage(Image logo) {
    this.setAvatar((PersonAvatar) logo);
  }

  @Override
  public Specification<Person> getSearchSpecification(SearchCriteria criteria) {
    return new PersonSearchSpecification(criteria);
  }

  @Override
  public Boolean isVerifiedOrcid() {
    return this.verifiedOrcid;
  }

  @Override
  public void setVerifiedOrcid(Boolean verifiedOrcid) {
    this.verifiedOrcid = verifiedOrcid;
  }
}
