/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - MetadataTest.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.tool;

import static org.junit.jupiter.api.Assertions.assertNotNull;

import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.util.XMLTool;

import ch.dlcm.DLCMMetadataVersion;

@ExtendWith(SpringExtension.class)
class MetadataTest extends AbstractXMLTest {
  private final String KO_SUFFIX = "-KO.xml";
  private final String OK_SUFFIX = "-OK.xml";

  private final String[] versions = { "1.0", "1.1", "2.0" };

  @Test
  void validationV11Test() throws SolidifyCheckingException {
    this.testFiles("1.1");
  }

  @Test
  void validationV1Test() throws SolidifyCheckingException {
    this.testFiles("1.0");
  }

  @Test
  void validationV2Test() throws SolidifyCheckingException {
    this.testFiles("2.0");
  }

  @Test
  void wellformedSchemasTest() throws URISyntaxException, SolidifyCheckingException {
    for (final String version : this.versions) {
      final List<URL> xsdList = this.loadSchema(DLCMMetadataVersion.fromVersion(version).getMetsSchema());
      try {
        for (final URL xsd : xsdList) {
          assertNotNull(Paths.get(xsd.toURI()));
          XMLTool.wellformed(Paths.get(xsd.toURI()));
        }
      } catch (final SolidifyCheckingException e) {
        this.returnException(e);
      }
    }
  }

  @Test
  void wellFormedTest() throws SolidifyCheckingException {
    for (final String version : this.versions) {
      for (final Path file : this.loadFilesTest(version)) {
        try {
          XMLTool.wellformed(file);
        } catch (final SolidifyCheckingException e) {
          this.returnException(e);
        }
      }
    }
  }

  private List<Path> loadFilesTest(String version) {
    final List<Path> list = new ArrayList<>();
    list.add(this.getPath("metadata/dlcm_mets-" + version + this.OK_SUFFIX));
    list.add(this.getPath("metadata/dlcm_mets-" + version + this.KO_SUFFIX));
    return list;
  }

  private void testFiles(String version) {
    final List<URL> xsdList = this.loadSchema(DLCMMetadataVersion.fromVersion(version).getMetsSchema());
    for (final Path file : this.loadFilesTest(version)) {
      try {
        XMLTool.validate(xsdList, file);
      } catch (final SolidifyCheckingException e) {
        if (!file.getFileName().toString().endsWith(this.KO_SUFFIX)) {
          this.returnException(e);
        }
      }
    }
  }
}
