/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Preservation Planning - AbstractJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.job;

import org.springframework.data.domain.PageRequest;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.JobProcessingService;

public abstract class AbstractJob {

  protected DLCMProperties configService;
  protected JobProcessingService jobProcessingService;
  protected MessageService messageService;

  protected AbstractJob(DLCMProperties configService, MessageService messageService, JobProcessingService jobProcessingService) {
    super();
    this.configService = configService;
    this.jobProcessingService = jobProcessingService;
    this.messageService = messageService;
  }

  public void run(String jobId, String executionId) {
    // Save starting info
    PreservationJob job;
    if (executionId == null) {
      job = this.jobProcessingService.createNewExecution(jobId);
      executionId = job.getLastExecution().getResId();
    } else {
      job = this.jobProcessingService.resumeExecution(jobId, executionId);
    }
    try {
      this.execute(job, executionId);
      // Save ending info
      this.jobProcessingService.completeJob(job, executionId);
    } catch (final Exception e) {
      // Save ending info
      this.jobProcessingService.completeInErrorJob(job, executionId, e);
    }
  }

  protected PreservationJob addIgnoredItem(PreservationJob job, String executionId, String aipId, String url, String errorMessage) {
    return this.jobProcessingService.addIgnoredItem(job, executionId, aipId, url, errorMessage);
  }

  protected PreservationJob addInErrorItem(PreservationJob job, String executionId, String aipId, String url, String errorMessage) {
    return this.jobProcessingService.addInErrorItem(job, executionId, aipId, url, errorMessage);
  }

  protected PreservationJob addProcessedItem(PreservationJob job, String executionId, String aipId, String url) {
    return this.jobProcessingService.addProcessedItem(job, executionId, aipId, url);
  }

  protected abstract void execute(PreservationJob job, String executionId);

  protected PageRequest getPageRequest() {
    return PageRequest.of(0, this.configService.getPreservationPlanning().getJobPageSize());
  }

  protected PageRequest getPageRequestForTotal() {
    return PageRequest.of(0, 1);
  }

  protected String getSipPublicUrl(String sipId) {
    return this.configService.getModule().getIngest().getPublicUrl() + SolidifyConstants.URL_SEP + ResourceName.SIP + SolidifyConstants.URL_SEP
            + sipId;
  }

  protected String getDepositPublicUrl(String depositId) {
    return this.configService.getModule().getPreingest().getPublicUrl() + SolidifyConstants.URL_SEP + ResourceName.DEPOSIT
            + SolidifyConstants.URL_SEP + depositId;
  }

  protected String getAipPublicUrl(String storageUrl, String aipId) {
    String publicStorageUrl = this.configService.getPublicArchivalStorageUrlFromInternalUrl(storageUrl);
    return publicStorageUrl + SolidifyConstants.URL_SEP + ResourceName.AIP + SolidifyConstants.URL_SEP + aipId;
  }

  protected String getOrderPublicUrl(String orderId) {
    return this.configService.getModule().getAccess().getPublicUrl() + SolidifyConstants.URL_SEP + ResourceName.ORDER
            + SolidifyConstants.URL_SEP + orderId;
  }

  protected String getDIPPublicUrl(String orderId) {
    return this.configService.getModule().getAccess().getPublicUrl() + SolidifyConstants.URL_SEP + ResourceName.DIP
            + SolidifyConstants.URL_SEP + orderId;
  }

  protected String getAIPDownloadPublicUrl(String orderId) {
    return this.configService.getModule().getAccess().getPublicUrl() + SolidifyConstants.URL_SEP + ResourceName.AIP
            + SolidifyConstants.URL_SEP + orderId;
  }
}
