/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Preservation Planning - BasicReplicationService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.replication.basic;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.PreservationPlanningController;
import ch.dlcm.exception.DLCMCompareException;
import ch.dlcm.exception.DLCMReplicationException;
import ch.dlcm.model.ChecksumCheck;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.replication.ReplicationService;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.AipService;
import ch.dlcm.service.DLCMService;
import ch.dlcm.service.rest.fallback.FallbackArchivalInfoPackageRemoteResourceService;

@Service
@ConditionalOnBean(PreservationPlanningController.class)
public class BasicReplicationService extends DLCMService implements ReplicationService {
  private static final Logger log = LoggerFactory.getLogger(BasicReplicationService.class);
  private final String[] archivalStorageNodes;
  private final FallbackArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  private final AipService aipService;
  private final String defaultArchivalStorageUrl;

  public BasicReplicationService(DLCMProperties dlcmProperties, MessageService messageService,
          FallbackArchivalInfoPackageRemoteResourceService aipRemoteResourceService, AipService aipService) {
    super(messageService);
    this.archivalStorageNodes = dlcmProperties.getArchivalStorageUrls();
    this.aipRemoteResourceService = aipRemoteResourceService;
    this.aipService = aipService;
    defaultArchivalStorageUrl = dlcmProperties.getDefaultArchivalStorageUrl();
  }

  /**
   * Check if a AIP is correctly replicated on all storage nodes
   */
  @Override
  public void checkAIP(String aipId) {
    // Load AIPs
    final Map<String, ArchivalInfoPackage> aipList = this.loadAIPs(aipId, true);
    // Find golden copy: latest AIP
    ArchivalInfoPackage sourceAip = this.aipRemoteResourceService.getAIP(aipId);

    // Check if all copies are available
    aipList.entrySet().stream().filter(entry -> entry.getValue() == null).forEach(entry -> {
      final String msg = "Missing copy of AIP (" + sourceAip.getResId() + ") on " + entry.getKey();
      throw new DLCMReplicationException(msg);
    });

    // Check if all copies are the same
    aipList.entrySet().stream().filter(entry -> entry.getValue() != null).forEach(entry -> this.aipService.compare(sourceAip, entry.getValue()));
  }

  /**
   * Replicate a AIP on all storage nodes
   */
  @Override
  public void replicateAIP(String aipId) {
    // Load AIPs
    final Map<String, ArchivalInfoPackage> aipList = this.loadAIPs(aipId, false);
    // Find golden copy: latest AIP
    final ArchivalInfoPackage sourceAip = this.aipRemoteResourceService.getAIP(aipId);
    // Test if missing a copy
    if (aipList.entrySet().stream().filter(entry -> entry.getValue() == null).count() == 0) {
      // Test if all copies already are up-to-date
      try {
        aipList.entrySet().stream().filter(entry -> entry.getValue() != null)
                .forEach(entry -> this.aipService.compare(sourceAip, entry.getValue()));
        throw new SolidifyCheckingException("All AIP copies are up-to-date for " + aipId);
      } catch (DLCMCompareException e) {
        log.info("AIP replication: starting for {} because of {}", aipId, e.getMessage());
      }
    } else {
      log.info("AIP replication: starting for {} because of missing copies", aipId);
    }

    // Replace archive for all existing AIP
    aipList.entrySet()
            .stream()
            .filter(entry -> entry.getValue() != null && entry.getValue().getInfo().getStatus().equals(PackageStatus.COMPLETED))
            .forEach(entry -> {
              try {
                // Compare AIP with source
                this.aipService.compare(sourceAip, entry.getValue());
              } catch (DLCMCompareException e) {
                // Replace AIP
                this.replaceAIPArchive(sourceAip, this.defaultArchivalStorageUrl, entry.getKey());
              } catch (SolidifyRestException | SolidifyResourceNotFoundException e) {
                final String msg = "Error in replicating tombstone AIP (" + sourceAip.getResId() + ") from " + defaultArchivalStorageUrl
                        + " to " + entry.getKey();
                throw new DLCMReplicationException(msg, e);
              }
            });

    // Create a copy for nodes without a copy
    aipList.entrySet()
            .stream()
            .filter(entry -> entry.getValue() == null)
            .forEach(entry -> {
              try {
                this.copyAIP(sourceAip, this.defaultArchivalStorageUrl, entry.getKey());
              } catch (SolidifyRestException | SolidifyResourceNotFoundException e) {
                final String msg = "Error in replicating AIP (" + sourceAip.getResId() + ") from " + defaultArchivalStorageUrl + " to "
                        + entry.getKey();
                throw new DLCMReplicationException(msg, e);
              }
            });
    log.info("AIP replication: completed for {}", aipId);
  }


  /**
   * Create a replica AIP on 'archivalStorage' node by copying AIP from 'masterNode' node
   *
   * @param aip
   * @param masterNode
   * @param archivalStorage
   */
  private void copyAIP(ArchivalInfoPackage aip, String masterNode, String archivalStorage) {
    final String downloadUrl = masterNode + SolidifyConstants.URL_SEP + ResourceName.AIP + SolidifyConstants.URL_SEP + aip.getResId()
            + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD;
    // Reset AIP status
    aip.resetForReplication();
    // Re-initialize the AIP status and checksum checking infos for re-storing
    aip.getInfo().setStatus(PackageStatus.IN_PREPARATION);
    aip.setChecksumCheck(new ChecksumCheck());
    if (Boolean.FALSE.equals(aip.getArchivalUnit())) {
      final List<ArchivalInfoPackage> collection = this.aipRemoteResourceService.getAIPCollectionList(masterNode, aip.getResId());
      aip.setCollection(collection);
    }
    // Create replicated AIP on secondary node
    this.aipRemoteResourceService.createAIP(aip, downloadUrl, aip.getCollectionIds(), archivalStorage);
  }

  /**
   * Update a replica AIP on 'archivalStorage' node by replicating AIP packages from 'masterNode' node
   *
   * @param aip
   * @param masterNode
   * @param archivalStorage
   */
  private void replaceAIPArchive(ArchivalInfoPackage aip, String masterNode, String archivalStorage) {
    final String downloadUrl = masterNode + SolidifyConstants.URL_SEP + ResourceName.AIP + SolidifyConstants.URL_SEP + aip.getResId()
            + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD;
    switch (aip.getInfo().getStatus()) {
      case COMPLETED:
        // Send AIP new package on secondary node
        this.aipRemoteResourceService.replicateAIP(aip, downloadUrl, archivalStorage, DLCMActionName.REPLICATE_PACKAGE);
        break;
      case DISPOSED:
        // Send AIP tombstone on secondary node
        this.aipRemoteResourceService.replicateAIP(aip, downloadUrl, archivalStorage, DLCMActionName.REPLICATE_TOMBSTONE);
        break;
      default:
        throw new DLCMReplicationException("Wrong AIP status for replication: " + aip.getInfo().getStatus());
    }
  }

  /**
   * Load AIP replicas form all archival storage nodes
   *
   * @param aipId
   * @param allAip
   * @return
   */
  private Map<String, ArchivalInfoPackage> loadAIPs(String aipId, boolean allAip) {
    // Check if replication enable
    if (this.archivalStorageNodes.length == 0) {
      throw new DLCMReplicationException(this.messageService.get("replication.disable"));
    }
    // Load AIP for each node
    final Map<String, ArchivalInfoPackage> aipList = new HashMap<>();
    boolean found = false;
    for (final String archivalStorage : this.archivalStorageNodes) {
      try {
        // Load AIP
        final ArchivalInfoPackage aip = this.aipRemoteResourceService.getAIP(archivalStorage, aipId);
        if (allAip || !aip.isInProgress()) {
          found = true;
          aipList.put(archivalStorage, aip);
        } else {
          aipList.put(archivalStorage, null);
        }
      } catch (final SolidifyResourceNotFoundException e) {
        aipList.put(archivalStorage, null);
      }
    }
    if (!found) {
      throw new DLCMReplicationException(this.messageService.get("replication.noaip", new Object[] { aipId }));
    }
    return aipList;
  }

}
