/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - DLCMDownloadTokenController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller;

import java.security.SecureRandom;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Base64;
import java.util.concurrent.ThreadLocalRandom;

import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.authentication.AnonymousAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.PathVariable;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.util.HashTool;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.security.DownloadTokenType;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.DownloadToken;
import ch.dlcm.repository.DownloadTokenRepository;
import ch.dlcm.service.rest.propagate.PropagateUserRemoteResourceService;

@UserPermissions
public class DLCMDownloadTokenController {

  // Token length should be a multiple of three to avoid padding with '=' character that could be
  // problematic in cookie value
  private static final int DOWNLOAD_TOKEN_LENGTH = 201;
  private static final String DOWNLOAD_TOKEN_ANONYMOUS = "anonymous";

  private final DownloadTokenRepository downloadTokenRepository;
  private final PropagateUserRemoteResourceService userRemoteResourceService;
  private final SecureRandom secureRandom;
  private final int downloadTokenLifeTimeMinutes;
  private final boolean secureCookie;

  public DLCMDownloadTokenController(DownloadTokenRepository downloadTokenRepository,
          PropagateUserRemoteResourceService userRemoteResourceService,
          DLCMProperties config) {
    this.downloadTokenRepository = downloadTokenRepository;
    this.userRemoteResourceService = userRemoteResourceService;
    this.downloadTokenLifeTimeMinutes = config.getSecurity().getDownloadTokenLifeTimeMinutes();
    this.secureCookie = config.getSecurity().getDownloadToken().getCookie().getSecure();
    this.secureRandom = new SecureRandom();
  }

  protected ResponseEntity<DownloadToken> getToken(@PathVariable String resourceId, DownloadTokenType type) {
    String userId = DOWNLOAD_TOKEN_ANONYMOUS + ThreadLocalRandom.current().nextInt();
    final Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
    if (authentication != null && !(authentication instanceof AnonymousAuthenticationToken)) {
      User user = this.userRemoteResourceService.getUserAuthenticated();
      userId = user.getResId();
    }
    if (type == null) {
      throw new SolidifyRuntimeException("Need a type defined");
    }

    // Create new download token
    final byte[] bytes = new byte[DOWNLOAD_TOKEN_LENGTH];
    this.secureRandom.nextBytes(bytes);
    final String token = Base64.getEncoder().encodeToString(bytes);
    final String tokenHash = HashTool.hash(token);
    DownloadToken downloadToken = new DownloadToken(userId, type, resourceId, tokenHash, token);

    synchronized (this) {
      // If there is already a download token for this resource and this user replace it
      DownloadToken existingToken = this.downloadTokenRepository.findByResourceIdAndUserId(resourceId, userId);
      if (existingToken != null) {
        this.downloadTokenRepository.delete(existingToken);
      }
      this.downloadTokenRepository.save(downloadToken);
    }
    final HttpHeaders headers = new HttpHeaders();
    final OffsetDateTime expiresDate = ZonedDateTime.now().toInstant().atOffset(ZoneOffset.UTC).plusMinutes(this.downloadTokenLifeTimeMinutes);
    final String expiresString = DateTimeFormatter.RFC_1123_DATE_TIME.format(expiresDate);
    final String secureFlag = secureCookie ? "; Secure" : "";

    headers.add(HttpHeaders.SET_COOKIE, "downloadToken-" + type.getLabel() + "-" + resourceId + "=" + token + "; Expires=" +
            expiresString + "; Path=/" + secureFlag + "; HttpOnly");
    return new ResponseEntity<>(headers, HttpStatus.OK);
  }
}
