/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - DLCMResourceWithLogoController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URLConnection;

import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

import ch.unige.solidify.controller.ResourceController;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.security.UserPermissions;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.ImageInterface;

@UserPermissions
public abstract class DLCMResourceWithLogoController<T extends Resource & ImageInterface> extends ResourceController<T> {

  @ResponseBody
  public HttpEntity<StreamingResponseBody> downloadLogo(@PathVariable String id) throws IOException {
    final T item = this.itemService.findOne(id);
    if (item.getImage() == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    // get content type from the byte[]
    InputStream is = new BufferedInputStream(new ByteArrayInputStream(item.getImage().getImage()));
    String mimeType = URLConnection.guessContentTypeFromStream(is);

    return this.buildDownloadResponseEntity(is, item.getImage().getFileName(), mimeType, item.getImage().getFileSize());
  }

  public HttpEntity<T> uploadLogo(@PathVariable String id, @RequestParam(DLCMConstants.FILE) MultipartFile file) throws IOException {
    T item = this.itemService.findOne(id);
    try {
      // create or update the logo
      item = this.updateOrCreateLogo(item, file);
    } catch (final IOException e) {
      final String message = "Error uploading file " + file;
      throw new SolidifyRuntimeException(message, e);
    }

    this.itemService.save(item);
    return new ResponseEntity<>(item, HttpStatus.OK);
  }

  public HttpEntity<T> deleteLogo(@PathVariable String id) {
    T item = this.itemService.findOne(id);
    item.setImage(null);

    this.itemService.save(item);
    return new ResponseEntity<>(item, HttpStatus.OK);
  }

  private T updateOrCreateLogo(T item, MultipartFile file) throws IOException {
    item.getNewIfNeededImage().setImage(file.getBytes());
    item.getNewIfNeededImage().setFileSize(file.getSize());
    item.getNewIfNeededImage().setFileName(file.getOriginalFilename());
    return item;
  }

}
