/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - DLCMService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.io.IOException;
import java.net.URI;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.logging.LogLevel;

import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyFileDeleteException;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.ZipTool;

import ch.dlcm.model.AbstractDataFile;
import ch.dlcm.model.AbstractDataFile.DataFileStatus;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.model.oais.InfoPackageInterface;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.model.preingest.Deposit;

public abstract class DLCMService {

  protected static final String ERROR_AIP_CONTAINER = "archival.aip.error.container";
  protected static final String LOG_OBJECT_WITH_STATUS_TEMPLATE = "{} '{}' ({}) (status: {}): {}";

  protected static final long BAGIT_INTERNAL_FILE_NUMBER = 8;
  protected static final long ZIP_COLLECTION_FILE_NUMBER = 1;
  protected static final long BAGIT_COLLECTION_FILE_NUMBER = BAGIT_INTERNAL_FILE_NUMBER + ZIP_COLLECTION_FILE_NUMBER;

  private static final Logger log = LoggerFactory.getLogger(DLCMService.class);

  protected final MessageService messageService;

  protected DLCMService(MessageService messageService) {
    this.messageService = messageService;
  }

  protected <T extends Resource> boolean belongsToList(T item, List<T> list) {
    return list.stream().anyMatch(o -> item.getResId().equals(o.getResId()));
  }

  protected long getPackageFileNumber(ArchiveContainer aipContainer, URI packageUri) {
    final ZipTool zip = new ZipTool(packageUri);
    switch (aipContainer) {
      case BAG_IT:
        return zip.getEntryNumber() - BAGIT_INTERNAL_FILE_NUMBER;
      case ZIP:
        return zip.getEntryNumber();
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { aipContainer }));
    }
  }

  protected void inError(InfoPackageInterface infoPackage, String msg) {
    infoPackage.getInfo().setStatusWithMessage(PackageStatus.IN_ERROR, msg);
    log.error("{} for {} {}", msg, this.getInfoPackageType(infoPackage), infoPackage.getResId());
  }

  protected void inError(InfoPackageInterface infoPackage, String msg, String cause) {
    infoPackage.getInfo().setStatusWithMessage(PackageStatus.IN_ERROR, msg + " : " + cause);
    log.error("{} ({}) for {} {}", msg, cause, this.getInfoPackageType(infoPackage), infoPackage.getResId());
  }

  protected void deleteFile(URI fileToDelete) {
    try {
      FileTool.deleteFile(fileToDelete);
    } catch (final IOException e) {
      throw new SolidifyFileDeleteException(fileToDelete.toString(), e);
    }
  }

  private String getInfoPackageType(InfoPackageInterface infoPackage) {
    if (infoPackage instanceof SubmissionInfoPackage) {
      return "SIP";
    } else if (infoPackage instanceof ArchivalInfoPackage) {
      return "AIP";
    } else if (infoPackage instanceof DisseminationInfoPackage) {
      return "DIP";
    }
    return "[unknown information package type]";
  }

  protected void logWillBeProcessedPackage(InfoPackageInterface infoPackage) {
    this.logPackageMessage(LogLevel.INFO, infoPackage, "will be processed");
  }

  protected void logProcessedPackage(InfoPackageInterface infoPackage) {
    if (infoPackage.getInfo().getStatus() == PackageStatus.COMPLETED) {
      this.logPackageMessage(LogLevel.INFO, infoPackage, "processed");
    } else if (infoPackage.getInfo().getStatus() == PackageStatus.IN_ERROR) {
      this.logPackageMessage(LogLevel.WARN, infoPackage, "processed (statusMessage: " + infoPackage.getInfo().getStatusMessage() + ")");
    } else {
      this.logPackageMessage(LogLevel.DEBUG, infoPackage, "processed");
    }
  }

  protected void logPackageMessage(LogLevel logLevel, InfoPackageInterface infoPackage, String message, Exception... e) {
    this.logObjectWithStatusMessage(logLevel, this.getInfoPackageType(infoPackage), infoPackage.getResId(),
            infoPackage.getInfo().getStatus().toString(), infoPackage.getInfo().getName(), message, e);
  }

  protected void logDepositMessage(LogLevel logLevel, Deposit deposit, String message, Exception... e) {
    this.logObjectWithStatusMessage(logLevel, "Deposit", deposit.getResId(), deposit.getStatus().toString(), deposit.getTitle(), message, e);
  }

  protected void logProcessedDeposit(Deposit deposit) {
    if (deposit.getStatus() == Deposit.DepositStatus.COMPLETED) {
      this.logDepositMessage(LogLevel.INFO, deposit, "processed");
    } else if (deposit.getStatus() == Deposit.DepositStatus.IN_ERROR) {
      this.logDepositMessage(LogLevel.WARN, deposit, "processed (statusMessage: " + deposit.getStatusMessage() + ")");
    } else {
      this.logDepositMessage(LogLevel.DEBUG, deposit, "processed");
    }
  }

  protected void logDataFileMessage(LogLevel logLevel, AbstractDataFile df, String message, Exception... e) {
    this.logObjectWithStatusMessage(logLevel, df.getClass().getSimpleName(), df.getResId(), df.getStatus().toString(), df.getFileName(), message,
            e);
  }

  protected void logProcessedDataFile(AbstractDataFile df) {
    if (df.getStatus() == DataFileStatus.READY) {
      this.logDataFileMessage(LogLevel.INFO, df, "processed");
    } else if (df.getStatus() == DataFileStatus.IN_ERROR) {
      this.logDataFileMessage(LogLevel.WARN, df, "processed (statusMessage: " + df.getStatusMessage() + ")");
    } else {
      this.logDataFileMessage(LogLevel.DEBUG, df, "processed");
    }
  }

  private void logObjectWithStatusMessage(LogLevel logLevel, String className, String resId, String status, String name, String message,
          Exception... exceptions) {
    switch (logLevel) {
      case TRACE:
        log.trace(LOG_OBJECT_WITH_STATUS_TEMPLATE, className, resId, name, status, message);
        break;
      case DEBUG:
        log.debug(LOG_OBJECT_WITH_STATUS_TEMPLATE, className, resId, name, status, message);
        break;
      case WARN:
        log.warn(LOG_OBJECT_WITH_STATUS_TEMPLATE, className, resId, name, status, message);
        break;
      case ERROR:
        Exception e = (exceptions.length > 0) ? exceptions[0] : null;
        log.error(LOG_OBJECT_WITH_STATUS_TEMPLATE, className, resId, name, status, message, e);
        break;
      case INFO:
      default:
        log.info(LOG_OBJECT_WITH_STATUS_TEMPLATE, className, resId, name, status, message);
        break;
    }
  }
}
