/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - RegistryService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.jms.annotation.JmsListener;
import org.springframework.stereotype.Service;

import ch.unige.solidify.model.ChangeInfo;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.message.RegistryMessage;
import ch.dlcm.model.index.RegistryMetadata;
import ch.dlcm.service.rest.trusted.TrustedRegistryMetadataRemoteResourceService;

@Service
@ConditionalOnProperty(value = "dlcm.parameters.registryEnabled", havingValue = "true")
public class RegistryService extends DLCMService {
  private static final Logger log = LoggerFactory.getLogger(RegistryService.class);

  private final String registryName;

  private final TrustedRegistryMetadataRemoteResourceService registryMetadataRemoteResourceService;

  public RegistryService(DLCMProperties dlcmProperties, MessageService messageService,
          TrustedRegistryMetadataRemoteResourceService registryMetadataRemoteResourceService) {
    super(messageService);
    this.registryName = dlcmProperties.getIndexing().getRegistryName();
    this.registryMetadataRemoteResourceService = registryMetadataRemoteResourceService;
  }

  @JmsListener(destination = "${dlcm.queue.registry}")
  public void registry(RegistryMessage event) {
    if (log.isTraceEnabled()) {
      log.trace("Reading message {}", event);
    }
    try {
      final RegistryMetadata item = this.convertMessage(event);
      // Test if creation
      if (!this.registryMetadataRemoteResourceService.checkItem(event.getResId())) {
        this.registryMetadataRemoteResourceService.indexItem(event.getResId(), item);
      } else {
        this.registryMetadataRemoteResourceService.updateItem(event.getResId(), item);
      }
    } catch (final Exception e) {
      log.error("Error in message " + event, e);
    }
  }

  private void addChangeInfo(Map<String, Object> info, String label, ChangeInfo changeInfo) {
    if (changeInfo != null) {
      info.put(label, changeInfo);
    }
  }

  private RegistryMetadata convertMessage(RegistryMessage msg) {
    final RegistryMetadata regMetadata = new RegistryMetadata();
    regMetadata.setIndex(this.registryName);
    regMetadata.setResId(msg.getResId());
    final Map<String, Object> info = new HashMap<>();
    info.put("type", msg.getType());
    this.addChangeInfo(info, "creation", msg.getCreation());
    this.addChangeInfo(info, "lastUpdate", msg.getLastUpdate());
    this.addChangeInfo(info, "disposition", msg.getDisposition());
    if (!msg.getMetadata().isEmpty()) {
      info.putAll(msg.getMetadata());
    }
    regMetadata.setMetadata(info);
    return regMetadata;
  }

}
