/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - RegistryMetadataRemoteResourceService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.rest.abstractservice;

import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Pageable;
import org.springframework.web.util.UriComponentsBuilder;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.service.SolidifyRestClientService;
import ch.unige.solidify.service.rest.abstractservice.IndexMetadataRemoteResourceService;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.index.RegistryMetadata;
import ch.dlcm.rest.ResourceName;

public abstract class RegistryMetadataRemoteResourceService extends IndexMetadataRemoteResourceService<String, RegistryMetadata> {
  private static final Logger log = LoggerFactory.getLogger(RegistryMetadataRemoteResourceService.class);

  private final DLCMProperties config;

  protected RegistryMetadataRemoteResourceService(DLCMProperties config, SolidifyRestClientService dlcmRestClientService) {
    super(RegistryMetadata.class, dlcmRestClientService);
    this.config = config;
  }

  @Override
  protected String getIndexUrl(String index) {
    return this.config.getModule().getDataMgmt().getUrl() + SolidifyConstants.URL_SEP + ResourceName.ITEM;

  }

  public RegistryMetadata getMetadata(String index, String resId) {
    return this.getIndexMetadata(index, resId);
  }

  @Override
  public RestCollection<RegistryMetadata> getIndexMetadataList(String index, Pageable pageable) {
    return this.getIndexMetadataList(index, new HashMap<>(), pageable);
  }

  @Override
  protected String buildUrl(String index, Map<String, Object> parameters) {
    String url = this.getIndexUrl(index) + SolidifyConstants.URL_SEP + ActionName.SEARCH;
    UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);

    if (parameters != null) {

      this.completeUri(uriBuilder, parameters);

      if (!StringTool.isNullOrEmpty((String) parameters.get(DLCMConstants.SEARCH_ORG_UNIT))) {
        uriBuilder.queryParam(DLCMConstants.SEARCH_ORG_UNIT, parameters.get(DLCMConstants.SEARCH_ORG_UNIT));
      }
    }

    return uriBuilder.toUriString();
  }

  public boolean indexItem(String id, RegistryMetadata md) {
    return this.createIndex(ResourceName.ITEM, id, md);
  }

  public boolean checkItem(String id) {
    return this.getIndex(ResourceName.ITEM, id);
  }

  public boolean updateItem(String id, RegistryMetadata md) {
    return this.updateIndex(ResourceName.ITEM, id, md);
  }

  private boolean createIndex(String resource, String id, RegistryMetadata md) {
    try {
      RegistryMetadata res = this.restClientService.postResource(
              this.config.getModule().getDataMgmt().getUrl() + SolidifyConstants.URL_SEP + resource,
              md, RegistryMetadata.class);
      if (res == null) {
        return false;
      }

    } catch (SolidifyResourceNotFoundException | SolidifyRestException e) {
      log.error("Error in indexing (" + id + ")", e);
      return false;
    }
    return true;
  }

  private boolean getIndex(String resource, String resId) {
    this.restClientService.checkResId(resId);
    try {
      RegistryMetadata res = this.restClientService.getResource(this.config.getModule().getDataMgmt().getUrl()
              + SolidifyConstants.URL_SEP + resource + SolidifyConstants.URL_SEP + resId,
              RegistryMetadata.class);
      if (res == null) {
        return false;
      }
    } catch (SolidifyResourceNotFoundException e) {
      return false;
    } catch (SolidifyRestException e) {
      log.error("Error in getting " + resource + " (" + resId + ")", e);
      return false;
    }
    return true;
  }

  private boolean updateIndex(String resource, String id, RegistryMetadata md) {
    try {
      this.restClientService.patchResource(this.config.getModule().getDataMgmt().getUrl() + SolidifyConstants.URL_SEP
              + resource + SolidifyConstants.URL_SEP + id, md, RegistryMetadata.class);
    } catch (SolidifyResourceNotFoundException | SolidifyRestException e) {
      log.error("Error in updating index (" + id + ")", e);
      return false;
    }
    return true;
  }

}
