/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Security - DownloadTokenPermissionService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.security;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;

import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.HashTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.security.DownloadTokenType;
import ch.dlcm.model.settings.DownloadToken;
import ch.dlcm.repository.DownloadTokenRepository;

@Service
public class DownloadTokenPermissionService {

  private final DownloadTokenRepository downloadTokenRepository;

  public DownloadTokenPermissionService(DownloadTokenRepository downloadTokenRepository) {
    this.downloadTokenRepository = downloadTokenRepository;
  }

  public boolean isAllowed(String resourceId, DownloadTokenType type) {
    if (this.canBypassPermissionCheck()) {
      return true;
    }

    ServletRequestAttributes servletRequestAttributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
    if (servletRequestAttributes == null) {
      throw new SolidifyRuntimeException("Can't get servlet request attributes");
    }

    HttpServletRequest httpServletRequest = servletRequestAttributes.getRequest();
    String tokenString = null;

    if (httpServletRequest.getCookies() == null) {
      throw new IllegalStateException("Download token cookie needed");
    }
    for (Cookie cookie : httpServletRequest.getCookies()) {
      String cookieName = cookie.getName();
      String cookieTypeAndResourceId = cookieName.substring(cookieName.indexOf('-') + 1);
      if (cookieTypeAndResourceId.equals(type.getLabel() + "-" + resourceId)) {
        tokenString = cookie.getValue();
      }
    }
    if (tokenString == null) {
      return false;
    }
    String tokenHash = HashTool.hash(tokenString);
    DownloadToken downloadToken = this.downloadTokenRepository.findByTokenHash(tokenHash);
    if (downloadToken != null && downloadToken.getType().equals(type) && downloadToken.getResourceId().equals(resourceId)) {
      return true;
    }
    return false;
  }

  private boolean canBypassPermissionCheck() {
    final String[] authorizedApplicationRoleNames = DLCMConstants.getRootAndTrustedRoleList();

    final Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
    final java.util.Collection<? extends GrantedAuthority> authorities = authentication.getAuthorities();

    for (final String roleName : authorizedApplicationRoleNames) {
      for (final GrantedAuthority authority : authorities) {
        if (authority.toString().equals(roleName)) {
          return true;
        }
      }
    }

    return false;
  }
}
