/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Solution - DepositDoc.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.resource;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.JsonFieldType;

import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.test.helper.AbstractResourceDoc;
import ch.unige.solidify.test.helper.FieldDescriptorBuilder;
import ch.unige.solidify.test.model.TypeField;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.model.Access;
import ch.dlcm.model.ComplianceLevel;
import ch.dlcm.model.DataTag;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.settings.ArchiveType;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.rest.UrlPath;

public class DepositDoc extends AbstractResourceDoc {

  @Override
  public List<String> getListLinks() {
    return Arrays.asList(ResourceName.SCHEMA, ActionName.VALUES);
  }

  @Override
  public String getName() {
    return UrlPath.PREINGEST_DEPOSIT;
  }

  @Override
  public List<String> getResourceLinks() {
    return Arrays.asList(ResourceName.CONTRIBUTOR, ResourceName.DATAFILE, ResourceName.AIP, ResourceName.SUBMISSION_AGREEMENT,
            ResourceName.LICENSE, ActionName.UPLOAD, ActionName.UPLOAD_ARCHIVE, ActionName.DOWNLOAD, ActionName.HISTORY);
  }

  @Override
  protected List<FieldDescriptor> getDocumentationFields() {
    // @formatter:off
    final List<FieldDescriptor> documentationFields =  new ArrayList<>(Arrays.asList(
      new FieldDescriptorBuilder(this)
        .name("title")
        .description("The title of the deposit")
        .mandatory(true)
        .type(TypeField.STRING)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("description")
        .description("The description of the deposit")
        .mandatory(true)
        .type(TypeField.STRING)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("publicationDate")
        .description("The publication date of the deposit")
        .mandatory(true)
        .type(TypeField.DATE)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("collectionBegin")
        .description("The beginning of the deposit collection")
        .type( TypeField.DATE_TIME)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("collectionEnd")
        .description("The end of the deposit collection")
        .type(TypeField.DATE_TIME)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("access")
        .description("The access level of the deposit")
        .mandatory(true)
        .enumValues(this.enumToList(Access.values()))
        .testValue(Access.PUBLIC.toString())
        .build(),
      new FieldDescriptorBuilder(this)
        .name("dataSensitivity")
        .description("The data sensitivity of the deposit")
        .mandatory(true)
        .enumValues(this.enumToList(DataTag.values()))
        .testValue(DataTag.BLUE.toString())
        .build(),
      new FieldDescriptorBuilder(this)
        .name("status")
        .description("The status of the deposit")
        .mandatory(true)
        .enumValues(this.enumToList(DepositStatus.values()))
        .testValue(DepositStatus.IN_PROGRESS.toString())
        .build(),
      new FieldDescriptorBuilder(this)
        .name("statusMessage")
        .description("The message related to the status")
        .type(TypeField.STRING)
        .build(),
      new FieldDescriptorBuilder(this)
        .name(DLCMConstants.DB_SIP_ID)
        .description("The generated SIP DLCM ID of the deposit")
        .type(TypeField.STRING)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("preparationId")
        .description("The preparation ID of the deposit")
        .type(TypeField.STRING)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("doi")
        .description("The Digital Object Identifier (DOI) of the deposit")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_DOI)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("language")
        .description("The language object of the deposit")
        .type(TypeField.UNDEFINED)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("languageId")
        .description("The language ID of the deposit")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_ID)
        .testValue("EN")
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("agreement")
        .description("The submission agreement of the deposit")
        .type(TypeField.BOOLEAN)
        .build().type(JsonFieldType.BOOLEAN),
      new FieldDescriptorBuilder(this)
        .name("submissionAgreementId")
        .description("The submission agreement of the deposit")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_ID)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("licenseId")
        .description("The license ID of the deposit")
        .type(TypeField.STRING)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("organizationalUnit")
        .description("The organizational unit object of the deposit")
        .type(TypeField.UNDEFINED)
        .build(),
      new FieldDescriptorBuilder(this)
        .name(DLCMConstants.ORG_UNIT_ID_FIELD)
        .description("The organizational unit ID of the deposit")
        .mandatory(true)
        .type(TypeField.STRING)
        .testValue(TEST_STRING_VALUE)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("submissionPolicy")
        .description("The submission policy object of the deposit")
        .type(TypeField.UNDEFINED)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("submissionPolicyId")
        .description("The submission policy ID of the deposit")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_ID)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("preservationPolicy")
        .description("The preservation policy object of the deposit")
        .type(TypeField.UNDEFINED)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("preservationPolicyId")
        .description("The preservation policy ID of the deposit")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_ID)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("keywords")
        .description("List of deposit keywords")
        .type(TypeField.ARRAY)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("complianceLevel")
        .description("The compliance level of the deposit")
        .enumValues(this.enumToList(ComplianceLevel.values()))
        .testValue(ComplianceLevel.NOT_ASSESSED.toString())
        .build(),
      new FieldDescriptorBuilder(this)
        .name("hasEmbargo")
        .description("If the deposit has an embargo")
        .type(TypeField.BOOLEAN)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("containsUpdatedMetadata")
        .description("If the deposit contains updated metadata")
        .type(TypeField.BOOLEAN)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("dataFileNumber")
        .description("Number of deposit data files")
        .type(TypeField.INT32)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("collectionSize")
        .description("Size of AIP collection")
        .type(TypeField.INT32)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("metadataVersion")
        .description("The metadata version of the deposit")
        .enumValues(this.enumToList(DLCMMetadataVersion.values()))
        .testValue(DLCMMetadataVersion.getDefaultVersion().getVersion())
        .build(),
      new FieldDescriptorBuilder(this)
        .name("additionalFieldsValues")
        .description("Some eventual additional fields values generated through an additional fields form")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_JSON)
        .testValue(TEST_JSON_ARRAY_VALUE)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("additionalFieldsFormId")
        .description("The additional fields form ID used to generate the additional fields values")
        .type(TypeField.STRING)
        .customConstraints("Mandatory if additionalFieldsValues are providen")
        .build(),
      new FieldDescriptorBuilder(this)
        .name("archiveType")
        .description("The archive type object of the deposit")
        .type(TypeField.UNDEFINED)
        .build(),
      new FieldDescriptorBuilder(this)
        .name("isIdenticalTo")
        .description("Indicates that Archive A is identical to B, used to register two separate instances of the same resource")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_DOI)
        .testValue(TEST_DOI_VALID)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("isReferencedBy")
        .description("Indicates A is used as a source of information by B")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_DOI)
        .testValue(TEST_DOI_VALID)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("isObsoletedBy")
        .description("Indicates A is replaced by B")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_DOI)
        .testValue(TEST_DOI_VALID)
        .build().type(JsonFieldType.STRING),
      new FieldDescriptorBuilder(this)
        .name("archiveTypeId")
        .description("The archive type ID of the deposit")
        .type(TypeField.STRING)
        .customConstraints(CONSTRAINT_VALID_ID)
        .testValue(ArchiveType.DATASET_ID)
        .build().type(JsonFieldType.STRING)));
    // @formatter:on
    documentationFields.addAll(new EmbargoInfoDoc().getDocumentationFields());
    return documentationFields;
  }

}
